using System;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;
using System.Diagnostics;

using NAssess.Core.Evaluating;
using NerdBank.Tools;
using NerdBank.Tools.Exslt;

namespace NAssess.Core.Engines.Evaluation
{
	/// <summary>
	/// The standard NAssess render for HTML evaluations.
	/// </summary>
	public class StandardHtml : NAssess.Core.Evaluating.HtmlRenderingEngine
	{
		private const string StandardHtmlXsltExtensionsNamespace = "http://nassess.byu.edu/rendering/evaluation/StandardHtml";

		#region Construction
		public StandardHtml()
		{
		}

		#endregion
		#region Attributes (that affect the style of rendering for the evaluation)
		internal const bool GroupSelectOneOptionsOnLeftDefault = true;
		/// <summary>
		/// Whether the radio buttons in a grid appear to the left of the question text.
		/// The default is true.
		/// </summary>
		public bool GroupSelectOneOptionsOnLeft = GroupSelectOneOptionsOnLeftDefault;
		internal const int GroupSelectOneOptionFrequencyDefault = 4;
		/// <summary>
		/// The number of rows of options between each header row
		/// that reminds the user of the response labels.  Default is 4.
		/// </summary>
		public int GroupSelectOneOptionFrequency = GroupSelectOneOptionFrequencyDefault;
		internal const bool IncludeClientValidationDefault = true;
		/// <summary>
		/// Whether the questionnaire should stream to the browser with
		/// client-side validations cripts to help the user not miss any questions.
		/// </summary>
		public bool IncludeClientValidation = IncludeClientValidationDefault;
		/// <summary>
		/// Gets/sets the length of the longest select-one response label
		/// before a drop down box becomes a radio list. Default is 75.
		/// </summary>
		public int DropdownToRadioListThreshold = 75;
		#endregion
		#region Properties
		private NAssess.Core.Evaluating.Evaluation evaluation;
		public override NAssess.Core.Evaluating.Evaluation Evaluation
		{
			get { return evaluation; }
			set { evaluation = value; }
		}
		/// <summary>
		/// Gets the number of pages that are in the questionnaire.  
		/// This number may depend on the responses already provided, and thus
		/// may change with each submission of responses.
		/// </summary>
		public override int PagesCount
		{
			get
			{
				if (Evaluation == null)
					throw new InvalidOperationException(Resources.EvaluationPropertyMustBeSetFirst);

				// Each section in the DDI is a page as HTML.  
				// So just count sections, as long as any conditionals 
				// within those sections are true.
				XPathNodeIterator nlSections = Evaluation.Codebook.Select("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='section']");
				int pageCount = 0;
				foreach (XPathNavigator section in nlSections)
					if (IsVarGrpActiveBasedOnData(section))
						pageCount++; // a qualifying section
				return pageCount;
			}
		}
		protected virtual bool IsVarGrpActiveBasedOnData(XPathNavigator varGrpNode)
		{
			// Fetch the condition tag
			XPathNavigator nCondition = Ddi.SelectSingleNode("ddi:defntn", varGrpNode);
			if (nCondition != null)
			{
				// A condition has been set.
				// Tokenize conditional string
				string[] tsCondition = StringUtilities.Tokenize(nCondition.Value);
				// Test the condition, always testing for the opposite condition as 
				// prescribed, so that we can immediately return false if the
				// condition is not satisfied.
				if (tsCondition[0] != "orLogic")
				{
					for (int i = 1; i < tsCondition.Length; i += 3)
					{
						object evalValue = Evaluation.Responses[tsCondition[i - 1]];
						string strValue = evalValue == null ? "" : evalValue.ToString();
						string cmpValue = tsCondition[i + 1];
						switch (tsCondition[i])
						{
							case "=": // an equality test
							case "==": // an equality test
								if (strValue != cmpValue) return false;
								break;
							case "<>": // an inequality test
							case "!=": // an inequality test
								if (strValue == cmpValue) return false;
								break;
							case ">":
								if (strValue == "" || Convert.ToDouble(strValue) <= Convert.ToDouble(cmpValue)) return false;
								break;
							case ">=":
								if (strValue == "" || Convert.ToDouble(strValue) < Convert.ToDouble(cmpValue)) return false;
								break;
							case "<":
								if (strValue == "" || Convert.ToDouble(strValue) >= Convert.ToDouble(cmpValue)) return false;
								break;
							case "<=":
								if (strValue == "" || Convert.ToDouble(strValue) > Convert.ToDouble(cmpValue)) return false;
								break;
						}
					}
				}
				// Test the condition as 
				// prescribed, so that we can immediately return true if the
				// condition is satisfied.
				else
				{
					bool result = false;
					for (int i = 4; i < tsCondition.Length && !result; i += 3)
					{
						object evalValue = Evaluation.Responses[tsCondition[i - 1]];
						string strValue = evalValue == null ? "" : evalValue.ToString();
						string cmpValue = tsCondition[i + 1];
						switch (tsCondition[i])
						{
							case "=": // an equality test
							case "==": // an equality test
								if (strValue == cmpValue) result = true;
								break;
							case "<>": // an inequality test
							case "!=": // an inequality test
								if (strValue != cmpValue) result = true;
								break;
							case ">":
								if (strValue == "" || Convert.ToDouble(strValue) > Convert.ToDouble(cmpValue)) result = true;
								break;
							case ">=":
								if (strValue == "" || Convert.ToDouble(strValue) >= Convert.ToDouble(cmpValue)) result = true;
								break;
							case "<":
								if (strValue == "" || Convert.ToDouble(strValue) < Convert.ToDouble(cmpValue)) result = true;
								break;
							case "<=":
								if (strValue == "" || Convert.ToDouble(strValue) <= Convert.ToDouble(cmpValue)) result = true;
								break;
						}
					}
					if (!result)
						return false;
				}

			}
			// This varGrp passes its own tests.  But what about child varGrp's?  
			// If ALL our children varGrp's are hidden as a result of data,
			// then we might as well be hidden as well.  

			// If there are variables directly in this varGrp, always display,
			// as this is a "leaf node" variable group.
			if (varGrpNode.GetAttribute("var", "").Length > 0) return true;
			// If there are no variables, check descending varGrp's first.
			if (varGrpNode.GetAttribute("varGrp", "").Length > 0)
			{
				// Get an array of all child varGrps
				string[] varGrpChildren = StringUtilities.Tokenize(varGrpNode.GetAttribute("varGrp", ""));
				// Assemble that array into a single XPath query
				StringBuilder searchString = new StringBuilder();
				searchString.Append("../ddi:varGrp[");
				foreach (string varGrpID in varGrpChildren)
					searchString.AppendFormat("@ID='{0}' or ", varGrpID);
				searchString.Length -= 4; // remove last OR
				searchString.Append("]");
				XPathNodeIterator nlVarGrpChildren = Ddi.Select(searchString.ToString(), varGrpNode);
				while (nlVarGrpChildren.MoveNext())
					// If at least one child varGrp is active, then we are
					if (IsVarGrpActiveBasedOnData(nlVarGrpChildren.Current)) return true;
				// No child varGrp was active.  Do not display us (the parent)
				return false;
			}

			// We don't expect the next statement to get executed, but if 
			// something unexpected comes our way, default to displaying this varGrp
			return true;
		}
		#endregion
		#region Operations
		/// <summary>
		/// Gets the raw HTML that represents a given page of this evaluation.
		/// If any answers have already been provided for the requested page,
		/// those answers will be preset in the HTML.
		/// </summary>
		/// <param name="page">
		/// The page number of the evaluation desired.  1-based index.
		/// </param>
		/// <param name="writer">
		/// The <see cref="XmlWriter"/> to use to write the resulting HTML.
		/// </param>
		/// <returns>
		/// The actual HTML of the evaluation, surrounded by a DIV tag.
		/// </returns>
		public override void ToHtml(int page, XmlWriter writer)
		{
			if (page < 1 || page > PagesCount)
				throw new ArgumentOutOfRangeException("page", page, Resources.PageOutOfRange);
			if (Evaluation == null)
				throw new InvalidOperationException(Resources.EvaluationPropertyMustBeSetFirst);

			PrepareRenderingSpecifics(page);

			XslCompiledTransform xslt = prepareXslt();
			xslt.Transform(prepareSource(), XsltArguments, writer);
		}

		[WebDirect]
		public static Stream GetValidationJS(out string contentType)
		{
			contentType = "text/javascript";
			string js = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Engines/Evaluation/xslt.html/validation.js", Project.DefaultNamespace);
			js = LocalizeJavaScript(js, Resources.ResourceManager);
			return StringToStream(js);
		}
		private static Stream StringToStream(string s)
		{
			MemoryStream ms = new MemoryStream(s.Length);
			StreamWriter sw = new StreamWriter(ms);
			sw.Write(s);
			sw.Flush(); // yes, this is necessary
			ms.Position = 0;
			return ms;
		}
		[WebDirect]
		[Obsolete("Use Page.RegisterOnSubmitStatement instead.")]
		public static Stream GetAspNetValidationJS(out string contentType)
		{
			Stream stream = EmbeddedResources.GetFileStreamFromAssembly("/Engines/Evaluation/xslt.html/aspnetValidation.js", Project.DefaultNamespace);
			contentType = "text/javascript";
			return stream;
		}

		private XPathNavigator GetSectionVarGrpToDisplay(int Page)
		{
			if (Page < 1) throw new ArgumentOutOfRangeException("Page", Page, Resources.PositiveValueRequired);
			XPathNodeIterator nlDisplayOrder = Evaluation.Codebook.Select("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='display']");
			if (nlDisplayOrder.Count != 1) // there must be exactly one display variable group to control order
				throw new HtmlRenderingException("There must be exactly one variable group of type \"display\" to define display order.");
			nlDisplayOrder.MoveNext();
			XPathNavigator nDisplayOrder = nlDisplayOrder.Current;
			XPathNavigator nDisplayOrderLabel = Ddi.SelectSingleNode("ddi:labl", nDisplayOrder);
			string sDisplayOrderLabel = null;
			if (nDisplayOrderLabel != null) sDisplayOrderLabel = nDisplayOrderLabel.Value.Trim();
			if (nDisplayOrder.GetAttribute("var", "").Length > 0)
				throw new HtmlRenderingException("There must be no variables defined as members of the \"display\" variable group \"" + sDisplayOrderLabel + "\" that defines order.");
			if (nDisplayOrder.GetAttribute("varGrp", "").Length == 0)
				throw new HtmlRenderingException("There must be a varGrp attribute defined in the \"display\" variable group \"" + sDisplayOrderLabel + "\".");

			// Extract space-delimited string of variable group names 
			string sDisplayOrderVarGrps = nDisplayOrder.GetAttribute("varGrp", "");
			// Tokenize string to get individual variable groups in order
			string[] OrderedSectionIDs = sDisplayOrderVarGrps.Split(' ');
			// Bring up the section we are to display
			XPathNavigator nDisplaySection = null;
			for (int candidatePage = 0; candidatePage < Page; candidatePage++)
			{
				nDisplaySection = Evaluation.Codebook.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@ID='" + OrderedSectionIDs[candidatePage] + "' and @type='section']");
				// If the considered page here is hidden, then it doesn't count in the page count,
				// and we should therefore skip forward.
				if (!IsVarGrpActiveBasedOnData(nDisplaySection))
					Page++;
			}
			if (nDisplaySection == null)
				throw new HtmlRenderingException("Could not find varGrp with ID=" + OrderedSectionIDs[Page - 1] + " and type=section");

			return nDisplaySection;
		}

		protected override void InitXsltArguments()
		{
			base.InitXsltArguments();

			PrepareResponses();
			PrepareXsltExtensions();
		}

		private bool prepareResponsesChangedHooked;
		private void PrepareResponses()
		{
			// The user responses act as an Xslt argument
			XmlDocument xmlResponses = new XmlDocument();
			xmlResponses.AppendChild(Evaluation.Responses.ToXml(xmlResponses));
			XsltArguments.AddParam("responses", string.Empty, xmlResponses.CreateNavigator().Select("."));
			if (!prepareResponsesChangedHooked) // don't double up on this event handler
				Evaluation.Responses.Changed += delegate
				{
					XsltArguments.RemoveParam("responses", string.Empty);
					PrepareResponses();
				};
			prepareResponsesChangedHooked = true;
		}
		private void PrepareXsltExtensions()
		{
			// Add extension objects to provide additional functionality
			XsltArguments.AddExtensionObject("http://fhss.byu.edu/exslt/strings", new ExsltStrings());
			XsltArguments.AddExtensionObject(StandardHtmlXsltExtensionsNamespace, new StandardHtmlXsltExtensionFunctions(this));
		}
		private void PrepareRenderingSpecifics(int Page)
		{
			XsltArgumentList args = XsltArguments;
			// Translate the page number into a section to render
			XPathNavigator nDisplaySection = GetSectionVarGrpToDisplay(Page);
			// Tell XSLT which section to render
			args.RemoveParam("section", string.Empty);
			args.AddParam("section", string.Empty, nDisplaySection.GetAttribute("ID", ""));

			// Pass on WebDirect Url
			args.RemoveParam("WebDirectUrl", string.Empty);
			if (WebDirectUrl != null) args.AddParam("WebDirectUrl", string.Empty, WebDirectUrl.ToString());

			// Pass on certain visual settings
			args.RemoveParam("GroupSelectOneOptionsOnLeft", string.Empty);
			args.AddParam("GroupSelectOneOptionsOnLeft", string.Empty, GroupSelectOneOptionsOnLeft);
			args.RemoveParam("varGrpSelectOneOptionFrequency", string.Empty);
			args.AddParam("varGrpSelectOneOptionFrequency", string.Empty, GroupSelectOneOptionFrequency);
			args.RemoveParam("DropdownToRadioListThreshold", string.Empty);
			args.AddParam("DropdownToRadioListThreshold", string.Empty, DropdownToRadioListThreshold);

			// Pass on other switches
			args.RemoveParam("clientvalidation", string.Empty);
			args.AddParam("clientvalidation", string.Empty, IncludeClientValidation);
		}
		private XPathNavigator prepareSource()
		{
			// The DDI acts as the source document
			return Evaluation.Codebook.XPath.CreateNavigator();
		}
		private static XmlResolver prepareResolver()
		{
			// Our custom InProjectResolver allows XML files to reference files 
			// embedded in the assembly.
			return new ProjectXmlResolver(Project.DefaultNamespace);
		}
		private const string xsltStartPoint = "/Engines/Evaluation/xslt.html/main.xslt";
		private const string xsltCacheName = "Xslt: " + xsltStartPoint;
		private static XslCompiledTransform prepareXslt()
		{
			// The XSLT produces the Html that we output
			XslCompiledTransform xslt;
			// First try to pull it out of the cache
			xslt = Project.GetFromCacheIfAvailable<XslCompiledTransform>(xsltCacheName);
			if (xslt == null)
			{
				xslt = new XslCompiledTransform(Project.IsXsltDebuggingEnabled);
				XsltSettings settings = new XsltSettings(true, true);
				using (Stream mainXslt = EmbeddedResources.GetFileStreamFromAssembly(xsltStartPoint, Project.DefaultNamespace))
					xslt.Load(new XmlTextReader(mainXslt), settings, prepareResolver());
				Project.SaveInCacheIfAvailable<XslCompiledTransform>(xsltCacheName, xslt);
			}
			return xslt;
		}
		/// <summary>
		/// Gets a list of the headings for each page in the questionnaire.
		/// </summary>
		/// <remarks>
		/// This list can be used to generate questionnaire navigation links.
		/// </remarks>
		public string[] GetPageHeadings()
		{
			if (Evaluation == null) throw new InvalidOperationException(Resources.EvaluationPropertyMustBeSetFirst);

			// Use the <ddi:labl> tags of <ddi:varGrp type=section> tags to generate the headings
			// for each page.  In our HTML generating XSLT files, these tags make up the 
			// heading that runs across the top of each page.
			string[] sHeadings = new string[PagesCount];
			for (int i = 0; i < sHeadings.Length; i++)
			{
				XPathNavigator nVarGrp = GetSectionVarGrpToDisplay(i + 1);
				XPathNavigator nLabl = Ddi.SelectSingleNode("ddi:labl", nVarGrp);
				if (nLabl != null)
					sHeadings[i] = nLabl.Value;
				else
					Trace.TraceWarning("varGrp[@type='section'] detected without a ddi:labl tag, resulting in an empty page heading.");
			}

			return sHeadings;
		}
		#endregion

		/// <summary>
		/// This class is plugged into the XSLT that performs the questionnaire
		/// rendering.  The purpose of this class is to provide a limited tie between
		/// code in the parent class and code in the XSLT file, to allow code reuse.
		/// </summary>
		public class StandardHtmlXsltExtensionFunctions : HtmlXsltExtensionFunctions
		{
			protected new StandardHtml Engine { get { return (StandardHtml)base.Engine; } }
			public StandardHtmlXsltExtensionFunctions(StandardHtml engine)
				: base(engine, NAssess.Core.Engines.Evaluation.Resources.ResourceManager)
			{
			}

			public bool IsVarGrpActiveBasedOnData(XPathNodeIterator xpathNodeIterator)
			{
				if (xpathNodeIterator == null) throw new ArgumentNullException("iXpath");
				// Convert to a node
				if (xpathNodeIterator.Count != 1) throw new ArgumentException(string.Format(Engines.Evaluation.Resources.ExactlyOneNodeRequired, xpathNodeIterator.Count));
				if (xpathNodeIterator.CurrentPosition < 1) xpathNodeIterator.MoveNext();
				XPathNavigator ipath = xpathNodeIterator.Current;
				XPathNodeIterator att = ipath.Select("@ID");
				att.MoveNext();
				string sID = att.Current.Value;
				XPathNavigator nVarGrp = Engine.Evaluation.Codebook.SelectSingleNode(
					"/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@ID='" + sID + "']");
				if (nVarGrp == null)
					throw new ApplicationException("varGrp could not be found in codebook.");

				// Call our parent function to do the work now that we have the node
				return Engine.IsVarGrpActiveBasedOnData(nVarGrp);
			}

			/// <summary>
			/// Used to find the length of the longest string in a nodeset
			/// </summary>
			/// <param name="nodeSet">The nodeset to find the longest string in.</param>
			/// <returns>The length of the longest string.</returns>
			public static int MaxStringLength(XPathNodeIterator nodeSet)
			{
				if (nodeSet == null) throw new ArgumentNullException("nodeSet");
				int MaxLengthSoFar = 0;

				while (nodeSet.MoveNext())
				{
					int curLength = nodeSet.Current.Value.Length;
					if (curLength > MaxLengthSoFar)
						MaxLengthSoFar = curLength;
				}
				return MaxLengthSoFar;
			}
		}
	}
}
