using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Resources;
using System.Xml;
using System.Xml.XPath;

using NerdBank.Tools;

namespace NAssess.Core.Engines
{
	/// <summary>
	/// Any class that implements this interface can render a report as Html.
	/// </summary>
	public abstract class HtmlRenderingEngine : RenderingEngine
	{
		private const string HtmlXsltExtensionsNamespace = "http://nassess.byu.edu/rendering/html";

		/// <summary>
		/// Gets the number of pages that are in the questionnaire.  
		/// This number may depend on the responses already provided, and thus
		/// may change with each submission of responses.
		/// </summary>
		public abstract int PagesCount { get; }
		public abstract void ToHtml(int page, XmlWriter writer);
		public virtual string ToHtml(int page)
		{
			using (MemoryStream ms = new MemoryStream(5 * 1024))
			{
				XmlTextWriter writer = new XmlTextWriter(ms, Encoding.Unicode);
				ToHtml(page, writer);
				writer.Flush();
				ms.Position = 0;
				StreamReader sr = new StreamReader(ms);
				return sr.ReadToEnd();
			}
		}
		public override RenderingEngine.Target RenderTarget
		{
			get
			{
				return RenderingEngine.Target.Html;
			}
		}
		protected override void InitXsltArguments()
		{
			base.InitXsltArguments();
			XsltArguments.AddExtensionObject(HtmlXsltExtensionsNamespace, new HtmlXsltExtensionFunctions(this, null));
		}
		public static string LocalizeJavaScript(string javascript, ResourceManager resources)
		{
			if (javascript == null) throw new ArgumentNullException("javascript");
			if (resources == null) throw new ArgumentNullException("resources");
			return Regex.Replace(javascript, @"\$(\w+)\$", delegate(Match m)
			{
				string s = resources.GetString(m.Groups[1].Value); // look up resource
				return s.Replace("\r\n", @"\n"); // Javascript friendly: newlines to just newline symbols
			});
		}

		public class HtmlXsltExtensionFunctions : CommonXsltExtensionsFunctions
		{
			protected readonly ResourceManager Resources;

			public HtmlXsltExtensionFunctions(RenderingEngine engine, ResourceManager resources)
				: base(engine)
			{
				Resources = resources;
			}

			public XPathNodeIterator IncludeClientJavaScript(string resourceFile, string webDirectRequest)
			{
				const string language = "JavaScript1.1";
				const string type = "text/javascript";

				if (resourceFile == null) throw new ArgumentNullException("resourceFile");
				if (webDirectRequest != null && webDirectRequest.Length == 0) webDirectRequest = null;

				XmlDocument xmlScript = new XmlDocument();
				xmlScript.AppendChild(xmlScript.CreateElement(null, "script", "http://www.w3.org/1999/xhtml"));
				if (language != null && language.Length > 0)
					xmlScript.DocumentElement.SetAttribute("language", language);
				if (type != null && type.Length > 0)
					xmlScript.DocumentElement.SetAttribute("type", type);
				if (Engine.WebDirectUrl == null || webDirectRequest == null) // If no WebDirect, embed script inline
				{
					string clientScript = EmbeddedResources.LoadFileFromAssemblyWithNamespace(resourceFile, Project.DefaultNamespace);
					if (Resources != null)
						clientScript = LocalizeJavaScript(clientScript, Resources);

					// We must prefix the script with a linefeed to prevent the first line from being
					// a compile error, due to a <!--var ... or some other collision.
					xmlScript.DocumentElement.AppendChild(xmlScript.CreateComment("\r\n" + clientScript));
				}
				else // Otherwise there IS a WebDirect Url specified
				{
					xmlScript.DocumentElement.SetAttribute("src", Engine.WebDirectUrl + webDirectRequest);
					xmlScript.DocumentElement.InnerText = " "; // prevent self-closing tag.
				}

				return xmlScript.CreateNavigator().Select("*");
			}

		}

		public class HtmlRenderingException : RenderingException
		{
			public HtmlRenderingException() { }
			public HtmlRenderingException(string msg) : base(msg) { }
		}
	}
}
