using System;
using System.Diagnostics;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;

using NAssess.Core.Engines;
using NAssess.Core.Evaluating;
using NerdBank.Tools;
using NerdBank.Tools.Exslt;

namespace NAssess.Core.Engines.Questionnaire
{
	/// <summary>
	/// Summary description for StandardFO.
	/// </summary>
	public class StandardFO : NAssess.Core.Evaluating.FORenderingEngine
	{
		public StandardFO()
		{
		}

		#region Attributes
		private NAssess.Core.Evaluating.Questionnaire questionnaire;
		public override NAssess.Core.Evaluating.Questionnaire Questionnaire
		{
			get { return questionnaire; }
			set { questionnaire = value; }
		}

		private const string XsltStartPoint = "/Engines/Questionnaire/xslt.fo/main.xslt";
		private const string XsltCacheName = "Xslt: " + XsltStartPoint;
		private XslCompiledTransform xslt;
		protected XslCompiledTransform Xslt
		{
			get
			{
				if (xslt == null)
				{
					// The XSLT that produces the XSL-FO that we output
					// First try to pull it out of the cache
					xslt = Project.GetFromCacheIfAvailable<XslCompiledTransform>(XsltCacheName);
					if (xslt == null)
					{
						xslt = new XslCompiledTransform(Project.IsXsltDebuggingEnabled);
						XsltSettings settings = new XsltSettings(true, true);
						using (Stream mainXslt = EmbeddedResources.GetFileStreamFromAssembly(XsltStartPoint, Project.DefaultNamespace))
							xslt.Load(new XmlTextReader(mainXslt), settings, CustomResolver);
						Project.SaveInCacheIfAvailable<XslCompiledTransform>(XsltCacheName, xslt);
					}
				}
				return xslt;
			}
		}
		private XmlResolver customResolver;
		protected XmlResolver CustomResolver
		{
			get
			{
				// Prepare custom InProjectResolver, so XSLT files can be retrieved directly from the assembly
				if (customResolver == null)
					customResolver = new ProjectXmlResolver(Project.DefaultNamespace);
				return customResolver;
			}
		}

		#endregion
		#region Operations
		public override void ToFO(System.IO.Stream outStream, bool monochrome, InlineFormatter inlineFOFormatter)
		{
			if (Questionnaire == null) throw new InvalidOperationException(Resources.QuestionnairePropertyRequired);
			if (outStream == null) throw new ArgumentNullException("outStream");
			if (!outStream.CanWrite) throw new ArgumentException(Resources.StreamUnwritable, "outStream");

			// Add other arguments for rendering, based on preset properties
			InitXsltParams(monochrome);

			// Inform all plug-ins that rendering is about to begin.
			OnRendering(null);

			// Transform questionnaire XML to FO
			using (MemoryStream ms = new MemoryStream(50 * 1024))
			{
				XmlTextWriter xtw = new XmlTextWriter(ms, Encoding.Unicode);
				Xslt.Transform(new XmlTextReader(new StringReader(Questionnaire.Codebook.XmlString)),
					XsltArguments, xtw);
				xtw.Flush();
				ms.Position = 0;

				// Transform XSL-FO to PDF now, if necessary.
				try
				{
					if (inlineFOFormatter != null)
						inlineFOFormatter(ms);
				}
				catch (Exception ex)
				{
					// Likely the cause of this is invalid FO being passed into the PDF renderer
					throw new FORenderingException("Error during FO formatting callback.", ex);
				}
				finally
				{
					// Inform all plug-ins that rendering has completed.
					OnRendered(null);
				}
			}
		}

		private void InitXsltParams(bool Monochrome)
		{
			// Set logo parameter, and assume it is rooted in the localized images directory
			// if the Uri is not fully qualified.
			if (LogoImageUrl != null)
				SetXsltParam("logoUrl", string.Empty, new Uri(ImagesWithTextDirectory, LogoImageUrl).AbsoluteUri);
		}
		protected override void InitXsltArguments()
		{
			base.InitXsltArguments();

			XmlDocument consent = new XmlDocument();
			consent.LoadXml(WrapHtmlFragment(Questionnaire.Consent));
			SetXsltParam("consent", string.Empty, consent.CreateNavigator());

			XmlDocument introduction = new XmlDocument();
			introduction.LoadXml(WrapHtmlFragment(Questionnaire.Introduction));
			SetXsltParam("introduction", string.Empty, introduction.CreateNavigator());

			XmlDocument closingStatement = new XmlDocument();
			closingStatement.LoadXml(WrapHtmlFragment(Questionnaire.ClosingStatement));
			SetXsltParam("closingStatement", string.Empty, closingStatement.CreateNavigator());

			// Add extension objects to provide additional functionality
			InitXsltExtensionObjects();
		}
		private static string WrapHtmlFragment(string s)
		{
			return "<div xmlns=\"http://www.w3.org/1999/xhtml\">" + s + "</div>";
		}
		private void InitXsltExtensionObjects()
		{
			SetXsltExtensionObject("http://fhss.byu.edu/exslt/strings", new ExsltStrings());
			SetXsltExtensionObject(CultureNamespace, new CultureAssist(typeof(StandardFO)));
		}

		#endregion
	}
}
