using System;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;

namespace NAssess.Core.Engines
{
	/// <summary>
	/// Summary description for RenderingEngine.
	/// </summary>
	public abstract class RenderingEngine
	{
		public const string CommonXsltExtensionsNamespace = "http://nassess.byu.edu/rendering";
		public const string CultureNamespace = "http://nassess.byu.edu/culture";
		#region RenderTarget specifiers
		public enum Target
		{
			Html,
			FO,
			Pdf
		}
		public abstract Target RenderTarget { get; }
		#endregion
		#region Events
		public event EventHandler Rendering;
		public event EventHandler Rendered;

		protected void OnRendering(EventArgs e)
		{
			if (Rendering != null)
				Rendering(this, e);
		}
		protected void OnRendered(EventArgs e)
		{
			if (Rendered != null)
				Rendered(this, e);
		}

		#endregion
		#region XSLT Arguments
		private XsltArgumentList xsltArguments;
		public XsltArgumentList XsltArguments
		{
			get
			{
				if (xsltArguments == null)
				{
					xsltArguments = new XsltArgumentList();
					InitXsltArguments();
				}
				return xsltArguments;
			}
			protected set
			{
				if (value == null)
					xsltArguments = null;
				else
					throw new ApplicationException(Resources.OnlyResetXsltArguments);
			}
		}
		protected virtual void InitXsltArguments()
		{
			XsltArguments.AddExtensionObject(CommonXsltExtensionsNamespace, new CommonXsltExtensionsFunctions(this));
			SetXsltParam("imagesURL", string.Empty, ImagesDirectory);
			SetXsltParam("imagesWithTextURL", string.Empty, ImagesWithTextDirectory);
		}
		protected void SetXsltParam(string name, string namespaceUri, object parameter)
		{
			// remove parameter first, to ensure we don't add the same
			// parameter twice and throw an exception.
			XsltArguments.RemoveParam(name, namespaceUri);
			if (parameter != null)
				XsltArguments.AddParam(name, namespaceUri, parameter);
		}
		protected void SetXsltExtensionObject(string namespaceUri, object extension)
		{
			XsltArguments.RemoveExtensionObject(namespaceUri);
			if (extension != null)
				XsltArguments.AddExtensionObject(namespaceUri, extension);
		}
		#endregion
		#region Plug-in support
		private PlugInExtensions plugins;
		protected PlugInExtensions Plugins
		{
			get
			{
				if (plugins == null) plugins = new PlugInExtensions(this);
				return plugins;
			}
		}
		public void PlugIn(IReportExtension plugin)
		{
			Plugins.Add(plugin);
		}
		#endregion
		#region External images lookup URLs
		// The following couple of properties control where the resulting report HTML points to download
		// report-specific images, both language-neutral and language-specific.
		private Uri imagesDirectory = StdImagesDirectory;
		public static readonly Uri StdImagesDirectory = new Uri("http://www.relate-institute.org/common/reports/images/");
		public Uri ImagesDirectory
		{
			get
			{
				return imagesDirectory;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("ImagesDirectory");
				if (!value.AbsoluteUri.EndsWith("/")) throw new ArgumentException(Resources.UrlRequiresTrailingSlash, "ImagesDirectory");
				imagesDirectory = value;
			}
		}

		private Uri imagesWithTextDirectory = StdImagesWithTextDirectory;
		public static readonly Uri StdImagesWithTextDirectory = new Uri(StdImagesDirectory, "en-US/");
		public Uri ImagesWithTextDirectory
		{
			get
			{
				return imagesWithTextDirectory;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("ImagesWithTextDirectory");
				if (!value.AbsoluteUri.EndsWith("/")) throw new ArgumentException(Resources.UrlRequiresTrailingSlash, "ImagesWithTextDirectory");
				imagesWithTextDirectory = value;
			}
		}
		#endregion
		public Uri WebDirectUrl;

		public class CommonXsltExtensionsFunctions
		{
			protected RenderingEngine Engine;

			public CommonXsltExtensionsFunctions(RenderingEngine engine)
			{
				Engine = engine;
			}
		}

		public class RenderingException : ApplicationException
		{
			public RenderingException() { }
			public RenderingException(string msg) : base(msg) { }
			public RenderingException(string msg, Exception innerException) : base(msg, innerException) { }
		}
	}
}
