using System;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// Summary description for BarGraph.
	/// </summary>
	public class BarGraph : GraphWindow
	{
		class Bar
		{
			public float Value;
			public float GreenCutoff = float.NaN;
			public float RedCutoff = float.NaN;

			public Bar(float value)
			{
				if (value < 0)
					throw new ArgumentOutOfRangeException("value", value, Resources.NonNegativeValueRequired);
				Value = value;
			}
			public Bar(Single value, Single greenCutoff, Single redCutoff)
				: this(value)
			{
				GreenCutoff = greenCutoff;
				RedCutoff = redCutoff;
			}
		}

		public ArrayList Bars = new ArrayList();

		public BarGraph()
		{
			for (int i = 0; i < CurveCorners.Length; i++)
				CurveCorners[i] = false;
			CurveCorners[(int)GraphCorner.TopRight] = true;
		}

		#region Coloring
		//private static Color[] _barColors = new Color[] {Color.SkyBlue, Color.MediumBlue, Color.Plum, Color.DarkViolet};
		private static Color[] barColors = new Color[] { Color.FromArgb(unchecked((int)0xFF3C4473)), Color.FromArgb(unchecked((int)0xFFA0262A)), Color.FromArgb(unchecked((int)0xFF74A2C5)), Color.FromArgb(unchecked((int)0xFFCB692E)) };
		private static Color[] barColorsBW = new Color[] { Color.Black, Color.Gray, Color.LightGray, Color.White };
		private static Color[] cutoffColors = new Color[] { Color.FromArgb(245, 162, 162), Color.FromArgb(192, 234, 192) };
		private static Color[] cutoffColorsBW = new Color[] { Color.DarkGray, Color.LightGray };
		private const int cutoffRed = 0;
		private const int cutOffGreen = 1;

		public static Color[] BarColors(bool monochromeMode)
		{
			if (monochromeMode)
				return barColorsBW;
			else
				return barColors;
		}
		public static Color[] CutoffColors(bool monochromeMode)
		{
			if (!monochromeMode)
				return cutoffColors;
			else
				return cutoffColorsBW;
		}

		public static Color BarColor(int index, int count, bool monochromeMode)
		{
			// in the case of just two bars, we want the second bar to be the color of the last bar.
			if (count == 2 && index == 1) return BarColors(monochromeMode)[3];
			return BarColors(monochromeMode)[index];
		}
		private Color BarColor(int index)
		{
			return BarColor(index, Bars.Count, MonochromeMode);
		}


		private bool gradientBars;
		public bool GradientBars
		{
			get
			{
				if (MonochromeMode) return false; // no gradient in monochrome mode 
				return gradientBars;
			}
			set
			{
				if (value && MonochromeMode)
					throw new InvalidOperationException(Resources.GradientAndMonochromeNotSupported);

				gradientBars = value;
			}
		}

		#endregion

		#region Scaling
		public enum BarGraphStyle
		{
			Proportional,
			Logarithmic
		}
		public const float StdMaxYScale = 5;
		public float MaxYScale = StdMaxYScale; // The highest point a bargraph could reach
		public const BarGraphStyle StdStyle = BarGraphStyle.Logarithmic;
		public BarGraphStyle Style = StdStyle;

		public int MaxBarWidth = 30; // this overrides EmptySpaceProportion
		public Single EmptySpaceProportion = 0.3F; // 30% of the horizontal space should be empty
		public int HashLength = 4; // the # of pixels the reference lines extend beyond the left side of the box
		private int barWidth
		{
			get
			{
				return (int)Math.Min(ContentArea.Width / Bars.Count * (1 - EmptySpaceProportion), MaxBarWidth);
			}
		}
		private int barSpacing
		{
			get
			{
				return (int)Math.Max(ContentArea.Width / (Bars.Count) * EmptySpaceProportion / 2, (ContentArea.Width - Bars.Count * MaxBarWidth) / Bars.Count / 2);
			}
		}

		public int CalcHeight(float preCalcHeight)
		{
			switch (Style)
			{
				case BarGraphStyle.Logarithmic:
					float lineInterval = (ContentArea.Height - 5) / 9;
					if (preCalcHeight > 3.0)
						return (int)((lineInterval * 3) + ((preCalcHeight - 3) * (lineInterval * 3)));
					else
						return (int)(preCalcHeight * lineInterval);
				case BarGraphStyle.Proportional:
					return (int)(preCalcHeight * (ContentArea.Height - 5) / MaxYScale);
				default:
					throw new NotSupportedException(Resources.StyleNotSupported);
			}
		}

		#endregion

		#region Rendering
		public override Bitmap Render()
		{
			int transLevel = 255;

			int iCounter;

			// Save room for the hash marks
			Rectangle rectArea = base.BoxArea;
			rectArea.Offset(HashLength, 0); rectArea.Width -= HashLength;
			base.BoxArea = rectArea;
			Bitmap bmpInput = base.Render();
			//return bmpInput;

			// Set the background as the canvas
			Graphics drawTool = Graphics.FromImage(bmpInput);
			drawTool.InterpolationMode = InterpolationMode.HighQualityBicubic;
			drawTool.CompositingMode = CompositingMode.SourceCopy;
			drawTool.CompositingQuality = CompositingQuality.HighQuality;
			drawTool.SmoothingMode = SmoothingMode.AntiAlias;

			drawCutOffs(drawTool);

			drawReferenceLines(drawTool);  // Draw the measurement reference lines

			// Draw the bars at the given measurements
			iCounter = 0;
			int barLeft = ContentArea.Left + barSpacing;
			foreach (Bar Bar in Bars)
			{
				if (float.IsNaN(Bar.Value)) // if no data
					drawNoData(barLeft, BarColor(iCounter), transLevel, Color.Black, drawTool);
				else // if there IS data for a bar
					drawBar(barLeft, CalcHeight(Bar.Value), BarColor(iCounter), transLevel, Color.Black, drawTool);
				barLeft += barWidth + barSpacing * 2;
				iCounter += 1;
			}

			drawTool.Dispose();

			return bmpInput;
		}
		private void drawReferenceLines(Graphics drawTool)
		{
			// Draw the measurement reference lines
			for (int y = 1; y <= MaxYScale; y++)
				drawTool.DrawLine(Pens.Gray, BoxArea.Left - HashLength, ContentArea.Bottom - CalcHeight(y), BoxArea.Right, ContentArea.Bottom - CalcHeight(y));
		}
		private void drawCutOffs(Graphics drawTool)
		{
			// Verify that either all or none of the green/red cutoffs are provided
			bool bGreenNotProvided = float.IsNaN(((Bar)Bars[0]).GreenCutoff);
			bool bRedNotProvided = float.IsNaN(((Bar)Bars[0]).RedCutoff);
			foreach (Bar b in Bars)
			{
				if (float.IsNaN(b.GreenCutoff) != bGreenNotProvided)
					throw new ArgumentException(Resources.BalancedCutoffsRequired);
				if (float.IsNaN(b.RedCutoff) != bRedNotProvided)
					throw new ArgumentException(Resources.BalancedCutoffsRequired);
			}

			Brush redBrush = new SolidBrush(CutoffColors(MonochromeMode)[cutoffRed]);
			Brush greenBrush = new SolidBrush(CutoffColors(MonochromeMode)[cutOffGreen]);

			// Draw the GREEN section of the background
			if (!float.IsNaN(((Bar)Bars[0]).GreenCutoff))
			{
				drawTool.FillRectangle(greenBrush, Rectangle.FromLTRB(BoxArea.Left + FeatherWidth - 1, ContentArea.Top + (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopLeft) - 1, ContentArea.Left + 1, ContentArea.Bottom - CalcHeight(((Bar)Bars[0]).GreenCutoff)));
				if (base.CurveCorners[(int)GraphCorner.TopLeft]) drawTool.FillPie(greenBrush, BoxArea.Left + FeatherWidth - 1, BoxArea.Top + FeatherWidth - 1, CurveRadius, CurveRadius + 2, 180, 90);
				int tempLeft = ContentArea.Left;
				for (int i = 0; i < Bars.Count; i++)
				{
					drawTool.FillRectangle(greenBrush, tempLeft, ContentArea.Top - 1, barWidth + barSpacing * 2 + 1, ContentArea.Height + 1 - CalcHeight(((Bar)Bars[i]).GreenCutoff));
					tempLeft += barWidth + barSpacing * 2;
				}
				drawTool.FillRectangle(greenBrush, Rectangle.FromLTRB(BoxArea.Right - CurveRadius / 2 - FeatherWidth - 1, ContentArea.Top + (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopRight) - 1, BoxArea.Right - FeatherWidth + 1, ContentArea.Bottom - CalcHeight(((Bar)Bars[Bars.Count - 1]).GreenCutoff)));
				if (base.CurveCorners[(int)GraphCorner.TopRight]) drawTool.FillPie(greenBrush, BoxArea.Right - FeatherWidth - CurveRadius - 1, BoxArea.Top + FeatherWidth - 1, CurveRadius + 2, CurveRadius + 2, 270, 90);
			}

			// Draw the RED section of the background
			if (!float.IsNaN(((Bar)Bars[0]).RedCutoff))
			{
				drawTool.FillRectangle(redBrush, Rectangle.FromLTRB(BoxArea.Left + FeatherWidth - 1, ContentArea.Bottom - CalcHeight(((Bar)Bars[0]).RedCutoff), ContentArea.Left + 1, ContentArea.Bottom - (CurveRadius / 2) * CurveCornersBit(GraphCorner.InsideLeft)));
				if (base.CurveCorners[(int)GraphCorner.InsideLeft]) drawTool.FillPie(redBrush, BoxArea.Left + FeatherWidth - 1, ContentArea.Bottom - CurveRadius - 1, CurveRadius, CurveRadius, 90, 90);
				int tempLeft = ContentArea.Left;
				for (int i = 0; i < Bars.Count; i++)
				{
					drawTool.FillRectangle(redBrush, Rectangle.FromLTRB(tempLeft, ContentArea.Bottom - CalcHeight(((Bar)Bars[i]).RedCutoff), tempLeft + barWidth + barSpacing * 2 + 1, ContentArea.Bottom));
					tempLeft += barWidth + barSpacing * 2;
				}
				drawTool.FillRectangle(redBrush, Rectangle.FromLTRB(BoxArea.Right - CurveRadius / 2 - FeatherWidth - 1, ContentArea.Bottom - CalcHeight(((Bar)Bars[Bars.Count - 1]).RedCutoff), BoxArea.Right - FeatherWidth + 1, ContentArea.Bottom - FeatherWidth - (CurveRadius / 2) * CurveCornersBit(GraphCorner.InsideRight) + 1));
				if (base.CurveCorners[(int)GraphCorner.InsideRight]) drawTool.FillPie(redBrush, BoxArea.Right - FeatherWidth - CurveRadius - 1, ContentArea.Bottom - FeatherWidth - CurveRadius, CurveRadius + 2, CurveRadius + 2, 0, 90);
			}

			redBrush.Dispose();
			greenBrush.Dispose();
		}
		private void drawBar(int leftSide, int barHeight, Color barColor, int transNumber, Color borderColor, Graphics drawTool)
		{
			Brush brushInner;
			if (GradientBars)
			{
				const float ratio = 0.5f;
				Color gradColor = Color.FromArgb(barColor.A, (int)(barColor.R * ratio), (int)(barColor.G * ratio), (int)(barColor.B * ratio));
				brushInner = new LinearGradientBrush(new PointF(leftSide, ContentArea.Bottom), new PointF(leftSide + barWidth, ContentArea.Bottom - barHeight), gradColor, Color.FromArgb(transNumber, barColor));
			}
			else
				brushInner = new SolidBrush(barColor);

			Pen penInner = new Pen(brushInner, barWidth);
			Pen penBorder = new Pen(Color.FromArgb(transNumber, borderColor), 1);

			drawTool.DrawLine(penInner, new PointF(leftSide + barWidth / 2, ContentArea.Bottom), new PointF(leftSide + barWidth / 2, ContentArea.Bottom - barHeight));

			// Draw the border of the bar (three lines that bound the rectangle on the left, right, and top)
			drawTool.DrawLine(penBorder, leftSide, ContentArea.Bottom - barHeight, leftSide, ContentArea.Bottom);
			drawTool.DrawLine(penBorder, leftSide, ContentArea.Bottom - barHeight, leftSide + barWidth, ContentArea.Bottom - barHeight);
			drawTool.DrawLine(penBorder, leftSide + barWidth, ContentArea.Bottom - barHeight, leftSide + barWidth, ContentArea.Bottom);

			brushInner.Dispose();
			penInner.Dispose();
			penBorder.Dispose();
		}

		private void drawNoData(int leftSide, Color barColor, int transNumber, Color borderColor, Graphics drawTool)
		{
			// Draw "NO DATA" vertically in the position the bar would be in

			// TODO: code here
		}
		/// <summary>
		/// Offers web browsers direct access to rendered bar graphs.
		/// </summary>
		/// <param name="contentType">Set to "image/jpeg" for every call</param>
		/// <param name="title">The title of the graph</param>
		/// <param name="style">Indicates whether a proportional or logarithmic scale is desired for the graph</param>
		/// <param name="gradient">True if a gradiant on each bar is desired.  Not available in monochrome mode.</param>
		/// <param name="monochrome">True if greyscale rendering is desired.</param>
		/// <param name="scales">An array of up to four values to use as data for the bars.</param>
		/// <param name="height">The vertical dimension of the physical size of the rendered graph in pixels.</param>
		/// <param name="MaxY">The maximum scale value.  For example, most scales go from 1-5, so 5 would go here.</param>
		/// <param name="width">The horizontal dimension of the physical size of the rendered graph in pixels.</param>
		/// <returns>A stream to the rendered Jpeg.</returns>		
		/// <example>
		/// URL parameters
		/// <list type="bullet">
		///		<item>?title=Relationship+Satisfaction&amp;style=logarithmic&amp;maxY=5&amp;scales=4.333333g4r2&amp;scales=5g4r2</item>
		///		<description>Logarithmic graph with two scales and cut off values.</description>
		///		<item>?title=Conflict+Resolution&amp;style=logarithmic&amp;maxY=5&amp;scales=4.285714g4.25r3.00&amp;scales=4.833333g4.00r3.00&amp;scales=4.428571g4.25r3.25&amp;scales=4.714286g4.00r3.00</item>
		///		<description>Logarithmic graph with four scales and cut off values.</description>
		///		<item>title=Approval&amp;style=proportional&amp;maxY=4&amp;width=145&amp;height=140&amp;scales=4gr&amp;scales=4gr</item>
		///		<description>Proportional graph with two scales and no cutoffs</description>
		/// </list>
		/// </example>
		/// <remarks>
		/// The scales string array must be either all scale values or all scale values with cutoff values specified.
		/// The format is thus:
		/// [ScaleValue]g[GreenCutoffValue]r[RedCutoffValue]
		/// or simply 
		/// [ScaleValue]
		/// for each array element.  
		/// </remarks>
		[WebDirect]
		public static Stream Render(out string contentType, string title, BarGraphStyle style, bool gradient, bool monochrome,
			float maxY, int width, int height, params string[] scales)
		{
			if (scales == null) throw new ArgumentNullException("scales", Resources.AtLeastOneScaleRequired);
			if (title == null) throw new ArgumentNullException("title");
			if (maxY < 0) throw new ArgumentOutOfRangeException("MaxY", maxY, Resources.PositiveNumberRequired);
			if (width < 0) throw new ArgumentOutOfRangeException("width", width, Resources.PositiveNumberRequired);
			if (height < 0) throw new ArgumentOutOfRangeException("height", height, Resources.PositiveNumberRequired);
			//if( style == null ) style = StyleEnum.Logarithmic;

			MemoryStream ms = new MemoryStream(5 * 1024);
			BarGraph bg = new BarGraph();

			bg.Title = title;
			bg.Style = style;
			bg.GradientBars = gradient;
			bg.MonochromeMode = monochrome;

			// Note that the following few settings can be optional parameters, since this method is called
			// via Reflection.  So only set each one if it was obviously set.
			if (width > 0) bg.Size.Width = width;
			if (height > 0) bg.Size.Height = height;
			if (maxY > 0) bg.MaxYScale = maxY;

			foreach (string sScale in scales)
			{
				Group mScale = Regex.Match(sScale, @"\A(?<scale>[\.\d]+)").Groups["scale"];
				Group mRed = Regex.Match(sScale, @"r(?<red>[\.\d]+)").Groups["red"];
				Group mGreen = Regex.Match(sScale, @"g(?<green>[\.\d]+)").Groups["green"];

				float scale = mScale.Success ? Convert.ToSingle(mScale.Value) : float.NaN;
				// There may be green and red ranges encoded in this scale
				if (scale > bg.MaxYScale)
					throw new ArgumentOutOfRangeException("scales", scale,
						Resources.ScaleAboveMaximum);
				Bar newBar = new Bar(scale);
				if (mGreen.Success)
					newBar.GreenCutoff = Convert.ToSingle(mGreen.Value);
				if (mRed.Success)
					newBar.RedCutoff = Convert.ToSingle(mRed.Value);
				bg.Bars.Add(newBar);
			}

			bg.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}
		#endregion

		#region Test routine
		[WebDirect]
		public new static Stream Test(out string contentType)
		{
			MemoryStream ms = new MemoryStream(5 * 1024);
			BarGraph bg = new BarGraph();

			bg.Bars.Add(new Bar(1, 4, 1));
			bg.Bars.Add(new Bar(2.25F, 3, 2));
			bg.Bars.Add(new Bar(3.5F, 4, 1));
			bg.Bars.Add(new Bar(5, 4.5F, 2));

			bg.Title = "Test scale";
			bg.Style = BarGraphStyle.Logarithmic;

			bg.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}
		#endregion
	}
}
