using System;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	public enum PointStyle
	{
		LightSquare,
		LightCircle,
		DarkSquare,
		DarkCircle,
	}
	/// <summary>
	/// Summary description for BullsEyeGraph.
	/// </summary>
	public class BullsEyeGraph : GraphWindow
	{
		public const int ScaleMax = 5;
		public const int ScaleMin = 1;

		private static Color[] colors = new Color[] {
										 Color.FromArgb(144,238,144), Color.FromArgb(135,206,250), 
										 Color.FromArgb(89,155,89), Color.FromArgb(173,216,230),
										 Color.FromArgb(232,208,252), Color.FromArgb(237,219,95),
										 Color.FromArgb(112,111,181), Color.FromArgb(255,166,85),
										 Color.FromArgb(201,152,93), Color.FromArgb(255,182,193),
										 Color.FromArgb(119,97,64), Color.FromArgb(240,128,128),
										 Color.FromArgb(119,130,132)
									 };
		protected static IList<Color> Colors = new ReadOnlyCollection<Color>(colors);

		public static Color DivisionColor(int index)
		{
			return Colors[index];
		}

		class Division
		{
			public Division() { }
			private readonly ArrayList chains = new ArrayList(2);
			public ArrayList Chains { get { return chains; } }
		}

		class Chain
		{
			public Chain(params float[] points)
			{
				foreach (float point in points)
					AddPoint(point);
			}
			private readonly ArrayList points = new ArrayList(2);
			public ArrayList Points { get { return points; } }
			private void AddPoint(float value)
			{
				if (value < ScaleMin || value > ScaleMax) throw new ArgumentOutOfRangeException("value", value.ToString());
				Points.Add(value);
			}
		}


		public BullsEyeGraph()
		{
			TargetArea = ContentArea;
			TargetArea.Height = TargetArea.Width = Math.Min(TargetArea.Width, TargetArea.Height);
			TargetArea.Inflate(-10, -10);
			TargetArea.Offset(0, 10);
		}

		private readonly ArrayList divisions = new ArrayList();
		public ArrayList Divisions { get { return divisions; } }
		protected Rectangle TargetArea;

		#region Rendering
		const float startingPoint = 0; // in radians
		protected float DivisionSize { get { return (float)(2 * Math.PI / Divisions.Count); } } // in radians
		protected PointF CenterPoint { get { return new Point((TargetArea.Right + TargetArea.Left) / 2, (TargetArea.Bottom + TargetArea.Top) / 2); } }
		float maxDistanceFromCenter { get { return TargetArea.Width / 2 - 5; } }
		public override Bitmap Render()
		{
			Bitmap bmp = base.Render();
			Graphics drawTool = Graphics.FromImage(bmp);

			drawTool.SmoothingMode = SmoothingMode.AntiAlias;
			drawTool.DrawEllipse(Pens.Black, TargetArea);
			HighlightDivisions(drawTool);
			DrawConcentricCircles(drawTool);
			PlotPoints(drawTool);

			return bmp;
		}
		private void HighlightDivisions(Graphics drawTool)
		{
			int iColor = 0;
			for (float offset = startingPoint; offset < 2 * Math.PI; offset += DivisionSize)
			{
				using (Brush brush = new SolidBrush(Colors[iColor++]))
					drawTool.FillPie(brush, TargetArea, radiansToDegrees(offset), radiansToDegrees(DivisionSize));
				iColor %= Colors.Count; // out-of-bounds protection
			}
		}
		private void DrawConcentricCircles(Graphics drawTool)
		{
			for (int i = ScaleMin; i < ScaleMax; i++)
				DrawCircle(drawTool, i);
		}
		private void DrawCircle(Graphics drawTool, float scaleValue)
		{
			RectangleF rect = new RectangleF(CenterPoint, new SizeF(0, 0));
			float inflateFactor = ScaleToDistanceFromCenter(scaleValue);
			rect.Inflate(inflateFactor, inflateFactor);
			drawTool.DrawEllipse(Pens.Brown, rect);
		}
		private float ScaleToDistanceFromCenter(float scale)
		{
			return maxDistanceFromCenter - ((scale - ScaleMin) / (ScaleMax - ScaleMin) * maxDistanceFromCenter);
		}
		private void PlotPoints(Graphics drawTool)
		{
			float offset = startingPoint;
			foreach (Division d in Divisions)
			{
				float chainOffset = offset + DivisionSize / (d.Chains.Count + 1);
				float chainAreaSize = DivisionSize / (d.Chains.Count + 1);

				int iChain = 0;
				foreach (Chain c in d.Chains)
				{
					float direction = chainOffset;
					float lastPt = float.NaN;
					foreach (float pt in c.Points)
					{
						if (!float.IsNaN(lastPt))
							DrawLine(drawTool, lastPt, pt, direction, iChain == 0 ? Pens.Black : Pens.White);
						PointStyle pointStyle;
						if (iChain % 2 != 0)
							pointStyle = float.IsNaN(lastPt) ? PointStyle.LightSquare : PointStyle.LightCircle;
						else
							pointStyle = float.IsNaN(lastPt) ? PointStyle.DarkSquare : PointStyle.DarkCircle;
						PlotPoint(drawTool, pointStyle, pt, direction, iChain == 0 ? Brushes.Black : Brushes.White);
						lastPt = pt;
					}
					chainOffset += chainAreaSize;
					iChain++;
				}

				offset += DivisionSize;
			}
		}
		private PointF ScaleToXY(float scaleValue, float direction)
		{
			return addPoints(CenterPoint, polarToRect(direction, ScaleToDistanceFromCenter(scaleValue)));
		}
		private const string pointImage1 = "/Engines/Reporting/graphs/Light-Blue.png";
		private const string pointImage2 = "/Engines/Reporting/graphs/Light-Orange.png";
		private const string pointImage3 = "/Engines/Reporting/graphs/Dark-Blue.png";
		private const string pointImage4 = "/Engines/Reporting/graphs/Dark-Orange.png";
		static internal Stream getStreamForPoint(PointStyle pointStyle)
		{
			string fileNameOfImage = null;
			switch (pointStyle)
			{
				case PointStyle.LightSquare:
					fileNameOfImage = pointImage1;
					break;
				case PointStyle.LightCircle:
					fileNameOfImage = pointImage2;
					break;
				case PointStyle.DarkSquare:
					fileNameOfImage = pointImage3;
					break;
				case PointStyle.DarkCircle:
					fileNameOfImage = pointImage4;
					break;
				default:
					throw new NotSupportedException(string.Format(Resources.PointStyleNotSupported, pointStyle));
			}
			return EmbeddedResources.GetFileStreamFromAssembly(fileNameOfImage, Project.DefaultNamespace);
		}
		private void PlotPoint(Graphics drawTool, PointStyle pointStyle, float scaleValue, float direction, Brush brush)
		{
			if (float.IsNaN(scaleValue)) return; // no data to plot a point for
			PointF pt = ScaleToXY(scaleValue, direction);

			using (Stream stream = getStreamForPoint(pointStyle))
			{
				using (Image img = Image.FromStream(stream))
				{
					pt.X -= img.Width / 2; pt.Y -= img.Height / 2;
					drawTool.DrawImage(img, pt);
				}
			}
		}

		private void DrawLine(Graphics drawTool, float from, float to, float direction, Pen pen)
		{
			if (float.IsNaN(from) || float.IsNaN(to)) return; // must have two points to draw a line
			PointF ptFrom = ScaleToXY(from, direction);
			PointF ptTo = ScaleToXY(to, direction);
			drawTool.DrawLine(pen, ptFrom, ptTo);
		}
		private const char PointInChainIterationSeparator = '-';
		private const char ChainInDivisionSeparator = 'c';
		/// <summary>
		/// Offers web browsers direct access to rendered bar graphs.
		/// </summary>
		/// <param name="contentType">Set to "image/jpeg" for every call</param>
		/// <param name="title">The title of the graph</param>
		/// <param name="monochrome">True if greyscale rendering is desired.</param>
		/// <param name="scales">An array of up to four values to use as data for the bars.</param>
		/// <param name="height">The vertical dimension of the physical size of the rendered graph in pixels.</param>
		/// <param name="width">The horizontal dimension of the physical size of the rendered graph in pixels.</param>
		/// <returns>A stream to the rendered Jpeg.</returns>		
		/// <example>
		/// URL parameters
		/// <list type="bullet">
		///		<item>?title=Relationship+Satisfaction&amp;style=logarithmic&amp;maxY=5&amp;scales=4.333333g4r2&amp;scales=5g4r2</item>
		///		<description>Logarithmic graph with two scales and cut off values.</description>
		///		<item>?title=Conflict+Resolution&amp;style=logarithmic&amp;maxY=5&amp;scales=4.285714g4.25r3.00&amp;scales=4.833333g4.00r3.00&amp;scales=4.428571g4.25r3.25&amp;scales=4.714286g4.00r3.00</item>
		///		<description>Logarithmic graph with four scales and cut off values.</description>
		///		<item>title=Approval&amp;style=proportional&amp;maxY=4&amp;width=145&amp;height=140&amp;scales=4gr&amp;scales=4gr</item>
		///		<description>Proportional graph with two scales and no cutoffs</description>
		/// </list>
		/// </example>
		/// <remarks>
		/// The scales string array must be either all scale values or all scale values with cutoff values specified.
		/// The format is thus:
		/// [ScaleValue]g[GreenCutoffValue]r[RedCutoffValue]
		/// or simply 
		/// [ScaleValue]
		/// for each array element.  
		/// </remarks>
		[WebDirect]
		public static Stream Render(out string contentType, string title, bool monochrome,
			int width, int height, params string[] scales)
		{
			if (scales == null) throw new ArgumentNullException("scales", Resources.AtLeastOneScaleRequired);
			if (title == null) throw new ArgumentNullException("title");
			if (width < 0) throw new ArgumentOutOfRangeException("width", width, Resources.NonNegativeValueRequired);
			if (height < 0) throw new ArgumentOutOfRangeException("height", height, Resources.NonNegativeValueRequired);
			//if( style == null ) style = StyleEnum.Logarithmic;

			MemoryStream ms = new MemoryStream(5 * 1024);
			BullsEyeGraph bg = new BullsEyeGraph();

			bg.Title = title;
			bg.MonochromeMode = monochrome;

			// Note that the following few settings can be optional parameters, since this method is called
			// via Reflection.  So only set each one if it was obviously set.
			if (width > 0) bg.Size.Width = width;
			if (height > 0) bg.Size.Height = height;

			foreach (string division in scales)
			{
				Division d;
				bg.Divisions.Add(d = new Division());
				foreach (string chain in division.Split(ChainInDivisionSeparator))
				{
					string[] sPoints = chain.Split(PointInChainIterationSeparator);
					float[] fPoints = new float[sPoints.Length];
					for (int i = 0; i < fPoints.Length; i++)
						fPoints[i] = Convert.ToSingle(sPoints[i]);
					d.Chains.Add(new Chain(fPoints));
				}
			}

			bg.Render(ms, ImageFormat.Png);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/png";
			return ms;
		}
		#endregion

		#region Math operations
		private static PointF polarToRect(float direction, float distance)
		{
			return new PointF((float)Math.Cos(direction) * distance, (float)Math.Sin(direction) * distance);
		}
		private static float radiansToDegrees(float radians)
		{
			return (float)(radians * 360 / (2 * Math.PI));
		}
		private static PointF addPoints(PointF pt1, PointF pt2)
		{
			return new PointF(pt1.X + pt2.X, pt1.Y + pt2.Y);
		}
		#endregion

	}
}
