using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// Summary description for Legend.
	/// </summary>
	public class BullsEyeLegend : GraphWindow
	{
		public ArrayList Labels = new ArrayList();
		public ArrayList PointLabels = new ArrayList();
		public static readonly Size StandardSize = new Size(180, 60);

		public BullsEyeLegend()
		{
			base.TitleSpace = 1.2F;
			base.BorderWidth = 1;
			base.TitleSize = 14;
			base.Size = StandardSize;
		}

		#region Rendering
		const int boxSize = 10;
		int ySpace;
		public override Bitmap Render()
		{
			if (Labels.Count == 0) throw new InvalidOperationException(Resources.LabelsRequiredFirst);

			// Reserve space for boxes
			Size.Height += (Labels.Count + PointLabels.Count) * (boxSize + 2);
			Bitmap bmpCanvas = base.Render();

			Graphics drawTool = Graphics.FromImage(bmpCanvas);

			// Generate locations for each label		
			ySpace = ContentArea.Height / (Labels.Count + 4); // vertical space available
			// Start at the top
			int ySpot = ContentArea.Top;
			int iCounter = 0;
			foreach (string Label in Labels)
			{
				// Draw box
				SolidBrush brush = new SolidBrush(BullsEyeGraph.DivisionColor(iCounter));
				Rectangle rectBox = new Rectangle(ContentArea.Left, ySpot + ySpace / 2 - boxSize / 2, boxSize, boxSize);
				drawTool.FillRectangle(brush, rectBox);
				// Draw label
				RectangleF rectLabel = RectangleF.FromLTRB(ContentArea.Left + boxSize + 5, ySpot, ContentArea.Right, ySpot + ySpace);
				drawLabel(drawTool, rectLabel, Label);

				brush.Dispose();
				ySpot += ySpace;
				iCounter += 1;
			}
			// draw a line
			drawTool.DrawLine(Pens.Black, ContentArea.Left, ySpot, ContentArea.Right, ySpot);
			ySpot += 2;
			DrawPointLabels(drawTool, ySpot);

			drawTool.Dispose();

			return bmpCanvas;
		}

		private void DrawPointLabels(Graphics drawTool, int yPos)
		{
			// Draw the categories (sample points)
			int idx = 0;
			foreach (string pointLabel in PointLabels)
			{
				PointStyle ps = PointStyle.DarkSquare;
				switch (idx++ % 4)
				{
					case 0:
						ps = PointStyle.DarkSquare;
						break;
					case 1:
						ps = PointStyle.DarkCircle;
						break;
					case 2:
						ps = PointStyle.LightSquare;
						break;
					case 3:
						ps = PointStyle.LightCircle;
						break;
				}
				DrawPoint(drawTool, yPos, ps, pointLabel);
				yPos += ySpace;
			}
		}
		private void DrawPoint(Graphics drawTool, int yPos, PointStyle pointStyle, string label)
		{
			using (Stream stream = BullsEyeGraph.getStreamForPoint(pointStyle))
			{
				using (Image img = Image.FromStream(stream))
				{
					Point pt = new Point(ContentArea.Left, yPos);
					drawTool.DrawImage(img, pt);
				}
			}
			// Draw label
			RectangleF rectLabel = RectangleF.FromLTRB(ContentArea.Left + boxSize + 5, yPos, ContentArea.Right, yPos + ySpace);
			drawLabel(drawTool, rectLabel, label);
		}

		private static void drawLabel(Graphics drawTool, RectangleF rectArea, string Label)
		{
			// Prepare font.  We size the font to max out at 10px, but it can get smaller if needed
			Font labelFont = new Font("SansSerif", Math.Min(10, rectArea.Height - 2), FontStyle.Regular, GraphicsUnit.Pixel);
			// We want to be very careful about drawing the title to make sure that it will fit.
			// Call Graphics.MeasureString to discover where this would go and shrink the font until it fits.
			while (drawTool.MeasureString(Label, labelFont).Width > rectArea.Width && labelFont.Size > 1)
			{
				labelFont.Dispose();
				labelFont = new Font(labelFont.FontFamily, labelFont.Size - 1, labelFont.Style, GraphicsUnit.Pixel);
			}

			// Prepare to format the string to center horizontally
			StringFormat strFormat = new StringFormat();
			strFormat.Alignment = StringAlignment.Near;
			strFormat.LineAlignment = StringAlignment.Center;

			drawTool.DrawString(Label, labelFont, Brushes.Black, rectArea, strFormat);
		}

		/// <summary>
		/// Generates a Legend box for browser viewing.
		/// </summary>
		/// <param name="contentType">
		/// The ContentType to pass through to the browser that is downloading the image.
		/// This is set by the type of image generated (jpeg/png/gif).
		/// </param>
		/// <param name="title">
		/// The cultured title to put on the Legend box.  This is typically "Graph Legend".
		/// </param>
		/// <param name="labels">
		/// An ordered array of elements to list in the legend.
		/// Typical contents: (example)
		///		Andrew describes Cheryl
		///		Cheryl describes Andrew
		///		Andrew describes Andrew
		///		Cheryl describes Cheryl
		/// </param>
		/// <param name="pointLabels">
		/// An ordered array of category names for each point style.  There are four of these.
		/// </param>
		/// <returns>
		/// A stream with the graphic to stream to the browser.
		/// </returns>
		/// <example>
		/// URL parameters: ?title=Graph+Legend&amp;labels=Andrew+describes+Andrew&amp;labels=Andrew+describes+Cheryl&amp;labels=-&amp;labels=Cheryl+describes+Andrew&amp;labels=Cheryl+describes+Cheryl
		/// </example>
		[WebDirect]
		public static Stream Render(out string contentType, string title, string[] labels, string[] pointLabels)
		{
			if (labels == null) throw new ArgumentNullException("labels");
			if (pointLabels == null) throw new ArgumentNullException("pointLabels");
			BullsEyeLegend l = new BullsEyeLegend();

			l.Title = title;

			foreach (string label in labels)
				l.Labels.Add(label);
			foreach (string pointLabel in pointLabels)
				l.PointLabels.Add(pointLabel);

			System.IO.MemoryStream ms = new System.IO.MemoryStream(5 * 1024);
			l.Render(ms, ImageFormat.Png);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/png";
			return ms;
		}
		#endregion

		#region Test routine
		[WebDirect]
		public new static System.IO.Stream Test(out string contentType)
		{
			Legend l = new Legend();
			l.Title = "Graph Legend";
			l.Size = Legend.DoubleSize;
			l.Labels.Add("Jake");
			l.Labels.Add("Rhonda");
			l.Labels.Add("Rebecca");
			l.Labels.Add("Sarah");
			l.Labels.Add("Andrew");

			System.IO.MemoryStream ms = new System.IO.MemoryStream(5 * 1024);
			l.Render(ms, ImageFormat.Png);
			ms.Position = 0;
			contentType = "image/png";
			return ms;
		}
		#endregion
	}
}
