using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using ZedGraph;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// A gradient version of BarGraph.
	/// Note: this graph generates the labels on the left unless
	/// the property DontDrawLabel is true. The space for the label
	/// is not retained in the size if it is not drawn. When it is 
	/// drawn, the ratio can be found by the property, LabelFraction.
	/// </summary>
	[Serializable]
	public class GradientGraph : GraphPane
	{

		#region attributes and properties
		// most of the color properties use the variable Monochrome to tell which
		// version of colors to use.  This eliminates the need of the programmer keeping track.


		private const float labelFraction = 1.1F / 3F;
		// gets the fraction of the graph reserved for a label when drawn.
		public static float LabelFraction
		{
			get
			{
				return labelFraction;
			}
		}

		private const int smallestFontSize = 10;
		private int overallFontSize = 13;
		// font size variables to set the right size for fonts
		public int OverallFontSize
		{
			get
			{
				return overallFontSize;
			}
			set
			{

				if (value > smallestFontSize)
					overallFontSize = value;
				else // cap it so it doesn't get unreadable
					overallFontSize = smallestFontSize;
			}
		}

		private bool isTwo;
		// if it is a two bar graph or not
		public bool IsTwo
		{
			get
			{
				return isTwo;
			}
			set
			{
				isTwo = value;
			}
		}

		private bool dontDrawLabel = true;
		// if the label is to be drawn
		public bool DontDrawLabel
		{
			get
			{
				return dontDrawLabel;
			}
			set
			{
				dontDrawLabel = value;
			}
		}

		private int max = 5;
		// the maximum score value
		public int Max
		{
			get
			{
				return max;
			}
			set
			{
				max = value;
				grids = new float[max];
				for (int i = 0; i < max; i++)
				{
					grids[i] = calcHeight(i + 1, max);
				}
			}
		}


		private bool monochrome;
		// true if the scale is monochrome
		public bool Monochrome
		{
			get
			{
				return monochrome;
			}
			set
			{
				monochrome = value;
			}
		}

		// the background coloring
		private static readonly Color[] backgroundColors = new Color[] { Color.Firebrick, Color.White, Color.White, Color.SeaGreen };
		private static readonly float[] backgroundPos = new float[] { 0, calcHeight(3, 5) / (float)5, calcHeight(4, 5) / (float)5, 1 };
		private static readonly Fill background = new Fill(backgroundColors, backgroundPos, -90F);
		private static readonly Color[] backgroundColorsBW = new Color[] { Color.Black, Color.LightGray, Color.White, Color.White };
		private static readonly Fill backgroundBW = new Fill(backgroundColorsBW, backgroundPos, -90F);

		// gets the array containing the 4 background colors
		public IList<Color> BackgroundColors
		{
			get
			{
				if (Monochrome)
					return new ReadOnlyCollection<Color>(backgroundColorsBW);
				else
					return new ReadOnlyCollection<Color>(backgroundColors);
			}
		}

		// gets the positions of the background colors in fractions [0,1]
		public static IList<float> BackgroundPos
		{
			get
			{
				return new ReadOnlyCollection<float>(backgroundPos);
			}
		}

		// gets the Fill of the background
		public Fill Background
		{
			get
			{
				if (Monochrome)
					return backgroundBW;
				else
					return background;
			}
		}

		// the array of the lines for 1,2,3,4,... scores
		private float[] grids;

		// shift the cut offs to make it more aesthetically pleasing
		private const float shift = .043F;


		private string[] strengthLabels = null;
		// get or set the labels
		public string[] CuttoffLabels
		{
			get
			{
				return strengthLabels;
			}
			set
			{
				if (value == null)
					throw new ArgumentNullException("value");
				if (value.Length < Max)
					throw new ArgumentOutOfRangeException("You did not specify enough cutoffs.");
				strengthLabels = new string[Max];
				for (int i = 0; i < strengthLabels.Length; i++)
				{
					strengthLabels[i] = value[i];
				}
			}
		}


		// the Font for the labels
		private FontSpec labelFontSpec = null;
		private FontSpec LabelFontSpec
		{

			get
			{
				if (labelFontSpec == null)
				{
					labelFontSpec = new FontSpec("Comic Sans MS", OverallFontSize, Color.Empty, true, false, false);
					labelFontSpec.Border.IsVisible = false;
				}
				return labelFontSpec;
			}
		}

		// Colors to match with preexisting Relate graphs
		private readonly static Color participantColor = Color.FromArgb(unchecked((int)0xFF3C4473));
		private readonly static Color participantPartnerColor = Color.FromArgb(unchecked((int)0xFFA0262A));
		private readonly static Color participant2PartnerColor = Color.FromArgb(unchecked((int)0xFF74A2C5));
		private readonly static Color participant2Color = Color.FromArgb(unchecked((int)0xFFCB692E));
		private readonly static Color participantColorBW = Color.Black;
		private readonly static Color participantPartnerColorBW = Color.Gray;
		private readonly static Color participant2PartnerColorBW = Color.LightGray;
		private readonly static Color participant2ColorBW = Color.White;
		private const float gradRatio = 0.5f;

		public Color ParticipantColor
		{
			get
			{
				if (Monochrome)
					return participantColorBW;
				else
					return participantColor;
			}
		}
		public Color ParticipantPartnerColor
		{
			get
			{
				if (Monochrome)
					return participantPartnerColorBW;
				else
					return participantPartnerColor;
			}
		}
		public Color Participant2PartnerColor
		{
			get
			{
				if (Monochrome)
					return participant2PartnerColorBW;
				else
					return participant2PartnerColor;
			}
		}
		public Color Participant2Color
		{
			get
			{
				if (Monochrome)
					return participant2ColorBW;
				else
					return participant2Color;
			}
		}

		public new string Title
		{
			get
			{
				return title;
			}
			set // the default title generation was not satisfactory so
			// we create one with a text element.
			{
				title = value;
				// Turn off the Default Title
				IsShowTitle = false;

				// Create the title
				TextItem TitleText = new TextItem(title, .50F, .005F,
					CoordType.PaneFraction, AlignH.Center, AlignV.Top);
				TitleText.FontSpec.Size = overallFontSize + 1;
				TitleText.FontSpec.IsBold = true;
				TitleText.FontSpec.Border.IsVisible = false;
				GraphItemList.Add(TitleText);
				MarginTop = 20;
			}
		}

		#endregion

		#region Constructors

		GradientGraph()
		{
			PaneRect = new RectangleF(0F, 0F, 341F, 300F);
			IsFontsScaled = false;
			Legend.IsVisible = false;
			monochrome = false;
			Max = 5;

		}

		#endregion

		#region Rendering

		// Renders the graph
		public void Render(System.IO.Stream stream)
		{
			Bitmap bmp = drawPane();
			bmp.Save(stream, ImageFormat.Jpeg);
			bmp.Dispose();
		}


		private Bitmap drawPane()
		{
			PaneBorder.IsVisible = false;

			// Makes the y axis numbers invisible
			YAxis.IsVisible = false;

			// Set the Y axis max line
			YAxis.Max = Max + shift;

			// Fill the axis background with a color gradient
			AxisFill = Background;

			if (!DontDrawLabel)
			{
				// leave some space for the labels on the left
				MarginLeft = LabelFraction * PaneRect.Width;
			}

			// Draw in the grid lines
			for (int index = 0; index < grids.Length; index++)
			{
				float cutoffValue = grids[index];
				// Add a BoxItem to show a colored band behind the graph data
				BoxItem cutOff = new BoxItem(new RectangleF((float)XAxis.Max, cutoffValue, (float)XAxis.Max - .5F, .0001F));
				cutOff.Fill = new Fill(Color.Black);
				cutOff.Fill.IsVisible = false;
				cutOff.Location.CoordinateFrame = CoordType.AxisXYScale;
				// Align the center-top of the box
				cutOff.Location.AlignH = AlignH.Left;
				cutOff.Location.AlignV = AlignV.Center;
				// place the box behind the axis items, so the bars are drawn on top of it
				cutOff.ZOrder = ZOrder.E_BehindAxis;
				GraphItemList.Add(cutOff);
				if (!DontDrawLabel)
				{
					// create the label for this cutoff on the left
					TextItem myText = new TextItem(CuttoffLabels[index], (float)XAxis.Max + .05F, cutoffValue,
					CoordType.AxisXYScale, AlignH.Right, AlignV.Center);
					myText.FontSpec.Size = LabelFontSpec.Size;
					myText.FontSpec.IsItalic = LabelFontSpec.IsItalic;
					myText.FontSpec.Border.IsVisible = LabelFontSpec.Border.IsVisible;
					GraphItemList.Add(myText);
				}
			}

			return Image;
		}


		/// <summary>
		/// Renders a 2 or 4 bar vertical graph.
		/// </summary>
		/// <param name="contentType">Out parameter set to the content type:"image/jpeg".</param>
		/// <param name="title">String title of the graph.</param>
		/// <param name="data">float data for the 2 or four graphs.</param>
		/// <param name="labels">String labels on the graph.</param>
		/// <param name="width">Int width in px.</param>
		/// <param name="height">Int height in px.</param>
		/// <param name="monochrome">Bool true or false. null implies false.</param>
		/// <param name="dontDrawLabel">Bool of true or false. null implies false.</param>
		/// <param name="MaxY">integer value of the max score possible.</param>
		/// <returns></returns>
		[WebDirect]
		public static Stream Render(out string contentType, string title, float[] data,
			string[] labels, float? width, float? height, bool monochrome, bool dontDrawLabel, int maxY)
		{
			if (data == null) throw new ArgumentNullException("data");
			if (title == null) throw new ArgumentNullException("title");
			if (labels == null) throw new ArgumentNullException("labels");
			if (maxY <= 0) throw new ArgumentOutOfRangeException("MaxY", "Must be a positive number.");

			if (!width.HasValue)
				width = 350;
			else if (width <= 0) throw new ArgumentOutOfRangeException("width", width, Resources.PositiveNumberRequired);
			if (!height.HasValue)
				height = 300;
			else if (height <= 0) throw new ArgumentOutOfRangeException("height", height, Resources.PositiveNumberRequired);
			//if (data.Length != 4 && data.Length != 2 && data.Length != 1) throw new ArgumentOutOfRangeException("data", "Only 1, 2, or 4 scores are allowed");

			MemoryStream ms = new MemoryStream(5 * 1024);

			GradientGraph graph = new GradientGraph();

			graph.Monochrome = monochrome;

			graph.DontDrawLabel = dontDrawLabel;

			graph.Max = maxY;

			if (labels.Length != graph.Max)
				throw new ArgumentOutOfRangeException("labels",
					"There must be MaxY labels.  In this case "
					+ graph.Max + ". " + labels.Length + " were provided.");
			graph.strengthLabels = labels;

			graph.Title = title;

			if (data.Length <= 2)
			{
				graph.IsTwo = true;
			}

			foreach (float var in data)
			{
				// ideally we would pass in the labels but the default
				// rendering of the text is unsatisfactory. We pass in
				// "" to prevent any rendering.
				graph.AddBar("", var);
			}
			graph.SetXAxis();
			graph.Resize(width.Value, height.Value);

			graph.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}

		#endregion

		#region Helper functions

		// Compute the pseudo-logrithimic scale
		private static float calcHeight(float preCalcHeight, float height)
		{
			float lineInterval = (float)height / 9F * 5F / height;
			// these factors are used to preserve the same effect with different heights.
			// the ideal is still with height 5.
			float scalingFactor = 3F * height / 5F;
			float scalingFactor2 = 3F * 75F / (height + 70F);

			if (preCalcHeight > scalingFactor)
				return (float)((lineInterval * scalingFactor) + ((preCalcHeight - scalingFactor) * (lineInterval * scalingFactor2)));
			else
				return (float)(preCalcHeight * lineInterval);

		}

		public void AddBar(string name, float score)
		{

			//  the AddBar method requires double[] so we will convert up for the score
			double[] ParticipantScores = new double[1];

			// convert the score to a pseudo Logrithimic scale
			ParticipantScores[0] = calcHeight(score, Max);
			Color parColor = new Color();
			switch (CurveList.NumBars)
			{
				case 0: parColor = ParticipantColor;
					break;
				case 1:
					if (!IsTwo) // skip to the partner
					{
						parColor = ParticipantPartnerColor;
					}
					else
					{
						parColor = Participant2Color;
					}
					break;
				case 2: parColor = Participant2PartnerColor;
					break;
				case 3: parColor = Participant2Color;
					break;
			}
			Color gradColor = Color.FromArgb(parColor.A, (int)(parColor.R * gradRatio), (int)(parColor.G * gradRatio), (int)(parColor.B * gradRatio));
			if (Monochrome) // no gradRatio needed
				gradColor = parColor;
			// Generate a bar with <Name> in the legend
			BarItem myCurve = AddBar(name, null, ParticipantScores, parColor);
			// Fill the bar with a gradient
			myCurve.Bar.Fill = new Fill(parColor, gradColor, 90F);
			if (Monochrome && parColor == Color.Black) // outline the black bar in a non-black color
				myCurve.Bar.Border.Color = Color.Gray;
		}

		public void SetXAxis()
		{
			// Set the XAxis labels
			XAxis.IsVisible = false;
			// Removes the X tics at the labels
			XAxis.IsTic = false;
			// Reverse the X axis data to display in the desirec order
			XAxis.IsReverse = true;

			//center the bars
			// draw the X axis up to 1.5
			XAxis.Max = 1.5;
			// draw the X axis from .5
			XAxis.Min = .5;

			// This is the part that makes the bars verticle
			BarBase = BarBase.X;
		}

		public void Resize(float width, float height)
		{
			// for fontsize changes, we only care about how wide the graph is
			float ResizeFactor = width / PaneRect.Width;
			OverallFontSize = (int)((float)OverallFontSize * ResizeFactor);
			PaneRect = new RectangleF(0, 0, width, height);

		}

		#endregion

		#region Testing

		[WebDirect]
		public static Stream Test(out string contentType)
		{

			string title = "Test Graph";
			float data1 = 3.4F;
			float data12 = .4F;
			float data21 = 4.4F;
			float data2 = 2.4F;
			string[] labels = { "label1", "label2", "label3", "label4", "label5" };
			int maxy = 5;
			int Width = 350;
			int Height = 300;
			bool mono = false;
			bool dontDrawL = false;

			MemoryStream ms = new MemoryStream(5 * 1024);
			GradientGraph test = new GradientGraph();
			test.DontDrawLabel = dontDrawL;
			test.CuttoffLabels = labels;
			test.Monochrome = mono;
			test.Max = maxy;
			test.Title = title;
			test.AddBar("", data1);
			test.AddBar("", data12);
			test.AddBar("", data21);
			test.AddBar("", data2);
			test.SetXAxis();
			test.Resize(Width, Height);
			test.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}

		#endregion
	}
}