using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections;
using System.Diagnostics;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// Summary description for Labels.
	/// </summary>
	public class GraphLabels
	{
		#region Attributes
		public ArrayList Labels = new ArrayList();
		public int Height = (new GraphWindow()).Size.Height;
		private int _Width;
		public int Width
		{
			get
			{
				if (_Width > 0) return _Width;
				// Calculate width based on the labels we have

				const float padding = 1.1F;

				// First create some dummy bitmap
				Bitmap dummyBitmap = new Bitmap(1, 1);
				Graphics drawTool = Graphics.FromImage(dummyBitmap);
				Font fontLabel = createFont();
				int lengthLongest = 0;
				try
				{
					drawTool.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

					// Now go through all the labels, measuring them, to see which is longest
					foreach (string label in Labels)
					{
						int iWidth = (int)drawTool.MeasureString(label, fontLabel).Width;
						lengthLongest = Math.Max(lengthLongest, iWidth);
					}
				}
				finally
				{
					fontLabel.Dispose();
					drawTool.Dispose();
					dummyBitmap.Dispose();
				}
				_Width = (int)(lengthLongest * padding); // cache what we've come up with
				return _Width;
			}
		}
		public BarGraph.BarGraphStyle Style = BarGraph.StdStyle;
		#endregion

		private static Font createFont()
		{
			return new Font("Helvetica, SansSerif", 10, FontStyle.Bold, GraphicsUnit.Pixel);
		}

		private static StringFormat createStringFormat()
		{
			StringFormat _strFormat;
			_strFormat = new StringFormat();
			_strFormat.Alignment = StringAlignment.Far;
			_strFormat.LineAlignment = StringAlignment.Center;
			_strFormat.Trimming = StringTrimming.None;

			return _strFormat;
		}

		private float MaxYScale
		{
			get
			{
				return Labels.Count;
			}
		}

		public GraphLabels()
		{
		}

		#region Render
		public Bitmap Render()
		{
			if (Width == 0 || Height == 0) throw new InvalidOperationException(Resources.SetHeightAndWidthFirst);
			Bitmap bmpCanvas = new Bitmap(Width, Height);
			Graphics drawTool = Graphics.FromImage(bmpCanvas);
			StringFormat strFormat = createStringFormat();
			Font fontLabel = createFont();
			try
			{
				drawTool.Clear(Color.White);
				drawTool.CompositingMode = CompositingMode.SourceOver;
				drawTool.SmoothingMode = SmoothingMode.HighQuality;
				drawTool.CompositingQuality = CompositingQuality.HighQuality;
				drawTool.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

				int iLabel = 1;
				BarGraph sampleBox = new BarGraph();
				//Scale Scale = new Scale(SurveyID, ScaleID, Context);
				sampleBox.Size.Height = this.Height;
				sampleBox.Style = Style;
				sampleBox.MaxYScale = MaxYScale;
				foreach (string label in Labels)
				{
					int y = sampleBox.ContentArea.Bottom - sampleBox.CalcHeight(iLabel);
					RectangleF rectText = new RectangleF(0, (float)(y - fontLabel.Height * 1.2 / 2), Width, (float)(fontLabel.Height * 1.2));
					drawTool.DrawString(label, fontLabel, Brushes.Black, rectText, strFormat);
					iLabel += 1;
				}
			}
			finally
			{
				fontLabel.Dispose();
				strFormat.Dispose();
				drawTool.Dispose();
			}
			return bmpCanvas;
		}
		public void Render(System.IO.Stream stream)
		{
			Bitmap bmp = Render();
			bmp.Save(stream, ImageFormat.Png);
			bmp.Dispose();
		}

		[WebDirect]
		public static Stream Render(out string contentType, BarGraph.BarGraphStyle style, int height, params string[] labels)
		{
			if (labels == null) throw new ArgumentNullException("labels");
			GraphLabels gl = new GraphLabels();

			foreach (string label in labels)
				gl.Labels.Add(label);

			gl.Style = style;
			if (height > 0) gl.Height = height;

			System.IO.MemoryStream ms = new System.IO.MemoryStream(5 * 1024);
			gl.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/png";
			return ms;
		}
		#endregion

		#region Test routine
		[WebDirect]
		public static Stream Test(out string contentType)
		{
			GraphLabels gl = new GraphLabels();

			gl.Labels.Add("Never");
			gl.Labels.Add("Rarely");
			gl.Labels.Add("Sometimes");
			gl.Labels.Add("Often");
			gl.Labels.Add("Very Often");

			gl.Style = BarGraph.BarGraphStyle.Logarithmic;

			MemoryStream ms = new MemoryStream(5 * 1024);
			gl.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/gif";
			return ms;
		}

		#endregion
	}
}
