using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections.Specialized;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// Summary description for GraphWindow.
	/// </summary>
	public class GraphWindow
	{
		public Size Size = new Size(250, 275);
		public string Title;
		public int TitleSize = 15; // the size in pixels of the graph title
		public int BorderWidth = 2;
		public bool[] CurveCorners = new bool[6];
		public enum GraphCorner
		{
			TopLeft = 0,
			TopRight,
			BottomLeft,
			BottomRight,
			InsideRight,
			InsideLeft,
		}
		public bool MonochromeMode;
		private Rectangle boxArea = new Rectangle(0, 0, 0, 0);
		public Rectangle BoxArea
		{
			get
			{
				if (!boxArea.IsEmpty) return boxArea;
				Rectangle rect = new Rectangle(0, 0, Size.Width, Size.Height);
				rect.Inflate(-FeatherWidth, -FeatherWidth);
				return rect;
			}
			set
			{
				if (value.Bottom > Size.Height || value.Right > Size.Width)
					throw new ArgumentOutOfRangeException("BoxArea", Resources.BoxBoundariesOutsideBitmap);
				boxArea = value;
			}
		}
		public Rectangle ContentArea
		{
			get
			{
				// pull in sides to leave room for curve and just asthetic space
				Rectangle rect = Rectangle.FromLTRB(BoxArea.Left, BoxArea.Top, BoxArea.Right, BoxArea.Bottom - titleArea.Height);
				rect.Inflate(-CurveRadius / 2, -FeatherWidth);
				return rect;
			}
		}
		public Single TitleSpace = 1.5F; // multiple of titleSize that makes up the space for the title

		protected int CurveCornersBit(GraphCorner corner)
		{
			return CurveCorners[(int)corner] ? 1 : 0;
		}

		public GraphWindow()
		{
			// Set standard corner-curve settings
			for (int i = 0; i < CurveCorners.Length; i++)
				CurveCorners[i] = true;
			CurveCorners[(int)GraphCorner.BottomLeft] = false;
			CurveCorners[(int)GraphCorner.InsideLeft] = false;
		}
		public void Render(System.IO.Stream stream)
		{
			Render(stream, ImageFormat.Jpeg);
		}
		public void Render(System.IO.Stream stream, ImageFormat imageFormat)
		{
			Bitmap bmp = Render();
			bmp.Save(stream, imageFormat);
			bmp.Dispose();
		}
		public virtual Bitmap Render()
		{
			// prepare the background bitmap
			Bitmap bmpInput = new Bitmap(Size.Width, Size.Height);

			// Set the background as the canvas
			Graphics drawTool = Graphics.FromImage(bmpInput);
			drawTool.CompositingQuality = CompositingQuality.HighQuality;
			drawTool.TextRenderingHint = TextRenderingHint.AntiAlias;
			drawTool.SmoothingMode = SmoothingMode.AntiAlias;

			// Make the background white
			drawTool.FillRectangle(Brushes.White, 0, 0, bmpInput.Width, bmpInput.Height);

			drawTitleBackground(drawTool);
			drawTitle(drawTool);

			Pen penBlackLine = new Pen(Color.Black, BorderWidth);

			//---------------------------------------------------------------------
			// Create the graph container
			//---------------------------------------------------------------------

			drawTool.DrawLine(penBlackLine, BoxArea.Left + (CurveRadius / 2) * CurveCornersBit(GraphCorner.InsideLeft), ContentArea.Bottom + FeatherWidth, BoxArea.Right - (CurveRadius / 2) * CurveCornersBit(GraphCorner.InsideRight), ContentArea.Bottom + FeatherWidth);
			drawTool.DrawLine(penBlackLine, BoxArea.Right, BoxArea.Top + (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopRight), BoxArea.Right, BoxArea.Bottom - (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomRight));
			drawTool.DrawLine(penBlackLine, BoxArea.Left, BoxArea.Top + (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopLeft), BoxArea.Left, BoxArea.Bottom - (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomLeft));
			drawTool.DrawLine(penBlackLine, BoxArea.Left + (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopLeft), BoxArea.Top, BoxArea.Right - (CurveRadius / 2) * CurveCornersBit(GraphCorner.TopRight), BoxArea.Top);
			drawTool.DrawLine(penBlackLine, BoxArea.Left + (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomLeft), BoxArea.Bottom, BoxArea.Right - (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomRight), BoxArea.Bottom);
			//    |---------D->--------|
			//    | 3                        2 |
			//    |                             |
			//    C                             B              LETTERS ABOVE
			//    \/ 4                       1 \/             NUMBERS BELOW
			//    |-------A->----------|
			//    | 5                       6 |
			//    |--------E->---------|
			if (CurveCorners[(int)GraphCorner.InsideRight]) drawTool.DrawArc(penBlackLine, BoxArea.Right - CurveRadius, ContentArea.Bottom - CurveRadius + FeatherWidth, CurveRadius, CurveRadius, 0, 90);
			if (CurveCorners[(int)GraphCorner.TopRight]) drawTool.DrawArc(penBlackLine, BoxArea.Right - CurveRadius, BoxArea.Top, CurveRadius, CurveRadius, 270, 90);
			if (CurveCorners[(int)GraphCorner.TopLeft]) drawTool.DrawArc(penBlackLine, BoxArea.Left, BoxArea.Top, CurveRadius, CurveRadius, 180, 90);
			if (CurveCorners[(int)GraphCorner.InsideLeft]) drawTool.DrawArc(penBlackLine, BoxArea.Left, ContentArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 90, 90);
			if (CurveCorners[(int)GraphCorner.BottomLeft]) drawTool.DrawArc(penBlackLine, BoxArea.Left, BoxArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 90, 90);
			if (CurveCorners[(int)GraphCorner.BottomRight]) drawTool.DrawArc(penBlackLine, BoxArea.Right - CurveRadius, BoxArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 0, 90);
			penBlackLine.Dispose();

			// Draw a box to indicate where the contentArea is
			Pen pen = new Pen(Color.Red, 2);
			pen.DashStyle = DashStyle.DashDot;
			// drawTool.DrawRectangle(pen, contentArea)
			pen.Dispose();

			drawTool.Dispose();

			return bmpInput;
		}

		protected int FeatherWidth
		{
			get
			{
				return BorderWidth;
			}
		}
		private Rectangle titleArea
		{
			get
			{
				return Rectangle.FromLTRB(BoxArea.Left, (int)(BoxArea.Bottom - TitleSize * TitleSpace), BoxArea.Right, BoxArea.Bottom);
			}
		}
		protected int CurveRadius
		{
			get
			{
				return (int)((Size.Width + Size.Height) / 2 * 0.15);
			}
		}
		private static Color titleBkgrndColor = Color.FromArgb(unchecked((int)0xFF767676));  // Teal
		private static Color titleBkgrndColorBW = Color.FromArgb(unchecked((int)0xFF767676)); // Silver
		protected static Color TitleBackgroundColor(bool monochromeMode)
		{
			if (!monochromeMode)
				return titleBkgrndColor;
			else
				return titleBkgrndColorBW;
		}
		private void drawTitleBackground(Graphics drawTool)
		{
			// Draw the title background
			SolidBrush drawBrush = new SolidBrush(TitleBackgroundColor(MonochromeMode));
			if (CurveCorners[(int)GraphCorner.BottomLeft]) drawTool.FillPie(drawBrush, BoxArea.Left, BoxArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 90, 90);
			if (CurveCorners[(int)GraphCorner.BottomRight]) drawTool.FillPie(drawBrush, BoxArea.Right - CurveRadius, BoxArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 0, 90);
			drawTool.FillRectangle(drawBrush, Rectangle.FromLTRB(BoxArea.Left + (CurveRadius / 2) - FeatherWidth - 1, ContentArea.Bottom, (BoxArea.Right - CurveRadius / 2) + FeatherWidth + 1, BoxArea.Bottom));
			drawTool.FillRectangle(drawBrush, Rectangle.FromLTRB(BoxArea.Left, ContentArea.Bottom - (CurveRadius / 2 - 1) * CurveCornersBit(GraphCorner.InsideLeft), BoxArea.Left + CurveRadius / 2 - FeatherWidth, BoxArea.Bottom - (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomLeft) + 1));
			drawTool.FillRectangle(drawBrush, Rectangle.FromLTRB(BoxArea.Right - CurveRadius / 2 + FeatherWidth, ContentArea.Bottom - (CurveRadius / 2 - 1) * CurveCornersBit(GraphCorner.InsideRight), BoxArea.Right, BoxArea.Bottom - (CurveRadius / 2) * CurveCornersBit(GraphCorner.BottomRight) + 1));
			if (CurveCorners[(int)GraphCorner.InsideRight]) drawTool.FillPie(Brushes.White, BoxArea.Right - CurveRadius - 1, ContentArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 0, 90);
			if (CurveCorners[(int)GraphCorner.InsideLeft]) drawTool.FillPie(Brushes.White, BoxArea.Left, ContentArea.Bottom - CurveRadius, CurveRadius, CurveRadius, 90, 90);

			drawBrush.Dispose();
		}
		private void drawTitle(Graphics drawTool)
		{
			//---------------------------------------------------------------------
			// Label the graph appropriately
			//---------------------------------------------------------------------

			// Calculate a smaller rectangle appropriate for the label to actually belong
			RectangleF rectArea = new RectangleF();
			rectArea.X = titleArea.X; rectArea.Y = titleArea.Y + FeatherWidth;
			rectArea.Width = titleArea.Width; rectArea.Height = titleArea.Height - FeatherWidth * 2;
			rectArea.Inflate(-CurveRadius / 4, 0);
			//drawTool.DrawRectangle(Pens.HotPink, rectArea.X, rectArea.Y, rectArea.Width, rectArea.Height);

			// Prepare to format the string to center horizontally
			StringFormat strFormat = new StringFormat();
			strFormat.Alignment = StringAlignment.Near;
			strFormat.LineAlignment = StringAlignment.Center;

			// Prepare font.  We size the font to max out at 15px, but it can get smaller if needed
			Font labelFont = new Font("SansSerif", Math.Min(15, rectArea.Height - 2), FontStyle.Bold, GraphicsUnit.Pixel);
			// We want to be very careful about drawing the title to make sure that it will fit.
			// Call Graphics.MeasureString to discover where this would go and shrink the font until it fits.
			while (drawTool.MeasureString(Title, labelFont).Width > rectArea.Width && labelFont.Size > 1)
			{
				labelFont.Dispose();
				labelFont = new Font(labelFont.FontFamily, labelFont.Size - 1, labelFont.Style, GraphicsUnit.Pixel);
			}

			drawTool.DrawString(Title, labelFont, Brushes.White, rectArea, strFormat);

			labelFont.Dispose();

		}

		[WebDirect]
		public static Stream Test(out string contentType)
		{
			MemoryStream ms = new MemoryStream(5 * 1024);
			(new GraphWindow()).Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}
	}
}
