using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// Summary description for Legend.
	/// </summary>
	public class Legend : GraphWindow
	{
		public ArrayList Labels = new ArrayList();
		public readonly static Size StandardSize = new Size( 180, 60 );
		public readonly static Size DoubleSize = new Size(180, 100);

		public Legend()
		{
			base.TitleSpace = 1.2F;
			base.BorderWidth = 1;
			base.TitleSize = 14;
			base.Size = StandardSize;
		}
		
		#region Rendering
		public override Bitmap Render()
		{
			if( Labels.Count == 0 ) throw new InvalidOperationException(Resources.LabelsRequiredFirst);

			Bitmap bmpCanvas = base.Render();

			Graphics drawTool = Graphics.FromImage(bmpCanvas);

			// Generate locations for each label
			int RealCount = 0;
			foreach(string label in Labels)
				if( label.Substring(0,1) != "-" ) RealCount += 1;
		
			int ySpace = ContentArea.Height / RealCount; // vertical space available
			// Reserve space for boxes
			int boxSize = 10;
			// Start at the top
			int ySpot = ContentArea.Top;
			int iCounter = 0;
			foreach( string Label in Labels )
			{
				if( Label.Substring(0,1) == "-" )
				{
					// draw a line instead of a box
					drawTool.DrawLine(Pens.Black, ContentArea.Left, ySpot, ContentArea.Right, ySpot);
				} 
				else
				{
					// Draw box
					SolidBrush brush = new SolidBrush(BarGraph.BarColor(iCounter, RealCount, MonochromeMode));
					Rectangle rectBox = new Rectangle(ContentArea.Left, ySpot + ySpace / 2 - boxSize / 2, boxSize, boxSize);
					drawTool.FillRectangle(brush, rectBox);
					drawTool.DrawRectangle(Pens.Black, rectBox);
					// Draw label
					RectangleF rectLabel = RectangleF.FromLTRB(ContentArea.Left + boxSize + 5, ySpot, ContentArea.Right, ySpot + ySpace);
					DrawLabel(drawTool, rectLabel, Label);

					brush.Dispose();
					ySpot += ySpace;
					iCounter += 1;
				}
			}

			drawTool.Dispose();

			return bmpCanvas;
		}

		private static void DrawLabel(Graphics drawTool, RectangleF rectArea, string Label)
		{
			// Prepare font.  We size the font to max out at 10px, but it can get smaller if needed
			Font labelFont = new Font("SansSerif", Math.Min(10, rectArea.Height - 2), FontStyle.Regular, GraphicsUnit.Pixel);
			// We want to be very careful about drawing the title to make sure that it will fit.
			// Call Graphics.MeasureString to discover where this would go and shrink the font until it fits.
			while( drawTool.MeasureString(Label, labelFont).Width > rectArea.Width && labelFont.Size > 1 )
			{
				labelFont.Dispose();
				labelFont = new Font(labelFont.FontFamily, labelFont.Size - 1, labelFont.Style, GraphicsUnit.Pixel);
			}

			// Prepare to format the string to center horizontally
			StringFormat strFormat = new StringFormat();
			strFormat.Alignment = StringAlignment.Near;
			strFormat.LineAlignment = StringAlignment.Center;

			drawTool.DrawString(Label, labelFont, Brushes.Black, rectArea, strFormat);
		}

		/// <summary>
		/// Generates a Legend box for browser viewing.
		/// </summary>
		/// <param name="contentType">
		/// The ContentType to pass through to the browser that is downloading the image.
		/// This is set by the type of image generated (jpeg/png/gif).
		/// </param>
		/// <param name="title">
		/// The cultured title to put on the Legend box.  This is typically "Graph Legend".
		/// </param>
		/// <param name="labels">
		/// An ordered array of elements to list in the legend.
		/// Typical contents: (example)
		///		Andrew describes Cheryl
		///		Cheryl describes Andrew
		///		Andrew describes Andrew
		///		Cheryl describes Cheryl
		/// </param>
		/// <returns>
		/// A stream with the graphic to stream to the browser.
		/// </returns>
		/// <example>
		/// URL parameters: ?title=Graph+Legend&amp;labels=Andrew+describes+Andrew&amp;labels=Andrew+describes+Cheryl&amp;labels=-&amp;labels=Cheryl+describes+Andrew&amp;labels=Cheryl+describes+Cheryl
		/// </example>
		[WebDirect]
		public static Stream Render(out string contentType, string title,bool monochrome, params string[] labels)
		{
			if( labels == null ) throw new ArgumentNullException("labels");
			Legend l = new Legend();
			l.MonochromeMode = monochrome;
			l.Title = title;
			l.Size = labels.Length > 2 ? Legend.DoubleSize : Legend.StandardSize;
			
			foreach( string label in labels )
				l.Labels.Add( label );

			System.IO.MemoryStream ms = new System.IO.MemoryStream(5 * 1024);
			l.Render(ms, ImageFormat.Png);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/png";
			return ms;
		}
		#endregion

		#region Test routine
		[WebDirect]
		public new static System.IO.Stream Test(out string contentType)
		{
			Legend l = new Legend();
			l.Title = "Graph Legend";
			l.Size = Legend.DoubleSize;
			l.Labels.Add("Jake describes Jake");
			l.Labels.Add("Jake describes Rhonda");
			l.Labels.Add("-");
			l.Labels.Add("Rhonda describes Jake");
			l.Labels.Add("Rhonda describes Rhonda");	

			System.IO.MemoryStream ms = new System.IO.MemoryStream(5 * 1024);
			l.Render(ms, ImageFormat.Png);
			ms.Position = 0;
			contentType = "image/png";
			return ms;
		}
		#endregion
	}
}
