using System;
using System.IO;
using System.Xml;
using System.Drawing;
using System.Drawing.Text;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Collections.Generic;
using ZedGraph;

using NerdBank.Tools;

namespace NAssess.Core.Engines.Reporting.Graphs
{
	/// <summary>
	/// A multipule horizontal bar graph image.
	/// </summary>
	[Serializable]
	public class MultiBar : GraphPane
	{

		#region variables

		// the maximum score value
		private const float Max = 5;

		private bool monochrome;

		public bool Monochrome
		{
			get
			{
				return monochrome;
			}
			set
			{
				monochrome = value;
			}
		}

		// the background coloring
		private static Color[] backgroundColors = new Color[] { Color.Salmon, Color.White, Color.White, Color.SeaGreen };
		private static float[] backgroundPos = new float[] { 0, calcHeight(3, Max) / (float)Max, calcHeight(4, Max) / (float)Max, 1 };
		private Fill background = new Fill(backgroundColors, backgroundPos);
		private static Color[] backgroundColorsBW = new Color[] { Color.DarkGray, Color.White, Color.White, Color.LightGray };
		private Fill backgroundBW = new Fill(backgroundColorsBW, backgroundPos);

		public static IList<Color> BackgroundColors
		{ get { return new ReadOnlyCollection<Color>(backgroundColors); } }

		public static Color[] BackgroundPos
		{
			get
			{
				return BackgroundPos;
			}
		}

		public Fill Background
		{
			get
			{
				if (Monochrome)
					return backgroundBW;
				else
					return background;
			}
		}

		private double[] grids = new double[] { 0, 1, 2, 3, 4, 5 };

		// shift the cut offs to make it more aesthetically pleasing
		private const float shift = .043F;

		private ReadOnlyCollection<string> cutoffLabels;
		/// <summary>
		/// Labels for challenge-neutral-asset.
		/// </summary>
		public ReadOnlyCollection<string> CutoffLabels
		{
			get
			{
				return cutoffLabels;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("CutoffLabels");
				if (value.Count < 3)
					throw new ArgumentOutOfRangeException("CutoffLabels", Resources.MoreCutoffsRequired);
				cutoffLabels = value;
			}
		}

		// Colors to match with preexisting Relate graphs
		private Color participantColor = Color.FromArgb(unchecked((int)0xFF3C4473));
		private Color participant2Color = Color.FromArgb(unchecked((int)0xFFCB692E));
		private Color participantColorBW = Color.Black;
		private Color participant2ColorBW = Color.White;
		private const float gradRatio = 0.5f;

		public Color ParticipantColor
		{
			get
			{
				if (Monochrome)
					return participantColorBW;
				else
					return participantColor;
			}
		}
		public Color Participant2Color
		{
			get
			{
				if (Monochrome)
					return participant2ColorBW;
				else
					return participant2Color;
			}
		}

		public string GraphTitle
		{
			get
			{
				return title;
			}
			set
			{
				base.Title = value;

				// Turn off the Default Title
				IsShowTitle = false;

				// Create the title
				TextItem TitleText = new TextItem(Title, .5F, .01F,
					CoordType.PaneFraction, AlignH.Center, AlignV.Top);
				TitleText.FontSpec.Size = 16;
				TitleText.FontSpec.IsBold = true;
				TitleText.FontSpec.Border.IsVisible = false;
				GraphItemList.Add(TitleText);
				MarginTop = 47;
			}
		}

		#endregion

		#region Constructors

		MultiBar()
		{
			for (int i = 0; i < grids.Length; i++)
				grids[i] = calcHeight(grids[i], Max);
			cutoffLabels = new ReadOnlyCollection<string>(new string[] { 
				Resources.CutoffLabelWeakness, Resources.CutoffLabelAverage, Resources.CutoffLabelStrength
			});
		}

		#endregion

		#region Rendering

		// Renders the graph
		public void Render(System.IO.Stream stream)
		{
			Bitmap bmp = DrawPane();
			bmp.Save(stream, ImageFormat.Jpeg);
			bmp.Dispose();
		}


		private Bitmap DrawPane()
		{
			PaneBorder.IsVisible = false;

			// Makes the x axis numbers invisible
			XAxis.IsVisible = false;

			// Set the X axis max line
			XAxis.Max = Max + shift;

			// Fill the axis background with a color gradient
			AxisFill = Background;

			// Draw in the grid lines
			foreach (float cutoff in grids)
			{
				// Add a BoxItem to show a colored band behind the graph data
				BoxItem CutOff = new BoxItem(new RectangleF(cutoff, .5F, .0001F, (float)YAxis.Max - .5F));
				CutOff.Fill = new Fill(Color.Black);
				CutOff.Fill.IsVisible = false;
				CutOff.Location.CoordinateFrame = CoordType.AxisXYScale;
				// Align the center-top of the box
				CutOff.Location.AlignH = AlignH.Center;
				CutOff.Location.AlignV = AlignV.Top;
				// place the box behind the axis items, so the grid is drawn on top of it
				CutOff.ZOrder = ZOrder.E_BehindAxis;
				GraphItemList.Add(CutOff);
			}

			// the Font for the challenge-neutral-asset labels
			FontSpec labelFontSpec =
				new FontSpec("Comic Sans MS", 15, Color.Empty, true, false, false);
			labelFontSpec.Border.IsVisible = false;

			// Create the first X-axis label
			TextItem myText = new TextItem(cutoffLabels[0], 0, 0,
				CoordType.AxisXYScale, AlignH.Left, AlignV.Top);
			myText.FontSpec.Size = labelFontSpec.Size;
			myText.FontSpec.IsItalic = labelFontSpec.IsItalic;
			myText.FontSpec.Border.IsVisible = labelFontSpec.Border.IsVisible;
			GraphItemList.Add(myText);

			// Create the second X-axis label
			myText = new TextItem(cutoffLabels[1], (float)calcHeight(3.5, Max), 0,
				CoordType.AxisXYScale, AlignH.Center, AlignV.Top);
			myText.FontSpec.Size = labelFontSpec.Size;
			myText.FontSpec.IsItalic = labelFontSpec.IsItalic;
			myText.FontSpec.Border.IsVisible = labelFontSpec.Border.IsVisible;
			GraphItemList.Add(myText);

			// Create the last X-axis label
			myText = new TextItem(cutoffLabels[2], (float)XAxis.Max, 0,
				CoordType.AxisXYScale, AlignH.Right, AlignV.Top);
			myText.FontSpec.Size = labelFontSpec.Size;
			myText.FontSpec.IsItalic = labelFontSpec.IsItalic;
			myText.FontSpec.Border.IsVisible = labelFontSpec.Border.IsVisible;
			GraphItemList.Add(myText);



			// Place the Legend at the bottom of the graph
			Legend.Position = LegendPos.BottomCenter;

			return Image;
		}


		[WebDirect]
		public static Stream Render(out string contentType, string title, string[] data1, string[] data2, string name1,
			string name2, string[] scales, string[] labels, string width, string height, bool monochrome)
		{
			if (scales == null) throw new ArgumentNullException("scales", Resources.AtLeastOneScaleRequired);
			if (title == null) throw new ArgumentNullException("title");
			if (name1 == null) throw new ArgumentNullException("name1");
			//if (name2 == null) throw new ArgumentNullException("name2");
			if (labels == null) throw new ArgumentNullException("labels");

			int Width = 450;
			if (width != null)
				Width = int.Parse(width);
			int Height = 338;
			if (height != null)
				Height = int.Parse(height);

			if (Width < 0) throw new ArgumentOutOfRangeException("width", width, Resources.NonNegativeValueRequired);
			if (Height < 0) throw new ArgumentOutOfRangeException("height", height, Resources.NonNegativeValueRequired);
			//if( style == null ) style = StyleEnum.Logarithmic;

			MemoryStream ms = new MemoryStream(5 * 1024);
			MultiBar test = new MultiBar();
			test.Monochrome = monochrome;
			if (labels.Length >= 3)
				test.cutoffLabels = new ReadOnlyCollection<string>(labels);
			test.GraphTitle = title;
			test.AddBar(name1, data1);
			if (name2 != null)
				test.AddBar(name2, data2);
			test.SetYAxis(scales);
			test.Resize(Width, Height);
			test.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}

		#endregion

		#region Functions

		// Compute the pseudo-logrithimic scale
		private static float calcHeight(double preCalcHeight, double height)
		{
			float lineInterval = (float)height / 9.0F;
			if (preCalcHeight > 3.0)
				return (float)((lineInterval * 3) + ((preCalcHeight - 3) * (lineInterval * 3)));
			else
				return (float)(preCalcHeight * lineInterval);

		}

		public void AddBar(string name, string[] scores)
		{
			if (scores == null) throw new ArgumentNullException("scores");
			double[] ParticipantScores = new double[scores.Length];

			// convert the scores to a pseudo Logrithimic scale
			for (int i = 0; i < ParticipantScores.Length; i++)
			{
				ParticipantScores[i] = calcHeight(double.Parse(scores[i]), Max);
			}

			Color parColor = new Color();
			if (CurveList.NumBars == 0)
			{
				parColor = ParticipantColor;
			}
			else
			{
				parColor = Participant2Color;
			}

			Color gradColor = Color.FromArgb(parColor.A, (int)(parColor.R * gradRatio), (int)(parColor.G * gradRatio), (int)(parColor.B * gradRatio));

			// Generate a bar with <Name> in the legend
			BarItem myCurve = AddBar(name, ParticipantScores, null, parColor);
			// Fill the bar with a gradient
			myCurve.Bar.Fill = new Fill(parColor, gradColor, 90F);
		}

		public void SetYAxis(string[] scales)
		{
			if (scales == null) throw new ArgumentNullException("scales");
			// label the Y axis
			YAxis.Title = "Scale";

			// Set the YAxis labels
			YAxis.TextLabels = scales;

			// Set the YAxis to Text type
			YAxis.Type = AxisType.Text;

			// Removes the Y tics at the labels
			YAxis.IsTic = false;
			// Reverse the Y axis data to display in the desirec order
			YAxis.IsReverse = true;
			// Set the desired steps inbetween each label to 1
			YAxis.Step = 1;
			//turn off the intermitiate tics
			YAxis.IsMinorTic = false;
			YAxis.IsMinorOppositeTic = false;
			// draw the Y axis up to the length of the scales
			YAxis.Max = scales.Length + .5;
			// draw the Y axis from .5
			YAxis.Min = .5;

			// This is the part that makes the bars horizontal
			BarBase = BarBase.Y;
		}

		public void Resize(float width, float height)
		{
			PaneRect = new RectangleF(0, 0, width, height);
		}

		#endregion

		#region Testing

		[WebDirect]
		public static Stream Test(out string contentType)
		{

			string title = "Test Graph";
			string[] data1 = { ".4", "1.4", "2.4", "3.4", "4.4", "2.6", "5" };
			string[] data2 = { "2.4", "3.2", "4.999", "2.456", "3.2", "3.45", "4.56" };
			string name1 = "bob";
			string name2 = "bob's wife";
			string[] scales = { "Satisfaction", "Self-Regulation", "Communication", "Kindness/Caring", "Conflict", "Sexual Intimacy", "Problem Areas" };
			string[] size = { "450", "338" };
			string[] labels = { "label1", "label2", "label3" };

			MemoryStream ms = new MemoryStream(5 * 1024);
			MultiBar test = new MultiBar();
			if (labels.Length >= 3)
				test.cutoffLabels = new ReadOnlyCollection<string>(labels);
			test.GraphTitle = title;
			test.AddBar(name1, data1);
			test.AddBar(name2, data2);
			test.SetYAxis(scales);
			test.Resize(float.Parse(size[0]), float.Parse(size[1]));
			test.Render(ms);
			ms.Position = 0; // reset to start of bitmap
			contentType = "image/jpeg";
			return ms;
		}

		#endregion
	}
}