using System;
using System.Diagnostics;
using System.Xml;
using System.Xml.XPath;
using NAssess.Core.Reporting;
using MetaData;

namespace NAssess.Core.Engines.Reporting
{
	/// <summary>
	/// Summary description for ReportingTools.
	/// </summary>
	public class ReportingTools
	{
		public const string NAssessReportingNamespace = 
			"http://nassess.byu.edu/report";
		public const string NAssessReportingExtensionsNamespace = 
			"http://nassess.byu.edu/report/extensions";
		public const string NAssessReportingStandardExtensionsNamespace = 
			"http://nassess.byu.edu/rendering/report/standardextensions";
		internal const string XHtmlNamespace =
			"http://www.w3.org/1999/xhtml";

		internal static XmlNode GetUserDataXml(Report report)
		{
			if( report == null ) throw new ArgumentNullException("report");

			/* The schema for this data should be:
					 * <participants>
					 *		<participant FriendlyName="Jack">
					 *			<evaluation id="787" questionnaireName="RELATE">
					 *				<responses v1='5' v2='3' />
					 *				<scales KINDNESS='3.5' ... />
					 *			</evaluation>
					 *			<evaluation id="788" questionnaireName="RELATE">
					 *				<responses v10='2' />
					 *			</evaluation>
					 *		</participant>
					 *		<participant FriendlyName="Jill">
					 *			<...>
					 *		</participant>
					 * </participants>
					 */
			XmlDocument xmlUserInfo = new XmlDocument();
			XmlElement nParticipants = (XmlElement) xmlUserInfo.AppendChild(xmlUserInfo.CreateElement("participants"));
			foreach( ReportParticipant participant in report.Participants )
			{
				XmlElement nParticipant = (XmlElement) nParticipants.AppendChild( xmlUserInfo.CreateElement("participant") );
				nParticipant.SetAttribute("FriendlyName", participant.Name);
				if( participant.Gender != Demographics.GenderType.Unknown )
					nParticipant.SetAttribute("Gender", participant.Gender == Demographics.GenderType.Male ? "M" : "F" );
				foreach( Evaluating.Evaluation eval in participant.Evaluations )
				{
					XmlElement nEvaluation = (XmlElement) nParticipant.AppendChild( xmlUserInfo.CreateElement("evaluation") );
					nEvaluation.SetAttribute("id", eval.Id.ToString());
					nEvaluation.SetAttribute("questionnaireName", eval.Questionnaire.Name);
					nEvaluation.AppendChild( eval.Responses.ToXml(nEvaluation.OwnerDocument) );
					nEvaluation.AppendChild( eval.Scales.ToXml(nEvaluation.OwnerDocument) );
				}
			}
			return xmlUserInfo;
		}
		/// <summary>
		/// Returns scale cutoff values for a specific questionnaire and gender.
		/// </summary>
		public static XmlDocument GetScaleCutoffs(Evaluating.Questionnaire questionnaire, Demographics.GenderType gender)
		{
			/* Format of resulting XML
			 * <cutoffs>
			 *		<green kind='3.5' happy='4' ... />
			 *		<red kind='1' happy='2' ... />
			 * </cutoffs>
			 */
			if (questionnaire == null) throw new ArgumentNullException("questionnaire");
			XmlDocument xmlCutOffs = new XmlDocument();
			XmlNode nCutoffs = xmlCutOffs.AppendChild( xmlCutOffs.CreateElement("cutoffs") );
			
			XmlElement nGreen = (XmlElement) nCutoffs.AppendChild( xmlCutOffs.CreateElement("green") );
			XmlElement nRed = (XmlElement) nCutoffs.AppendChild( xmlCutOffs.CreateElement("red") );

			XPathNodeIterator nlScales = questionnaire.Codebook.Select(
				"/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='analysis'][ddi:concept[@vocab]]");
			while( nlScales.MoveNext() )
			{
				XPathNavigator nScale = nlScales.Current;
				addCutoffAttribute( nScale, nGreen, gender, 'G', questionnaire );
				addCutoffAttribute( nScale, nRed, gender, 'R', questionnaire );
			}

			return xmlCutOffs;
		}
		private static void addCutoffAttribute( XPathNavigator nScale, XmlElement nCutoffs, 
			Demographics.GenderType gender, char colorSuffix, Evaluating.Questionnaire questionnaire )
		{
			// Get scale name
			string scaleName = nScale.GetAttribute("ID","");

			// Derive gender character
			string genderChar = Demographics.ConvertGenderTypeToChar(gender).ToString();
			if( genderChar == " " ) genderChar = string.Empty;

			XPathNavigator nGenderSpecific = Evaluating.Ddi.SelectSingleNode("ddi:concept[@vocab='" + genderChar + colorSuffix + "']", nScale);
			XPathNavigator nGenderNeutral = Evaluating.Ddi.SelectSingleNode("ddi:concept[@vocab='" + colorSuffix + "']", nScale);

			if( nGenderSpecific != null ) // If a gender-specific cutoff is available, use it
				nCutoffs.SetAttribute( scaleName, nGenderSpecific.Value );
			else if( nGenderNeutral != null ) // Try for the gender neutral cutoff
				nCutoffs.SetAttribute( scaleName, nGenderNeutral.Value );
		}

		#region Graph generation - EXPERIMENTAL
/*
		public static Graphs.BarGraph GetScaleGraph(Report report, string scaleName, int width, int height)
		{
			if( report == null ) throw new ArgumentNullException("report");
			if( scaleName == null || scaleName.Trim().Length == 0 ) throw new ArgumentNullException("scaleName");
			if( scaleName.IndexOf('\'') >= 0 ) throw new ArgumentException("Cannot have apostrophes in scale name.");

			foreach( Evaluating.Codebook cb in report.Questionnaire.Codebooks )
			{
				XmlNode nScale = cb.xmlDDIDoc.SelectSingleNode(string.Format(
					"/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='analysis'][@ID='{0}']", scaleName), cb.getDDINamespace());
				if( nScale == null ) continue;
				
				Graphs.BarGraph bg = new Graphs.BarGraph();
				bg.Style = getScaleStyle( cb, nScale );
				bg.MaxYScale = getScaleMaxY( cb, nScale );
				if( width > 0 ) bg.Size.Width = width;
				if( height> 0 ) bg.Size.Height = height;
				bg.Title = getScaleTitle( cb, nScale );

				return bg;
			}
			throw new ArgumentOutOfRangeException("scaleName", scaleName, "Could not find a scale by that name in any of the codebooks.");
		}

		private static Graphs.BarGraph.StyleEnum getScaleStyle(Evaluating.Codebook codebook, XmlNode nScale)
		{
			Debug.Assert( codebook != null && nScale != null );

			XmlNode nScaleStyle = nScale.SelectSingleNode("ddi:defntn", codebook.getDDINamespace());
			if( nScaleStyle == null ) return Graphs.BarGraph.StyleEnum.Proportional; // default

			return (Graphs.BarGraph.StyleEnum) Enum.Parse(typeof(Graphs.BarGraph.StyleEnum), nScaleStyle.Value, true);
		}

		private static float getScaleMaxY(Evaluating.Codebook codebook, XmlNode nScale)
		{
			Debug.Assert( codebook != null && nScale != null );

			XmlNode nFirstVar = getFirstVarInVarGrp( codebook, nScale );
			return float.Parse( nFirstVar.SelectSingleNode("ddi:valrng/ddi:range[1]/@max", codebook.getDDINamespace()).Value );
		}

		private static XmlNode getFirstVarInVarGrp( Evaluating.Codebook codebook, XmlNode nVarGrp )
		{
			Debug.Assert( codebook != null && nVarGrp != null );

			XmlAttribute xaVar = nVarGrp.Attributes["var"];
			XmlAttribute xaVarGrp = nVarGrp.Attributes["varGrp"];
			if( xaVar != null )
				return nVarGrp.SelectSingleNode(string.Format(
					"../ddi:var[@ID='{0}']", xaVar.Value.Split(' ')[0]), codebook.getDDINamespace());
			else if( xaVarGrp != null )
				return getFirstVarInVarGrp( codebook, nVarGrp.SelectSingleNode(string.Format(
					"../ddi:varGrp[@ID='{0}']", xaVarGrp.Value.Split(' ')[0]), codebook.getDDINamespace()) );
			else
				throw new ApplicationException("No var or varGrp attributes found.");
		}

		private static string getScaleTitle(Evaluating.Codebook codebook, XmlNode nScale)
		{
			Debug.Assert( codebook != null && nScale != null );

			return nScale.SelectSingleNode("ddi:labl", codebook.getDDINamespace()).Value;
		}
*/
		#endregion
	}
}
