using System;
using System.Diagnostics;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;

using NAssess.Core.Engines;
using NAssess.Core.Reporting;
using NerdBank.Tools;
using NerdBank.Tools.Exslt;

namespace NAssess.Core.Engines.Reporting
{
	/// <summary>
	/// The standard NAssess render for PDF reports.
	/// </summary>
	public class StandardFO : NAssess.Core.Reporting.FORenderingEngine
	{
		protected XmlResolver CustomResolver;
		protected XslCompiledTransform xslt;

		public StandardFO()
		{
		}

		#region Attributes
		private Report _report;
		public override Report Report
		{
			get
			{
				return _report;
			}
			set
			{
				if (_report != null)
				{
					// Detach from old report's Changed event.  
					// This actually does work. See ms-help://MS.VSCC.2003/MS.MSDNQTR.2003FEB.1033/csref/html/vcwlkEventsTutorial.htm
					_report.Changed -= new EventHandler(OnReportConfigurationChanged);
				}

				_report = value;

				if (value != null)
				{
					value.Changed += new EventHandler(OnReportConfigurationChanged);
					InitReport();
				}
			}
		}

		private XmlNode _UserDataXml;
		public XmlNode UserDataXml
		{
			get
			{
				if (_UserDataXml == null)
					_UserDataXml = ReportingTools.GetUserDataXml(Report);
				return _UserDataXml;
			}
		}

		public string ReportTitle;
		public string ReportSubTitle;
		#endregion
		#region Events
		private void OnReportConfigurationChanged(object sender, EventArgs e)
		{
			// invalidate the data arguments we pass into the transform, to force recreation
			XsltArguments = null;
		}
		#endregion
		#region Helper methods
		protected override void InitXsltArguments()
		{
			base.InitXsltArguments();

			// Add extension objects to provide additional functionality
			InitXsltExtensionObjects();
		}

		protected void InitReport()
		{
			if (Report == null) throw new InvalidOperationException(Resources.SetReportFirst);
			_UserDataXml = null; // re-initialize

			ReportTitle = Report.Assessment.Description != null && Report.Assessment.Description.Length > 0
				? Report.Assessment.Description : Report.Assessment.Name;
			ReportSubTitle = null;

			// Prepare custom InProjectResolver, so XSLT files can be retrieved directly from the assembly
			CustomResolver = new ProjectXmlResolver(Project.DefaultNamespace);

			InitXslt();
		}
		private const string XsltStartPoint = "/Engines/Reporting/xslt.fo/main.xslt";
		private const string XsltCacheName = "Xslt: " + XsltStartPoint;
		protected void InitXslt()
		{
			// The XSLT that produces the XSL-FO that we output
			// First try to pull it out of the cache
			xslt = Project.GetFromCacheIfAvailable<XslCompiledTransform>(XsltCacheName);
			if( xslt == null )
			{
				xslt = new XslCompiledTransform();
				XsltSettings settings = new XsltSettings(true, true);
				using( Stream mainXslt = EmbeddedResources.GetFileStreamFromAssembly(XsltStartPoint, Project.DefaultNamespace) )
					xslt.Load( new XmlTextReader( mainXslt ), settings, CustomResolver );
				Project.SaveInCacheIfAvailable<XslCompiledTransform>(XsltCacheName, xslt);
			}
		}
		protected void InitXsltExtensionObjects()
		{
			SetXsltExtensionObject("http://fhss.byu.edu/exslt/strings", new ExsltStrings());
			SetXsltExtensionObject(CultureNamespace, new CultureAssist(typeof(StandardHtml)));
			SetXsltExtensionObject(ReportingTools.NAssessReportingStandardExtensionsNamespace, new XsltStdExtensions());
			SetXsltExtensionObject(ReportingTools.NAssessReportingExtensionsNamespace, Plugins);
		}
		public void InitXsltParams(bool monochrome)
		{
			string suffixMonochrome = "";
			if (monochrome) suffixMonochrome = "monochrome=true&";

			if (WebDirectUrl != null)
			{
				SetXsltParam(StandardHtml.GradientGraphParamName, string.Empty, WebDirectUrl + StandardHtml.StdGradientGraphSource + suffixMonochrome);
				SetXsltParam(StandardHtml.MultiBarGraphParamName, string.Empty, WebDirectUrl + StandardHtml.StdMultiBarGraphSource + suffixMonochrome);
				SetXsltParam(StandardHtml.BarGraphParamName, string.Empty, WebDirectUrl + StandardHtml.StdBarGraphSource + suffixMonochrome);
				SetXsltParam(StandardHtml.BarGraphLabelsParamName, string.Empty, WebDirectUrl + StandardHtml.StdBarGraphLabelsSource + suffixMonochrome);
				SetXsltParam(StandardHtml.BarGraphLegendParamName, string.Empty, WebDirectUrl + StandardHtml.StdBarGraphLegendSource + suffixMonochrome);
				SetXsltParam(StandardHtml.BullsEyeGraphParamName, string.Empty, WebDirectUrl + StandardHtml.StdBullsEyeGraphSource + suffixMonochrome);
				SetXsltParam(StandardHtml.BullsEyeLegendParamName, string.Empty, WebDirectUrl + StandardHtml.StdBullsEyeLegendSource + suffixMonochrome);
				SetXsltParam(StandardHtml.ChecklistIconParamName, string.Empty, WebDirectUrl + StandardHtml.StdChecklistIconSource);
			}

			// PDF specific arguments
			if (ReportTitle != null) SetXsltParam("ReportTitle", string.Empty, ReportTitle);
			if (ReportSubTitle != null) SetXsltParam("ReportSubTitle", string.Empty, ReportSubTitle);
			SetXsltParam("TodaysDate", string.Empty, DateTime.Now.ToString("MMMM d, yyyy"));
			if (LogoImageUrl != null)
			{
				// add cultural specific directory to seal URL if not fully qualified.
				Uri fullyqualifiedSealUrl = new Uri(ImagesWithTextDirectory, LogoImageUrl);
				SetXsltParam("SealUrl", string.Empty, fullyqualifiedSealUrl.AbsoluteUri);
			}
		}
		#endregion

		#region Operations
		public override void ToFO(Stream outStream, bool Monochrome, InlineFormatter inlineFOFormatter)
		{
			Debug.Assert(Report.Assessment.Questionnaire != null);
			if (Report == null) throw new InvalidOperationException(Resources.SetReportFirst);
			if (outStream == null) throw new ArgumentNullException("outStream");
			if (!outStream.CanWrite) throw new ArgumentException(Resources.StreamUnwritable, "outStream");

			// Add other arguments for rendering, based on preset properties
			InitXsltParams(Monochrome);

			// The user(s)' information, including responses and scales, act as an Xslt argument
			SetXsltParam("participants", string.Empty, UserDataXml.CreateNavigator().Select("/participants/*"));

			// The codebook data, so labels and text can be used directly out of the questionnaire
			SetXsltParam("codebook", string.Empty, Report.Assessment.Questionnaire.Codebook.XPath.CreateNavigator().Select("*"));

			// Inform all plug-ins that rendering is about to begin.
			OnRendering(null);

			// Transform report XML to FO
			using (MemoryStream ms = new MemoryStream(50 * 1024))
			{
				XmlTextWriter xtw = new XmlTextWriter(ms, Encoding.Unicode);
				xslt.Transform(new XmlTextReader(new StringReader(Report.Assessment.Template.XmlString)), 
					XsltArguments, xtw);
				xtw.Flush();
				ms.Position = 0;

				// Transform XSL-FO to PDF now, if necessary.
				try
				{
					if (inlineFOFormatter != null)
						inlineFOFormatter(ms);
				}
				catch (Exception ex)
				{
					// Likely the cause of this is invalid FO being passed into the PDF renderer
					throw new FORenderingException("Error during FO formatting callback.", ex);
				}
				finally
				{
					// Inform all plug-ins that rendering has completed.
					OnRendered(null);
				}
			}
		}

		#endregion
	}
}
