using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using System.Xml.XPath;
using System.Diagnostics;

using NerdBank.Tools;
using NerdBank.Tools.Exslt;
using NAssess.Core.Reporting;

namespace NAssess.Core.Engines.Reporting
{
	/// <summary>
	/// The standard NAssess render for HTML reports.
	/// </summary>
	public class StandardHtml : NAssess.Core.Reporting.HtmlReportRenderingEngine
	{
		protected XmlResolver CustomResolver;
		protected XslCompiledTransform Xslt;

		public StandardHtml()
		{
		}

		#region Attributes
		protected internal const string AssessmentNameParamName = "AssessmentName";

		private Report report;
		public override Report Report
		{
			get
			{
				return report;
			}
			set
			{
				if (report != null)
				{
					// Detach from old report's Changed event.  
					// This actually does work. See ms-help://MS.VSCC.2003/MS.MSDNQTR.2003FEB.1033/csref/html/vcwlkEventsTutorial.htm
					report.Changed -= new EventHandler(onReportConfigurationChanged);
				}

				report = value;

				if (value != null)
				{
					value.Changed += new EventHandler(onReportConfigurationChanged);
					InitReport();
				}
			}
		}

		public override int PagesCount
		{
			get
			{
				if (Report == null)
					throw new InvalidOperationException(Resources.SetReportFirst);

				// We consider a page for every first-level element.
				return Report.Assessment.Template.Select("/nassess:report/nassess:level").Count;
			}
		}
		private XmlNode userDataXml;
		public XmlNode UserDataXml
		{
			get
			{
				if (userDataXml == null)
					userDataXml = ReportingTools.GetUserDataXml(Report);
				return userDataXml;
			}
		}

		// The following few properties control where the resulting report HTML points to download bar graphs and other
		// NAssess generated images
		public const string StdBarGraphSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.BarGraph/Render&";
		private string _BarGraphSource = StdBarGraphSource;
		public string BarGraphSource
		{
			get
			{
				return _BarGraphSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("BarGraphSource");

				_BarGraphSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string BarGraphParamName = "graphURL";

		public const string StdGradientGraphSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.GradientGraph/Render&";
		private string _GradientGraphSource = StdGradientGraphSource;
		public string GradientGraphSource
		{
			get
			{
				return _GradientGraphSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("GradientGraphSource");

				_GradientGraphSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string GradientGraphParamName = "gGraphURL";

		public const string StdMultiBarGraphSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.MultiBar/Render&";
		private string _MultiBarGraphSource = StdMultiBarGraphSource;
		public string MultiBarGraphSource
		{
			get
			{
				return _MultiBarGraphSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("MultiBarGraphSource");

				_MultiBarGraphSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string MultiBarGraphParamName = "multibarURL";


		public const string StdBarGraphLegendSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.Legend/Render&";
		private string _BarGraphLegendSource = StdBarGraphLegendSource;
		public string BarGraphLegendSource
		{
			get
			{
				return _BarGraphLegendSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("BarGraphLegendSource");

				_BarGraphLegendSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string BarGraphLegendParamName = "graphLegendURL";

		public const string StdBarGraphLabelsSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.GraphLabels/Render&";
		private string _BarGraphLabelsSource = StdBarGraphLabelsSource;
		public string BarGraphLabelsSource
		{
			get
			{
				return _BarGraphLabelsSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("BarGraphLabelsSource");

				_BarGraphLabelsSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string BarGraphLabelsParamName = "graphLabelsURL";

		public const string StdBullsEyeGraphSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.BullsEyeGraph/Render&";
		private string _BullsEyeGraphSource = StdBullsEyeGraphSource;
		public string BullsEyeGraphSource
		{
			get
			{
				return _BullsEyeGraphSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("BullsEyeGraphSource");

				_BullsEyeGraphSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string BullsEyeGraphParamName = "bullsEyeGraphURL";

		public const string StdBullsEyeLegendSource = "?NAssess/NAssess.Core.Engines.Reporting.Graphs.BullsEyeLegend/Render&";
		private string _BullsEyeLegendSource = StdBullsEyeLegendSource;
		public string BullsEyeLegendSource
		{
			get
			{
				return _BullsEyeLegendSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("BullsEyeLegendSource");

				_BullsEyeLegendSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string BullsEyeLegendParamName = "bullsEyeLegendURL";

		public const string StdChecklistIconSource = "?NAssess/NAssess.Core.Engines.Reporting.StandardHtml/GetImage&filename=BlueSquare.png";
		private string _ChecklistIconSource = StdChecklistIconSource;
		public string ChecklistIconSource
		{
			get
			{
				return _ChecklistIconSource;
			}
			set
			{
				if (value == null || value.Length == 0)
					throw new ArgumentNullException("ChecklistIconSource");

				_ChecklistIconSource = value;

				onReportConfigurationChanged(this, null);
			}
		}
		public const string ChecklistIconParamName = "checkImageURL";

		// The following properties describe the link to use for HTML linking within the report, but on different pages.
		// Setting this property is optional, but without it set, no links across pages will be enabled.
		private Uri linkWithinReportUrl;
		public Uri LinkWithinReportUrl
		{
			get
			{
				return linkWithinReportUrl;
			}
			set
			{
				linkWithinReportUrl = value;

				onReportConfigurationChanged(this, null);
			}
		}

		// The following map is for mapping report FriendlyName references within the current
		// report to actual report displaying web pages for hyperlinking between reports.
		private Dictionary<Assessment, Uri> externalReportUrls = new Dictionary<Assessment, Uri>(3);
		public Dictionary<Assessment, Uri> ExternalReportUrls
		{
			get { return externalReportUrls; }
			private set { externalReportUrls = value; }
		}
		#endregion
		#region Events
		private void onReportConfigurationChanged(object sender, EventArgs e)
		{
			// invalidate the data arguments we pass into the transform, to force recreation
			userDataXml = null;
			XsltArguments = null;
		}
		#endregion
		#region Operations
		/// <summary>
		/// Gets the number of pages that are in the report.  
		/// </summary>
		public override void ToHtml(int page, XmlWriter writer)
		{
			if (Report == null) throw new InvalidOperationException(Resources.SetReportFirst);
			if (page < 0 || page > PagesCount) throw new ArgumentOutOfRangeException("page",
												   page, Resources.PageOutOfRange);

			// Translate the page # to render to the top-level to render
			XPathNavigator nTopLevel = Report.Assessment.Template.SelectSingleNode("/nassess:report/nassess:level[" + page.ToString() + "]");
			Debug.Assert(nTopLevel != null, "Somehow the page # did not map to a page in the report.");

			// Add argument specifying the top-level to render
			SetXsltParam("TopLevelId", string.Empty, nTopLevel.GetAttribute("id", ""));

			// The user(s)' information, including responses and scales, act as an Xslt argument
			SetXsltParam("participants", string.Empty, UserDataXml.CreateNavigator().Select("/participants/*"));

			// The codebook data, so labels and text can be used directly out of the questionnaire
			SetXsltParam("codebook", string.Empty, Report.Assessment.Questionnaire.Codebook.XPath.CreateNavigator().Select("*"));

			// Perform render and output to stream
			OnRendering(null);
			Xslt.Transform(Report.Assessment.Template.XPath.CreateNavigator(), XsltArguments, writer);
			OnRendered(null);
		}
		public IList<HeadingNode> GetHeadings()
		{
			XPathNodeIterator iterator = Report.Assessment.Template.Select("/nassess:report/nassess:level//xhtml:h1/text()");
			List<HeadingNode> headings = new List<HeadingNode>();
			while (iterator.MoveNext())
			{
				XPathNodeIterator subiterator = ReportTemplate.Select("nassess:level//xhtml:h2/text()", iterator.Current);
				List<HeadingNode> subheadings = new List<HeadingNode>();
				while (subiterator.MoveNext())
					subheadings.Add(new HeadingNode(subiterator.Current.Value, null));
				headings.Add(new HeadingNode(iterator.Current.Value, subheadings));
			}
			return headings;
		}
		public class HeadingNode
		{
			internal HeadingNode(string title, IList<HeadingNode> childNodes)
			{
				this.title = title;
				this.childNodes = childNodes != null ? childNodes : new List<HeadingNode>();
			}
			private string title;
			public string Title { get { return title; } }
			private IList<HeadingNode> childNodes;
			public IList<HeadingNode> ChildNodes { get { return childNodes; } }
		}
		#endregion
		#region Helper methods
		protected void InitReport()
		{
			if (Report == null) throw new InvalidOperationException(Resources.SetReportFirst);
			userDataXml = null; // re-initialize

			// Prepare custom InProjectResolver, so XSLT files can be retrieved directly from the assembly
			CustomResolver = new ProjectXmlResolver(Project.DefaultNamespace);

			InitXslt();
		}
		protected override void InitXsltArguments()
		{
			base.InitXsltArguments();

			// Add arguments for rendering, based on preset properties
			initXsltParams();

			// Add extension objects to provide additional functionality
			initXsltExtensionObjects();
		}
		private void initXsltExtensionObjects()
		{
			XsltArguments.AddExtensionObject("http://fhss.byu.edu/exslt/strings", new ExsltStrings());
			XsltArguments.AddExtensionObject(CultureNamespace, new CultureAssist(this));
			XsltArguments.AddExtensionObject(ReportingTools.NAssessReportingStandardExtensionsNamespace, new XsltStdExtensions());
			XsltArguments.AddExtensionObject(ReportingTools.NAssessReportingExtensionsNamespace, Plugins);
		}
		private void initXsltParams()
		{
			Debug.Assert(Report != null, "Report must be set first.");
			SetXsltParam(AssessmentNameParamName, string.Empty, Report.Assessment.Name);
			if (WebDirectUrl != null)
			{
				SetXsltParam(MultiBarGraphParamName, string.Empty, WebDirectUrl + MultiBarGraphSource);
				SetXsltParam(GradientGraphParamName, string.Empty, WebDirectUrl + GradientGraphSource);
				SetXsltParam(BarGraphParamName, string.Empty, WebDirectUrl + BarGraphSource);
				SetXsltParam(BarGraphLabelsParamName, string.Empty, WebDirectUrl + BarGraphLabelsSource);
				SetXsltParam(BarGraphLegendParamName, string.Empty, WebDirectUrl + BarGraphLegendSource);
				SetXsltParam(BullsEyeGraphParamName, string.Empty, WebDirectUrl + BullsEyeGraphSource);
				SetXsltParam(BullsEyeLegendParamName, string.Empty, WebDirectUrl + BullsEyeLegendSource);
				SetXsltParam(ChecklistIconParamName, string.Empty, WebDirectUrl + ChecklistIconSource);
			}
			SetXsltParam("imagesURL", string.Empty, ImagesDirectory);
			SetXsltParam("imagesWithTextURL", string.Empty, ImagesWithTextDirectory);
			SetXsltParam("linkWithinReportURL", string.Empty, LinkWithinReportUrl != null ? LinkWithinReportUrl.ToString() : null);

			if (ExternalReportUrls.Count > 0)
			{
				XmlDocument listOfExternalReports = new XmlDocument();
				XmlElement nRoot = (XmlElement)listOfExternalReports.AppendChild(listOfExternalReports.CreateElement("externalReports"));
				foreach (Assessment assessment in ExternalReportUrls.Keys)
				{
					// Don't allow adding an external report reference to ourselves
					if (assessment == Report.Assessment) continue; // skip it
					if (ExternalReportUrls[assessment] == null) continue; // protect from null values

					XmlElement nExtReport = listOfExternalReports.CreateElement("report");
					nExtReport.SetAttribute("FriendlyName", assessment.Name);
					nExtReport.SetAttribute("URL", ExternalReportUrls[assessment].ToString());

					// Go fetch the actual report XML so that this report can calculate page numbers with it and such
					nExtReport.InnerXml = assessment.Template.Xml.OuterXml;

					nRoot.AppendChild(nExtReport);
				}
				// Set parameter to node set of <report> tags (no root <externalReports> tag)
				SetXsltParam("linkExternalReportURLs", string.Empty, nRoot.CreateNavigator().Select("*"));
			}
			else
			{
				XsltArguments.RemoveParam("linkExternalReportURLs", string.Empty);
			}
		}
		private const string XsltStartPoint = "/Engines/Reporting/xslt.html/main.xslt";
		private const string XsltCacheName = "Xslt: " + XsltStartPoint;
		protected void InitXslt()
		{
			// The XSLT that produces the Html that we output
			// First try to pull it out of the cache
			Xslt = Project.GetFromCacheIfAvailable<XslCompiledTransform>(XsltCacheName);
			if (Xslt == null)
			{
				Xslt = new XslCompiledTransform();
				XsltSettings settings = new XsltSettings(true, true);
				using (Stream mainXslt = EmbeddedResources.GetFileStreamFromAssembly(XsltStartPoint, Project.DefaultNamespace))
					Xslt.Load(new XmlTextReader(mainXslt), settings, CustomResolver);
				Project.SaveInCacheIfAvailable(XsltCacheName, Xslt);
			}
		}
		#endregion
		#region Methods for obtaining embedded data
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1021:AvoidOutParameters"), WebDirect]
		public static Stream GetLegendPopupJS(out string contentType)
		{
			Stream stream = EmbeddedResources.GetFileStreamFromAssembly("/Engines/Reporting/xslt.html/legendboxscript.js", Project.DefaultNamespace);
			contentType = "text/javascript";
			return stream;
		}
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1021:AvoidOutParameters"), WebDirect]
		public static Stream GetImage(string fileName, out string contentType)
		{
			if (fileName == null) throw new ArgumentNullException("filename");
			// security check
			// TODO: Fix this to check for allowed sequences rather than disallowed ones.  
			//       (Escaped characters could still get through this.)
			if (fileName.IndexOf(Path.DirectorySeparatorChar) >= 0 || 
				fileName.IndexOf(Path.AltDirectorySeparatorChar) >= 0)
				throw new ArgumentException(Resources.InvalidFileName, "fileName");

			Stream stream = EmbeddedResources.GetFileStreamFromAssembly("/Engines/Reporting/images/" + fileName, Project.DefaultNamespace);
			contentType = Project.GetMimeType(fileName);
			return stream;
		}
		#endregion
	}
}
