using System;
using System.Diagnostics;
using System.IO;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;
using System.Data;

using NerdBank.Tools;

namespace NAssess.Core.Evaluating
{
	public class Ddi
	{
		/// <summary>
		/// The URI for the DDI codebook XML schema.
		/// </summary>
		public const string DdiUri = "http://www.icpsr.umich.edu/DDI";

		private Ddi() { } // NHibernate support
		public Ddi(XmlDocument xml)
		{
			Xml = xml;
		}

		private int id; // just for NHibernate
		private int Id
		{
			get { return id; }
			set { id = value; }
		}

		private string xmlString;
		/// <remarks>
		/// For NHibernate support.
		/// </remarks>
		public string XmlString
		{
			get
			{
				if (xmlString == null && xml != null)
					xmlString = xml.OuterXml;
				return xmlString;
			}
			set
			{
				xmlString = value;
				xml = null;
				xpath = null;
			}
		}

		private XmlDocument xml;
		public XmlDocument Xml
		{
			get
			{
				if (xml == null && xmlString != null)
				{
					xml = new XmlDocument();
					xml.LoadXml(xmlString);
				}
				return xml;
			}
			private set
			{
				ValidateDdi(value);
				xml = value;
				xmlString = null;
				xpath = null;
			}
		}

		private XPathDocument xpath;
		public XPathDocument XPath
		{
			get
			{
				if (xpath == null && XmlString != null)
					xpath = new XPathDocument(new StringReader(XmlString));
				return xpath;
			}
		}

		public static implicit operator XmlDocument(Ddi ddi)
		{
			if (ddi == null) return null;
			return ddi.Xml;
		}

		/// <summary>
		/// Converts a user input value for a specific variable to the type
		/// appropriate to be stored in that variable.
		/// This may mean from a string to an integer, or from a float 
		/// to a rounded number.
		/// </summary>
		/// <param name="variableName">
		/// The variable in the codebook whose data type should determine
		/// the ending type of the value.
		/// </param>
		/// <param name="value">
		/// The value that is about to be assigned to the specified variable.
		/// </param>
		/// <returns>
		/// The converted value.
		/// </returns>
		public virtual object PrepareDataValue(string variableName, object value)
		{
			string valueString = value as string;
			if (valueString != null && valueString.Length == 0) value = valueString = null;
			if (value == null) return value; // nothing more to do
			XPathNavigator var = FindVariableByName(variableName);
			Type desiredType = GetDataTypeFromDdiVar(var);
			try
			{
				value = Convert.ChangeType(value, desiredType);
				// trim string length if too long
				if (desiredType == typeof(string) && value != null)
				{
					XPathNavigator width = SelectSingleNode("ddi:location/@width", var);
					if (width != null && valueString.Length > width.ValueAsInt)
						value = valueString.Substring(0, width.ValueAsInt);
				}
			}
			catch
			{
				// Try to round floating points to integers, if applicable
				if (desiredType == typeof(int))
				{
					try
					{
						value = (int)Math.Round(Convert.ToSingle(value), 0);
					}
					catch
					{
						value = null; // failed to convert.  Just ignore.
					}
				}
				else
					value = null; // failed to convert.  Just ignore.
			}
			return value;
		}

		#region DDI schema validation
		private System.Text.StringBuilder ddiValidationErrors;
		protected virtual void ValidateDdi(XmlDocument ddi)
		{
			// Validate the incoming report XML to make sure it is fully
			// compliant with workable code.
			XmlReaderSettings settings = new XmlReaderSettings();

			// Prepare to validate incoming DDI and other filtering schemas
			settings.Schemas.Add(DdiUri,
				new XmlTextReader(EmbeddedResources.GetFileStreamFromAssembly(
					"/schema/ddi2.xsd", Project.DefaultNamespace)));
			settings.Schemas.Add("http://nassess.byu.edu/conditional",
				new XmlTextReader(EmbeddedResources.GetFileStreamFromAssembly(
					"/schema/conditional.xsd", Project.DefaultNamespace)));
			settings.ValidationType = ValidationType.Schema;
			settings.ValidationEventHandler += new ValidationEventHandler(ddiReader_ValidationEventHandler);

			using (XmlReader reader = XmlReader.Create(new StringReader(ddi.OuterXml), settings))
			{
				XmlDocument newDDI = new XmlDocument();
				newDDI.PreserveWhitespace = true;
				ddiValidationErrors = new StringBuilder();
				newDDI.Load(reader);
				if (ddiValidationErrors.Length > 0)
				{
					// There were schema validation errors.
					throw new ArgumentException(string.Format(Resources.DdiValidationErrorSummary,
						ddiValidationErrors));
				}

				SuperValidateNewDdi(newDDI);
			}
		}
		protected virtual void SuperValidateNewDdi(XmlDocument newDdi)
		{
			// TODO: fix this
			//CheckNoVarNameConflict( newDDI, Table );

			CheckForDisplayVarGrp(newDdi);
			CheckSectionVarGrpsForNoVars(newDdi);
			CheckValues(newDdi);

			// Check to make sure that all scales are made up of variables with the same
			// range of possible values (so we don't get a scale of v1, v2, v3 where v2 is 1-10
			// while v1 and v3 are 1-5, because that wouldn't make sense).
			// TODO: code here

			// Check all scales that have reversed variables to see that the variables that 
			// appear in the reversed variable list also appear in the regular variable list.
			// TODO: code here

			// Check that stacked grids have sub varGrps that have equal numbers of variables in them.
			// TODO: code here

			// TODO: More checks here
		}
		/// <summary>
		/// Make sure that all variable names are unique within the codebook.
		/// </summary>
		internal virtual void CheckNoVarNameConflict(XmlDocument ddi)
		{
			XmlNamespaceManager nsmgr = AddNamespaces(ddi);
			XmlNodeList vars = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var[@name]", nsmgr);
			// Check that all our own variable names are unique
			foreach (XmlNode var in vars)
			{
				if (ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var[@name='" + var.Attributes["name"].Value + "']", nsmgr).Count > 1)
					throw new ArgumentException(var.Attributes["name"].Value,
						"Variable name is not unique within codebook.  No variable name can be used twice.");
			}
		}
		/// <summary>
		/// Check to make sure that there is exactly one display varGrp, and that
		/// there is something in it.
		/// </summary>
		/// <param name="ddi"></param>
		protected virtual void CheckForDisplayVarGrp(XmlDocument ddi)
		{
			Debug.Assert(ddi != null);
			XmlNamespaceManager nsmgr = AddNamespaces(ddi);
			XmlNodeList varGrps = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='display']", nsmgr);
			if (varGrps.Count != 1)
				throw new ArgumentException(string.Format(Resources.ExactlyOneDisplayVarGroup, varGrps.Count));

			CheckForSectionVarGrpsInDisplayVarGrp(ddi, varGrps[0]);
		}
		protected virtual void CheckForSectionVarGrpsInDisplayVarGrp(XmlDocument ddi, XmlNode displayNode)
		{
			Debug.Assert(ddi != null && displayNode != null);
			XmlNamespaceManager nsmgr = AddNamespaces(ddi);

			// Without section varGrp's within the display varGrp, no pages will ever be rendered.
			if (displayNode.Attributes["varGrp"] == null)
				throw new ArgumentException(Resources.DisplayVarGrpRequiresVarGrpAttribute);
			string[] subVarGrpIDs = displayNode.Attributes["varGrp"].Value.Trim().Split(' ');
			foreach (string subVarGrpID in subVarGrpIDs)
			{
				XmlNode nSubVarGrp = ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@ID='" + subVarGrpID + "']", nsmgr);
				if (nSubVarGrp.Attributes["type"].Value != "section")
					throw new ArgumentException(string.Format(Resources.DdiValidationNonSectionDisplayVarGrp, subVarGrpID));
			}
		}
		protected virtual void CheckSectionVarGrpsForNoVars(XmlDocument ddi)
		{
			Debug.Assert(ddi != null);
			XmlNamespaceManager nsmgr = AddNamespaces(ddi);

			XmlNodeList nl = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='section'][@var]", nsmgr);
			if (nl.Count == 0) return; // no problems detected

			StringBuilder errors = new StringBuilder();
			errors.Append("The following varGrp type='section' nodes were found containing var attributes, which is not supported:\n");
			foreach (XmlNode n in nl)
				errors.Append(n.Attributes["ID"].Value + "\n");

			throw new ArgumentException(errors.ToString());
		}
		/// <summary>
		/// Check that all variable's values are explicitly specified
		/// </summary>
		/// <param name="ddi"></param>
		protected virtual void CheckValues(XmlDocument ddi)
		{
			Debug.Assert(ddi != null);
			XmlNamespaceManager nsmgr = AddNamespaces(ddi);

			XmlNodeList nl = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var[ddi:catgry/ddi:catValu/text() = '']", nsmgr);
			if (nl.Count == 0) return; // no problems detected

			StringBuilder errors = new StringBuilder();
			errors.Append("The following variables were found with a blank value:\n");
			foreach (XmlNode n in nl)
				errors.Append(n.Attributes["ID"].Value + "\n");

			throw new ArgumentException(errors.ToString());
		}
		private void ddiReader_ValidationEventHandler(object sender, ValidationEventArgs e)
		{
			ddiValidationErrors.Append(e.Message + "\n");
		}
		private void verifySimilarFormDDI(string masterDDI, string candidateDDI)
		{
			XmlDocument mDDI = new XmlDocument();
			XmlDocument cDDI = new XmlDocument();

			mDDI.LoadXml(masterDDI);
			cDDI.LoadXml(candidateDDI);

			XmlNode n1 = mDDI.SelectSingleNode("/ddi:codeBook/ddi:dataDscr", AddNamespaces(mDDI));
			XmlNode n2 = cDDI.SelectSingleNode("/ddi:codeBook/ddi:dataDscr", AddNamespaces(cDDI));
			if (!isNodeSimilarInForm(n1, n2))
				throw new ArgumentException(Resources.DdiValidationNonmatchingStructure);
		}
		private bool isNodeSimilarInForm(XmlNode n1, XmlNode n2)
		{
			Debug.Assert(n1 != null); Debug.Assert(n2 != null);
			if (n1.Attributes == null && n2.Attributes == null) return true; // we don't consider text
			if ((n1.Attributes == null) != (n2.Attributes == null)) return false;

			if (n1.ChildNodes.Count != n2.ChildNodes.Count) return false;
			if (n1.Attributes.Count != n2.Attributes.Count) return false;

			// Compare node attributes
			for (int i = 0; i < n1.Attributes.Count; i++)
				if (n1.Attributes[i].Name != n2.Attributes[i].Name) return false;

			// Compare child nodes
			for (int i = 0; i < n1.ChildNodes.Count; i++)
			{
				if (!isNodeSimilarInForm(n1.ChildNodes[i], n2.ChildNodes[i])) return false;
			}

			return true; // similar in form
		}
		#endregion
		#region DDI conversion operations
		/// <summary>
		/// Adds the required basic varGrp definitions to a DDI file.
		/// </summary>
		/// <param name="ddi">
		/// The DDI file with variables and no variable groups.
		/// </param>
		/// <param name="nsmgr">
		/// Optional.  Must contain the uri for DDI, with prefix ddi.
		/// </param>
		public static void PrepareEssentialVarGrps(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) nsmgr = AddNamespaces(ddi);

			XmlElement nData = ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr", nsmgr) as XmlElement;
			if (nData == null)
				throw new ApplicationException(string.Format(Resources.DdiValidationMissingSection, "/ddi:codeBook/ddi:dataDscr"));

			XmlElement nDisplayVarGrp = ddi.CreateElement("ddi", "varGrp", DdiUri);
			XmlElement nPageVarGrp = ddi.CreateElement("ddi", "varGrp", DdiUri);
			XmlElement nSubjectVarGrp = CreateVarGrpWithAllVariables(ddi, nsmgr);

			nPageVarGrp.SetAttribute("type", "section");
			nPageVarGrp.SetAttribute("ID", "Page1");
			nPageVarGrp.SetAttribute("varGrp", nSubjectVarGrp.GetAttribute("ID"));
			SetLabelOfVarOrVarGrp(nPageVarGrp, nsmgr, "Page 1");

			nDisplayVarGrp.SetAttribute("type", "display");
			nDisplayVarGrp.SetAttribute("varGrp", nPageVarGrp.GetAttribute("ID"));
			SetLabelOfVarOrVarGrp(nDisplayVarGrp, nsmgr, "Display order");

			XmlNode nFirstDataNode = nData.FirstChild;
			nData.InsertBefore(nDisplayVarGrp, nFirstDataNode);
			nData.InsertBefore(nPageVarGrp, nFirstDataNode);
			nData.InsertBefore(nSubjectVarGrp, nFirstDataNode);
		}

		/// <summary>
		/// Finds groups of multiple-choice variables with common response options,
		/// and creates variable groups to put them into grids.
		/// </summary>
		/// <returns>
		/// The number of groups created.
		/// </returns>
		public static int GroupSimilarVariables(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			XmlNodeList nlVars = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var", nsmgr);
			StringBuilder varList = new StringBuilder(150);
			XmlElement nLastVar = null;
			XmlNodeList nlLastVarCatgrys = null;
			int groupCount = 0;
			foreach (XmlElement nVar in nlVars)
			{
				XmlNodeList nlCatgrys = nVar.SelectNodes("ddi:catgry", nsmgr);

				bool sameAsLast = true; // assume same until we find a difference.
				if (nLastVar != null)
				{
					// Make sure the variable types are the same.
					sameAsLast &= nLastVar.SelectSingleNode("ddi:varFormat/@type", nsmgr).Value
						== nVar.SelectSingleNode("ddi:varFormat/@type", nsmgr).Value;

					// Compare this variable's multiple response options to the last
					sameAsLast &= nlLastVarCatgrys.Count == nlCatgrys.Count;
					for (int i = 0; sameAsLast && i < nlCatgrys.Count; i++)
					{
						sameAsLast &= nlCatgrys[i].SelectSingleNode("ddi:catValu", nsmgr).InnerText.Trim()
							== nlLastVarCatgrys[i].SelectSingleNode("ddi:catValu", nsmgr).InnerText.Trim();
						sameAsLast &= nlCatgrys[i].SelectSingleNode("ddi:labl", nsmgr).InnerText.Trim()
							== nlLastVarCatgrys[i].SelectSingleNode("ddi:labl", nsmgr).InnerText.Trim();
					}
				}

				// If this variable does not belong to the same group,
				// wrap up the last one and start a new one.
				if (!sameAsLast)
				{
					Debug.Assert(varList.Length > 0); // there must have been a previous variable in the list
					varList.Length -= 1; // remove trailing space

					// Only define a group if there is at least two similar variables together.
					if (varList.ToString().IndexOf(' ') >= 0)
						DefineSimilarVarGrp(ddi, nsmgr, varList.ToString(), nLastVar, ++groupCount);

					varList.Length = 0;
					nLastVar = null;
				}

				// If this is the first variable, start the group with it.
				// If this is the same as the previous variable, continue the group.
				if (nLastVar == null || sameAsLast)
					varList.Append(nVar.GetAttribute("ID") + " ");

				nLastVar = nVar;
				nlLastVarCatgrys = nlCatgrys;
			}

			if (varList.Length == 0) return groupCount; // no variables
			varList.Length -= 1; // remove trailing space
			// Define any last remaining variables as a group
			DefineSimilarVarGrp(ddi, nsmgr, varList.ToString(), nLastVar, ++groupCount);
			return groupCount;
		}
		protected static void DefineSimilarVarGrp(XmlDocument ddi, XmlNamespaceManager nsmgr, string vars, XmlElement lastVarNode, int suffix)
		{
			if (vars == null) throw new ArgumentNullException("vars");
			if (lastVarNode == null) throw new ArgumentNullException("lastVarNode");
			string type = "subject";

			// If there were at least two variables in the last group, 
			// and if they were multiple choice numerics, make this a grid.
			if (vars.IndexOf(' ') >= 0
				&& lastVarNode.SelectSingleNode("ddi:varFormat/@type", nsmgr).Value == "numeric"
				&& lastVarNode.SelectNodes("ddi:catgry", nsmgr).Count > 0)
				type = "grid";

			DefineVariableGroup(ddi, nsmgr, type + suffix.ToString(), type,
				vars, type + " " + suffix.ToString(), null);
		}
		public static int GroupUngroupedVariables(XmlDocument ddi, XmlNamespaceManager nsmgr, int startingGroupOffset)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			Dictionary<string, XmlElement> varsInGrps = new Dictionary<string, XmlElement>(ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr", nsmgr).ChildNodes.Count);

			// Iterate through each varGrp that contains variables, and add each referenced variable ID
			// to a common list.
			foreach (XmlElement nVarGrp in ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@var]", nsmgr))
			{
				foreach (string sVarName in StringUtilities.Tokenize(nVarGrp.GetAttribute("var")))
					varsInGrps.Add(sVarName, nVarGrp);
			}

			// Now go through each variable in file order and see if it is in any variable groups.
			StringBuilder varList = new StringBuilder(300);
			int groupCount = startingGroupOffset;
			foreach (XmlElement nVar in ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var", nsmgr))
			{
				if (!varsInGrps.ContainsKey(nVar.GetAttribute("ID")))
					varList.Append(nVar.GetAttribute("ID") + " ");
				else
				{
					XmlElement inVarGrp = varsInGrps[nVar.GetAttribute("ID")];
					// We ran into a variable in a varGrp already.
					if (varList.Length == 0) continue; // no stray variables were found
					varList.Length -= 1; // trim last space
					DefineDissimilarVarGrp(ddi, nsmgr, varList.ToString(), inVarGrp, ++groupCount);
					varList.Length = 0; // start over
				}
			}
			return groupCount;
		}
		protected static void DefineDissimilarVarGrp(XmlDocument ddi, XmlNamespaceManager nsmgr, string vars, XmlElement beforeNode, int suffix)
		{
			string type = "subject";

			DefineVariableGroup(ddi, nsmgr, type + suffix.ToString(), type,
				vars, type + " " + suffix.ToString(), beforeNode);
		}
		/// <summary>
		/// Corrects many mistakes in variable definitions common when converting
		/// SPSS codebooks to DDI.
		/// </summary>
		/// <param name="ddi">
		/// The DDI to correct variables within.
		/// </param>
		/// <param name="nsmgr">
		/// Optional.  Must contain the uri for DDI, with prefix ddi.
		/// </param>
		public static void FixCommonVariableProblems(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) nsmgr = AddNamespaces(ddi);

			FixContinuousMultipleChoiceVars(ddi, nsmgr);
			FixNumericVariableLengths(ddi, nsmgr);
		}
		/// <summary>
		/// Adds the necessary varGrp tags and adjusts the var tags as necessary
		/// to prepare a DDI file recently converted from SPSS for NAssess consumption.
		/// </summary>
		/// <param name="ddi">
		/// The DDI file with variables and no variable groups.
		/// </param>
		/// <param name="nsmgr">
		/// Optional.  Must contain the uri for DDI, with prefix ddi.
		/// </param>
		public static void PrepareDdiConvertedFromSpss(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) nsmgr = AddNamespaces(ddi);

			FixCommonVariableProblems(ddi, nsmgr);
			int groups = GroupSimilarVariables(ddi, nsmgr);
			GroupUngroupedVariables(ddi, nsmgr, groups);
			PrepareEssentialVarGrps(ddi, nsmgr);
		}

		private static XmlElement CreateVarGrpWithAllVariables(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			// If the variables have already been put into groups, reference those instead.
			if (ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp", nsmgr).Count > 0)
				return CreateVarGrpWithAllVarGrps(ddi, nsmgr);

			XmlElement nSubjectVarGrp = ddi.CreateElement("ddi", "varGrp", DdiUri);
			nSubjectVarGrp.SetAttribute("type", "subject");
			nSubjectVarGrp.SetAttribute("ID", "group1");

			XmlNodeList nlVars = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:var", nsmgr);
			StringBuilder varList = new StringBuilder(nlVars.Count * 8);
			foreach (XmlElement nVar in nlVars)
				varList.Append(nVar.GetAttribute("ID") + " ");
			nSubjectVarGrp.SetAttribute("var", varList.ToString());

			SetLabelOfVarOrVarGrp(nSubjectVarGrp, nsmgr, "All variables");

			return nSubjectVarGrp;
		}
		private static XmlElement CreateVarGrpWithAllVarGrps(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			XmlElement nSubjectVarGrp = ddi.CreateElement("ddi", "varGrp", DdiUri);
			nSubjectVarGrp.SetAttribute("type", "subject");
			nSubjectVarGrp.SetAttribute("ID", "group0");

			XmlNodeList nlVarGrps = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp", nsmgr);
			StringBuilder varGrpList = new StringBuilder(nlVarGrps.Count * 8);
			foreach (XmlElement nVarGrp in nlVarGrps)
				varGrpList.Append(nVarGrp.GetAttribute("ID") + " ");
			nSubjectVarGrp.SetAttribute("varGrp", varGrpList.ToString());

			SetLabelOfVarOrVarGrp(nSubjectVarGrp, nsmgr, "All variables");

			return nSubjectVarGrp;
		}

		/// <summary>
		/// Changes all continuous variables that have integer valued
		/// multiple choice responses to discrete variables.
		/// </summary>
		protected static void FixContinuousMultipleChoiceVars(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			// Compile a list of all continuous numeric variables with multiple choice responses.
			XmlNodeList nlContinuousMultipleChoiceNumerics = ddi.SelectNodes(
				"/ddi:codeBook/ddi:dataDscr/ddi:var[@intrvl='contin'][ddi:varFormat/@type='numeric'][ddi:catgry]", nsmgr);

			// Look at each variable, to see if we should change it's type from
			// continuous to discrete.
			foreach (XmlElement nVar in nlContinuousMultipleChoiceNumerics)
			{
				// Test whether all the multiple choice responses have integer values.
				try
				{
					foreach (XmlElement nCatValu in nVar.SelectNodes("ddi:catgry/ddi:catValu", nsmgr))
						Int32.Parse(nCatValu.InnerText.Trim());

					// They must have been all integers, because Parse never threw an exception.
					// Change the variable type to discrete.
					nVar.SetAttribute("intrvl", "discrete");
				} // safely ignore the exceptions, as they have already performed their work.
				catch (FormatException)
				{ }
				catch (OverflowException)
				{ }
				catch (ArgumentNullException)
				{ }
			}

		}
		/// <summary>
		/// Adjusts the length reserved for multiple choice numeric variables 
		/// as necessary.
		/// </summary>
		protected static void FixNumericVariableLengths(XmlDocument ddi, XmlNamespaceManager nsmgr)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");

			// Find all numeric, multiple-choice variables.
			XmlNodeList nlMultipleChoiceNumerics = ddi.SelectNodes(
				"/ddi:codeBook/ddi:dataDscr/ddi:var[ddi:varFormat/@type='numeric'][ddi:catgry]", nsmgr);

			// Go through each variable, to see what the location/@width value should be.
			foreach (XmlElement nVar in nlMultipleChoiceNumerics)
			{
				// Go through each multiple choice value and track the maximum length.
				int maxLength = 0;
				foreach (XmlElement nCatValu in nVar.SelectNodes("ddi:catgry/ddi:catValu", nsmgr))
					maxLength = Math.Max(maxLength, nCatValu.InnerText.Trim().Length);

				// Set the location attribute.
				SetLocationWidthOfVar(nVar, nsmgr, maxLength);
			}
		}

		/// <summary>
		/// Sets the ddi:labl element within a ddi:var or ddi:varGrp tag.  
		/// The ddi:labl element is created if necessary.
		/// </summary>
		protected static void SetLabelOfVarOrVarGrp(XmlElement varOrVarGroupNode, XmlNamespaceManager nsmgr, string label)
		{
			if (varOrVarGroupNode == null) throw new ArgumentNullException("nVarOrVarGrp");
			if (varOrVarGroupNode.LocalName != "varGrp" && varOrVarGroupNode.LocalName != "var")
				throw new ArgumentOutOfRangeException("nVarOrVarGrp", varOrVarGroupNode.LocalName, Resources.VarOrVarGrpExpected);
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");
			if (label == null) throw new ArgumentNullException("label");

			XmlElement labl = (XmlElement)varOrVarGroupNode.SelectSingleNode("ddi:labl", nsmgr);
			if (labl == null)
				labl = (XmlElement)varOrVarGroupNode.AppendChild(varOrVarGroupNode.OwnerDocument.CreateElement("ddi", "labl", DdiUri));
			labl.InnerText = label;
		}

		/// <summary>
		/// Sets the ddi:location/@width attribute within a ddi:var tag.
		/// The ddi:location element is created if necessary.
		/// </summary>
		protected static void SetLocationWidthOfVar(XmlElement variableNode, XmlNamespaceManager nsmgr, int width)
		{
			if (variableNode == null) throw new ArgumentNullException("nVar");
			if (variableNode.LocalName != "var")
				throw new ArgumentOutOfRangeException("nVar", variableNode.LocalName, Resources.VariableNotFound);
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");
			if (width < 1) throw new ArgumentOutOfRangeException("width", width, Resources.PositiveValueRequired);

			XmlElement location = (XmlElement)variableNode.SelectSingleNode("ddi:location", nsmgr);
			if (location == null)
				location = (XmlElement)variableNode.AppendChild(variableNode.OwnerDocument.CreateElement("ddi", "location", DdiUri));
			location.SetAttribute("width", width.ToString());
		}

		/// <summary>
		/// Adds a new varGrp tag to the DDI document.
		/// </summary>
		protected static void DefineVariableGroup(XmlDocument ddi, XmlNamespaceManager nsmgr, string id, string type, string vars, string label, XmlElement nodeBefore)
		{
			if (ddi == null) throw new ArgumentNullException("ddi");
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");
			if (id == null || id.Length == 0) throw new ArgumentNullException("id");
			if (type == null || type.Length == 0) throw new ArgumentNullException("type");
			if (vars == null || vars.Length == 0) throw new ArgumentNullException("vars");
			if (label == null) throw new ArgumentNullException("label");

			XmlElement nVarGrp = ddi.CreateElement("ddi", "varGrp", DdiUri);
			SetLabelOfVarOrVarGrp(nVarGrp, nsmgr, label);
			nVarGrp.SetAttribute("ID", id);
			nVarGrp.SetAttribute("type", type);
			nVarGrp.SetAttribute("var", vars);

			XmlElement nData = (XmlElement)ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr", nsmgr);
			if (nodeBefore == null) nodeBefore = (XmlElement)nData.SelectSingleNode("ddi:var", nsmgr);
			nData.InsertBefore(nVarGrp, nodeBefore);
		}
		#endregion
		#region DDI namespace generators
		private XmlNamespaceManager namespaceManager;
		protected XmlNamespaceManager NamespaceManager
		{
			get
			{
				if (namespaceManager == null)
					namespaceManager = AddNamespaces(Xml.NameTable);
				return namespaceManager;
			}
		}
		protected internal static void AddNamespaces(XmlNamespaceManager namespaceManager)
		{
			if (namespaceManager == null) throw new ArgumentNullException("namespaceManager");
			namespaceManager.AddNamespace("ddi", DdiUri);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNameTable nameTable)
		{
			if (nameTable == null) throw new ArgumentNullException("nametable");
			XmlNamespaceManager nsmgr = new XmlNamespaceManager(nameTable);
			AddNamespaces(nsmgr);
			return nsmgr;
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNode node)
		{
			if (node == null) throw new ArgumentNullException("node");
			XmlDocument nodeDoc = node as XmlDocument;
			return AddNamespaces(nodeDoc != null ? nodeDoc.NameTable : node.OwnerDocument.NameTable);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNodeList nodeList)
		{
			if (nodeList == null) throw new ArgumentNullException("nodeList");
			if (nodeList.Count == 0) return null;
			return AddNamespaces(nodeList[0].OwnerDocument.NameTable);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XPathNavigator xpath)
		{
			if (xpath == null) throw new ArgumentNullException("xpath");
			return AddNamespaces(xpath.NameTable);
		}
		#endregion
		#region DDI lookup functions
		protected Dictionary<string, XPathNavigator> VariableCache;
		protected Dictionary<XPathNavigator, Type> VariableTypeCache = new Dictionary<XPathNavigator, Type>();
		protected void GenerateVariableCache()
		{
			if (VariableCache == null) VariableCache = new Dictionary<string, XPathNavigator>(150);
			foreach (XPathNavigator var in FindAllVariables())
				VariableCache[var.GetAttribute("name", string.Empty)] = var;
		}
		protected XPathNodeIterator FindAllVariables()
		{
			return Select("/ddi:codeBook/ddi:dataDscr/ddi:var");
		}
		/// <summary>
		/// Used to determine whether a variable of a specified name exists in the codebook.
		/// </summary>
		/// <param name="varName">
		/// The name of the variable being sought.
		/// </param>
		/// <returns>
		/// True if the variable exists in the codebook.  Otherwise, False.
		/// </returns>
		public bool ContainsVariable(string varName)
		{
			if (VariableCache == null) GenerateVariableCache();
			return VariableCache.ContainsKey(varName);
		}
		/// <summary>
		/// Checks for the existence of a scale by a given id.
		/// </summary>
		public bool ContainsScale(string scaleId)
		{
			return FindScaleById(scaleId) != null;
		}
		/// <summary>
		/// Performs a fast XPath search of the codebook.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <param name="startNode">
		/// The node to consider the root node in the search.
		/// </param>
		/// <returns>
		/// A <see cref="XPathNodeIterator">nodeset</see> of all search hits.
		/// </returns>
		public static XPathNodeIterator Select(string xpath, XPathNavigator startNode)
		{
			if (xpath == null || xpath.Length == 0) throw new ArgumentNullException("xpath");
			if (startNode == null) throw new ArgumentNullException("startNode");
			XPathExpression exp = startNode.Compile(xpath);
			exp.SetContext(AddNamespaces(startNode));
			return startNode.Select(exp);
		}
		/// <summary>
		/// Performs a fast XPath search of the codebook, 
		/// considering the root node to be the starting node.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <returns>
		/// A <see cref="XPathNodeIterator">nodeset</see> of all search hits.
		/// </returns>
		public XPathNodeIterator Select(string xpath)
		{
			return Select(xpath, XPath.CreateNavigator());
		}
		/// <summary>
		/// Performs a fast XPath search of the codebook.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <param name="startNode">
		/// The node to consider the root node in the search.
		/// </param>
		/// <returns>
		/// An <see cref="XPathNavigator"/> pointing at the first search hit.
		/// </returns>
		public static XPathNavigator SelectSingleNode(string xpath, XPathNavigator startNode)
		{
			XPathNodeIterator nodeset = Select(xpath, startNode);
			return nodeset.MoveNext() ? nodeset.Current : null;
		}
		/// <summary>
		/// Performs a fast XPath search of the codebook,
		/// considering the root node to be the starting node.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <returns>
		/// An <see cref="XPathNavigator"/> pointing at the first search hit.
		/// </returns>
		public XPathNavigator SelectSingleNode(string xpath)
		{
			return SelectSingleNode(xpath, XPath.CreateNavigator());
		}
		/// <summary>
		/// Find a variable in the DDI codebook.
		/// </summary>
		/// <param name="varName">
		/// The name of the variable.  That is, the search is performed on the variable name attribute (not the ID attribute).
		/// </param>
		/// <returns>
		/// The variable found, or null if none found.
		/// </returns>
		public XPathNavigator FindVariableByName(string varName)
		{
			if (VariableCache != null)
			{
				try
				{
					return VariableCache[varName];
				}
				catch (KeyNotFoundException)
				{
					return null;
				}
			}
			else
				return SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@name='" + varName + "']");
		}
		public void CheckConstraints(string variableName, object value)
		{
			if (variableName == null) throw new ArgumentOutOfRangeException("variableName", variableName, Resources.VariableNotFound);
			XPathNavigator nVar = FindVariableByName(variableName);

			if (value == null) return; // don't check null values

			// Check to make sure that the value being assigned to the variable is one within the allowed range.
			XPathNodeIterator validRanges = Select("ddi:valrng/ddi:range", nVar);
			XPathNodeIterator invalidRanges = Select("ddi:invalrng/ddi:range", nVar);

			if (validRanges.Count > 0)
			{
				double v;
				if (!double.TryParse(value.ToString(), out v))
					return; // just ignore, because something else will probably catch it later.
				double min, max, minExclusive, maxExclusive;

				// Make sure that the value falls within allowed ranges
				bool pass = false;
				while (validRanges.MoveNext())
				{
					fillRanges(validRanges.Current, out min, out max, out minExclusive, out maxExclusive);
					pass |= isWithinRange(v, min, max, minExclusive, maxExclusive);
					if (pass) break; // Quit early.  We definitely found a line that includes this value
				}
				if (!pass) throw new ArgumentOutOfRangeException("value", value, Resources.ValueOutsideCodebookAllowedRange);

				// Make sure that the value does NOT fall within the disallowed ranges
				Debug.Assert(pass);
				while (invalidRanges.MoveNext())
				{
					fillRanges(invalidRanges.Current, out min, out max, out minExclusive, out maxExclusive);
					if (isWithinRange(v, min, max, minExclusive, maxExclusive))
						throw new ArgumentOutOfRangeException("value", value, Resources.ValueInExplicitlyDisallowedRange);
				}
			}
		}
		private static void fillRanges(XPathNavigator nRange, out double min, out double max, out double minExclusive, out double maxExclusive)
		{
			string aMin = nRange.GetAttribute("min", "");
			string aMax = nRange.GetAttribute("max", "");
			string aMinExclusive = nRange.GetAttribute("minExclusive", "");
			string aMaxExclusive = nRange.GetAttribute("maxExclusive", "");

			min = aMin.Length > 0 ? double.Parse(aMin) : double.NaN;
			max = aMax.Length > 0 ? double.Parse(aMax) : double.NaN;
			minExclusive = aMinExclusive.Length > 0 ? double.Parse(aMinExclusive) : double.NaN;
			maxExclusive = aMaxExclusive.Length > 0 ? double.Parse(aMaxExclusive) : double.NaN;
		}
		private static bool isWithinRange(double value, double min, double max, double minExclusive, double maxExclusive)
		{
			if (double.IsNaN(min) && double.IsNaN(max) && double.IsNaN(minExclusive) && double.IsNaN(maxExclusive))
				return false; // an empty range doesn't include everything!

			bool pass = true;
			pass &= double.IsNaN(min) || value >= min;
			pass &= double.IsNaN(max) || value <= max;
			pass &= double.IsNaN(minExclusive) || value > minExclusive;
			pass &= double.IsNaN(maxExclusive) || value < maxExclusive;
			return pass;
		}
		/// <summary>
		/// Gets a list of all variables that appear on the specified page of the questionnaire.
		/// </summary>
		/// <param name="questionnairePage">
		/// The page of the questionnaire being queried.  1-based index.
		/// </param>
		/// <param name="includeOtherFields">
		/// Whether to include fields referenced by Other keywords.  
		/// Dropdown boxes with "Other (specify)" link these text box fields in.
		/// </param>
		/// <returns>
		/// A space-delimited list of variables that appear on the page.
		/// </returns>
		protected internal string[] GetVarListForPage(int questionnairePage, bool includeOtherFields)
		{
			if (questionnairePage < 1)
				throw new ArgumentOutOfRangeException("questionnairePage", questionnairePage, "Should be between 1 and PagesCount.");
			string[] varGrpIDsForPages = SelectSingleNode(
				"/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='display']/@varGrp").Value.Trim().Split(' ');
			if (questionnairePage > varGrpIDsForPages.Length)
				throw new ArgumentOutOfRangeException("questionnairePage", "Value is higher than last page.");
			XPathNavigator nVarGrpForPage = SelectSingleNode(string.Format(
				"/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='section'][@ID='{0}']", varGrpIDsForPages[questionnairePage - 1]));
			Debug.Assert(nVarGrpForPage != null);
			List<string> varList = new List<string>(10);
			GetVarsInVarGrp(nVarGrpForPage, varList, includeOtherFields);
			return varList.ToArray();
		}
		protected internal void GetVarsInVarGrp(XPathNavigator varGrpNode, IList<string> varList, bool includeOtherFields)
		{
			Debug.Assert(varGrpNode != null);
			Debug.Assert(varList != null);

			if (varGrpNode.GetAttribute("var", "").Length > 0)
			{
				string[] varIDs = varGrpNode.GetAttribute("var", "").Trim().Split(' ');
				foreach (string varID in varIDs)
				{
					// Translate varID's to their column name equivalents.
					XPathNavigator nVar = SelectSingleNode(string.Format(
						"../ddi:var[@ID='{0}']", varID), varGrpNode);
					Debug.Assert(nVar != null);
					if (!includeOtherFields && SelectSingleNode("ddi:varFormat[@type='character' and @category='other']", nVar) != null)
						continue; // skip, because it's an "other" field
					varList.Add(nVar.GetAttribute("name", ""));
				}
			}
			if (varGrpNode.GetAttribute("varGrp", "").Length > 0)
			{
				string[] varGrpIDs = varGrpNode.GetAttribute("varGrp", "").Trim().Split(' ');
				foreach (string varGrpID in varGrpIDs)
				{
					XPathNavigator nInnerVarGrp = SelectSingleNode(string.Format(
						"../ddi:varGrp[@ID='{0}']", varGrpID), varGrpNode);
					GetVarsInVarGrp(nInnerVarGrp, varList, includeOtherFields);
				}
			}
		}

		#endregion
		#region Codebook Questions and Answers
		protected const string ScaleVarGrpFilter = "[@type='analysis'][@var or @varGrp]"; // [ddi:labl]
		protected internal XPathNodeIterator FindAllScales()
		{
			return Select("/ddi:codeBook/ddi:dataDscr/ddi:varGrp" + ScaleVarGrpFilter);
		}
		/// <param name="index">
		/// A 1-based index of the scale to retrieve.
		/// </param>
		protected internal XPathNavigator FindScaleByIndex(int index)
		{
			if (index < 1) throw new ArgumentOutOfRangeException("index", index, Resources.PositiveValueRequired);
			return SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp" + ScaleVarGrpFilter + "[" + index.ToString() + "]");
		}
		protected internal XPathNavigator FindScaleById(string scaleId)
		{
			if (scaleId == null || scaleId.Length == 0) throw new ArgumentNullException("scaleId");

			return SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp" +
				ScaleVarGrpFilter + "[@ID=" + StringUtilities.QuoteXPathString(scaleId) + "]");
		}
		protected XPathNavigator FindVarById(string varId)
		{
			if (varId == null || varId.Length == 0) throw new ArgumentNullException("varId");

			return SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var" +
				"[@ID=" + StringUtilities.QuoteXPathString(varId) + "]");
		}
		protected internal void GetAllVariablesInScale(XPathNavigator scaleNode, out XPathNavigator[] variables, out StringCollection reversedVariableIdsList)
		{
			if (scaleNode == null) throw new ArgumentNullException("nScale");
			if (!IsScale(scaleNode)) throw new ArgumentOutOfRangeException("nScale", scaleNode.GetAttribute("ID", ""), "Not a varGrp that defines a scale.");

			List<XPathNavigator> listVariables = new List<XPathNavigator>();
			reversedVariableIdsList = new StringCollection();
			GetAllVariablesInScale(scaleNode, listVariables, reversedVariableIdsList);

			variables = listVariables.ToArray();
		}
		private void GetAllVariablesInScale(XPathNavigator scaleNode, System.Collections.Generic.IList<XPathNavigator> variables, StringCollection reversedVariables)
		{
			Debug.Assert(scaleNode != null);
			Debug.Assert(variables != null);
			Debug.Assert(reversedVariables != null);
			Debug.Assert(IsScale(scaleNode), "Supplied XmlElement is not a scale.");

			XPathNavigator nReversed = SelectSingleNode("ddi:concept[@vocab='reversed']", scaleNode);

			string vars = scaleNode.GetAttribute("var", "");
			string varsReversed = (nReversed == null) ? "" : nReversed.Value;
			string varGrps = scaleNode.GetAttribute("varGrp", "");

			// Add any variables that are in the scale
			foreach (string varId in StringUtilities.Tokenize(vars))
				variables.Add(FindVarById(varId));

			// Add any variables that are reversed
			foreach (string varId in StringUtilities.Tokenize(varsReversed))
				reversedVariables.Add(varId);

			// Add any sub-scales that have variables
			foreach (string subScale in StringUtilities.Tokenize(varGrps))
				GetAllVariablesInScale(FindScaleById(subScale), variables, reversedVariables);
		}
		protected static bool IsScale(XPathNavigator varGrp)
		{
			if (varGrp == null) throw new ArgumentNullException("varGrp");
			return varGrp.NamespaceURI == Ddi.DdiUri && varGrp.LocalName == "varGrp"
				&& varGrp.GetAttribute("type", "") == "analysis";
		}
		#endregion

		#region Codebook DataTable construction
		/// <summary>
		/// Builds a new codebook custom data table without columns for scales.
		/// </summary>
		public DataTable BuildDataTable()
		{
			return BuildDataTable(false);
		}
		/// <summary>
		/// Builds a new codebook custom data table.
		/// </summary>
		/// <param name="includeScales">
		/// Whether to include columns for scales in the table.
		/// </param>
		public DataTable BuildDataTable(bool includeScales)
		{
			// Create a DataTable that we will use to build up a schema to describe
			// the new table we wish to create.
			DataTable dt = new DataTable(XPath.CreateNavigator().GetAttribute("ID", DdiUri));
			dt.Locale = System.Globalization.CultureInfo.InvariantCulture;

			// Go through the codebook and add all the variables.
			foreach (XPathNavigator nVar in FindAllVariables())
				dt.Columns.Add(ConstructDataColumnFromDdiVariable(nVar));

			if (includeScales)
			{
				foreach (XPathNavigator nScale in FindAllScales())
					dt.Columns.Add(nScale.GetAttribute("ID", string.Empty), typeof(double)).AllowDBNull = true;
			}

			return dt;
		}
		protected DataColumn ConstructDataColumnFromDdiVariable(XPathNavigator variableNode)
		{
			if (variableNode == null) throw new ArgumentNullException("variableNode");
			DataColumn dcVar = new DataColumn(variableNode.GetAttribute("name", string.Empty));
			dcVar.AllowDBNull = true; // all questions are optional
			SetDataTypeFromDdiVar(dcVar, variableNode);
			return dcVar;
		}
		protected Type GetDataTypeFromDdiVar(XPathNavigator variableNode)
		{
			if (variableNode == null) throw new ArgumentNullException("nVar");
			Type type;
			if (VariableTypeCache.TryGetValue(variableNode, out type))
				return type;

			XPathNavigator aVarType = variableNode.SelectSingleNode("ddi:varFormat/@type", NamespaceManager);
			XPathNavigator aVarFormat = variableNode.SelectSingleNode("ddi:varFormat/@formatname", NamespaceManager);
			if (aVarType == null) Trace.WriteLine("WARNING: varFormat/@type not specified for variable " + variableNode.GetAttribute("name", DdiUri)
									  + ".  Assuming type \"character\".");
			string sVarType = aVarType != null ? aVarType.Value : "character";
			string sVarFormat = aVarFormat != null ? aVarFormat.Value : "";
			switch (sVarType)
			{
				case "character":
					type = typeof(string);
					break;
				case "numeric":
					// even though the type says numeric, it could actually be a date field.
					if (sVarFormat.IndexOf("date") >= 0)
						type = typeof(DateTime);
					else // numeric
					{
						if (variableNode.GetAttribute("intrvl", DdiUri) == "contin")
							type = typeof(float);
						else
							type = typeof(int);
					}
					break;
				default:
					throw new NotSupportedException("The variable data type " +
						aVarType + " is not yet supported.");
			}
			return VariableTypeCache[variableNode] = type;
		}
		protected void SetDataTypeFromDdiVar(DataColumn dataColumn, XPathNavigator variableNode)
		{
			if (dataColumn == null) throw new ArgumentNullException("dataColumn");
			if (variableNode == null) throw new ArgumentNullException("nVar");

			dataColumn.DataType = GetDataTypeFromDdiVar(variableNode);
			if (dataColumn.DataType == typeof(string))
			{
				XPathNavigator aWidth = SelectSingleNode("ddi:location/@width", variableNode);
				dataColumn.MaxLength = (aWidth != null) ? Convert.ToInt32(aWidth.Value) : 255;
			}
		}
		#endregion

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Ddi other = obj as Ddi;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}
	}
}
