using System;
using System.Data;
using System.Data.SqlClient;
using System.Collections.Specialized;
using System.Collections.Generic;
using System.Xml;
using System.Xml.XPath;
using System.Diagnostics;
using System.Text;
using System.Text.RegularExpressions;
using System.Globalization;

using Iesi.Collections;

using MyNHibernateContrib.TwoWayCollections;
using MyNHibernateContrib.GenericCollections;
using NAssess.Core.FollowUp;

namespace NAssess.Core.Evaluating
{
	/// <summary>
	/// An object that represents a questionnaire that is in the process of being filled out.
	/// </summary>
	public class Evaluation : IParentReferencingNode<Evaluation, Questionnaire>,
		IReferenceableParent<EvaluationTracker, Evaluation>
	{
		#region Attributes
		private int id;
		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}

		private Questionnaire questionnaire;
		/// <summary>
		/// The <see cref="Questionnaire"/> to which this Evaluation belongs.
		/// </summary>
		public virtual Questionnaire Questionnaire
		{
			get { return questionnaire; }
			set { questionnaire = value; }
		}

		private CultureInfo culture;
		/// <summary>
		/// Gets/sets the culture of the questionnaire associated with this evaluation.
		/// </summary>
		public virtual CultureInfo Culture
		{
			get { return culture; }
			set { culture = value; }
		}

		/// <summary>
		/// Gets the codebook in active use for this evaluation.
		/// </summary>
		public virtual Ddi Codebook
		{
			get
			{
				if (Questionnaire == null) return null;
				if (Culture != null)
					return Questionnaire.Codebooks[Culture];
				else
					return Questionnaire.Codebook;
			}
		}

		private bool includeInDataSet = true;
		/// <summary>
		/// Indicates whether calls to EvaluationsCollection.GetDataSet() and similar
		/// methods will return this row in the dataset.
		/// </summary>
		/// <remarks>
		/// This property is useful when you wish to mark a row "deleted" so it does not
		/// affect data, but do not actually want to destroy the data, in case it need to be 
		/// recovered.
		/// </remarks>
		public virtual bool IncludeInDataSet
		{
			get { return includeInDataSet; }
			set { includeInDataSet = value; }
		}

		private bool consentGiven;
		/// <summary>
		/// Gets/sets whether the respondent accepted the terms of the
		/// <see cref="NAssess.Core.Evaluating.Questionnaire.Consent">consent form</see>.
		/// </summary>
		public virtual bool ConsentGiven
		{
			get { return consentGiven; }
			set { consentGiven = value; }
		}

		private DateTime startDate = DateTime.Now;
		/// <summary>
		/// The date that this evaluation was started.
		/// </summary>
		public virtual DateTime StartDate
		{
			get { return startDate; }
			set { startDate = value; }
		}

		private DateTime? endDate;
		/// <summary>
		/// The date that this evaluation was completed.
		/// </summary>
		public virtual DateTime? EndDate
		{
			get { return endDate; }
			set
			{
				DateTime? oldValue = endDate;

				endDate = value;

				if (value.HasValue != oldValue.HasValue)
				{
					if (value.HasValue)
						OnCompleted();
					else
						OnResume();
				}
			}
		}
		public virtual void ImportEndDate(DateTime endDate) { this.endDate = endDate; }

		/// <summary>
		/// Gets the time taken to complete this evaluation.
		/// </summary>
		public TimeSpan? Duration
		{
			get
			{
				if (!EndDate.HasValue) return null;
				return EndDate - StartDate;
			}
		}

		/// <summary>
		/// Gets/sets whether this evaluation has been completed.
		/// </summary>
		/// <remarks>
		/// Setting this property updates the <see cref="EndDate"/> property.
		/// </remarks>
		public bool IsCompleted
		{
			get { return EndDate.HasValue; }
			set { EndDate = value ? DateTime.Now : (DateTime?)null; }
		}

		/// <summary>
		/// Gets whether this evaluation has been completed and if its associated <see cref="Questionnaire"/> has any usable reports.
		/// </summary>
		public virtual bool IsReportAvailable
		{
			get
			{
				return IsCompleted && Questionnaire != null && Questionnaire.Assessments.Count > 0;
			}
		}

		private HtmlRenderingEngine htmlEngine;
		public virtual HtmlRenderingEngine HtmlEngine
		{
			get
			{
				if (htmlEngine == null)
				{
					htmlEngine = new Engines.Evaluation.StandardHtml();
					htmlEngine.Evaluation = this;
				}
				return htmlEngine;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("HtmlEngine", Resources.HtmlEngineRequired);
				htmlEngine = value;
				htmlEngine.Evaluation = this;
			}
		}
		/// <summary>
		/// Gets the number of pages that are in the questionnaire, given the
		/// currently supplied responses.  
		/// </summary>
		/// <remarks>
		/// This number may depend on the responses already provided, and thus
		/// may change with each submission of responses.
		/// </remarks>
		public int PagesCount
		{
			get
			{
				return HtmlEngine.PagesCount;
			}
		}

		/// <summary>
		/// Gets the percent (between 0 and 1) that the specified page has been filled out with answers.
		/// </summary>
		/// <param name="page">
		/// The page of the questionnaire being queried.  1-based index.
		/// </param>
		/// <param name="totalQuestions">
		/// Outputs the total number of questions on the page.  
		/// Useful for doing a weighted average across the entire evaluation.
		/// </param>
		/// <returns>
		/// A floating-point number between zero and one that represents the fraction of the page that has been filled out.
		/// </returns>
		/// <remarks>
		/// This method does not consider "Other" fields in its figuring.
		/// </remarks>
		protected float PageCompletion(int page, out int totalQuestions)
		{
			string[] varList = Codebook.GetVarListForPage(page, false);
			totalQuestions = varList.Length;

			// Protect ourselves from a divison by zero error later.
			if (varList.Length == 0) throw new InvalidOperationException(Resources.NoQuestionsOnPage);

			// Count up how many responses have been given
			int NumberOfNonNullResponses = 0;
			foreach (string varName in varList)
				if (Responses[varName] != null) NumberOfNonNullResponses++;

			return (float)NumberOfNonNullResponses / varList.Length;
		}
		/// <summary>
		/// Gets the percent (between 0 and 1) that the specified page has been filled out with answers.
		/// </summary>
		/// <param name="page">
		/// The page of the questionnaire being queried.  1-based index.
		/// </param>
		/// <returns>
		/// A floating-point number between zero and one that represents the fraction of the page that has been filled out.
		/// </returns>
		/// <remarks>
		/// This method does not consider "Other" fields in its figuring.
		/// </remarks>
		public float PageCompletion(int page)
		{
			int totalQuestions;
			return PageCompletion(page, out totalQuestions);
		}
		/// <summary>
		/// Gets the proportion (between 0 and 1) of the evaluation that has been completed.
		/// </summary>
		public float AmountCompleted
		{
			get
			{
				int pagesCount = PagesCount;
				int totalQuestions = 0;
				float evalProportion = 1;
				for (int iPage = 1; iPage < pagesCount; iPage++)
				{
					int pageQuestionsCount;
					float pageProportion = PageCompletion(iPage, out pageQuestionsCount);
					evalProportion = (evalProportion * totalQuestions + pageProportion * pageQuestionsCount)
						/ (totalQuestions + pageQuestionsCount);
					totalQuestions += pageQuestionsCount;
				}

				return evalProportion;
			}
		}

		private EventTracker followUpEvent;
		/// <summary>
		/// The follow-up study event responsible for the start of this evaluation, if any.
		/// </summary>
		public virtual EventTracker FollowUpEvent
		{
			get { return followUpEvent; }
			protected internal set { followUpEvent = value; }
		}

		#endregion

		#region Events
		/// <summary>
		/// Fired when the <see cref="Evaluation"/> has been completed by the user.
		/// </summary>
		public event EventHandler Completed;
		/// <summary>
		/// Fired when an <see cref="Evaluation"/> which was previously marked as 
		/// completed is now reversed to an in-progress state.
		/// </summary>
		public event EventHandler Resume;

		/// <summary>
		/// Fires the <see cref="Completed"/> event.
		/// </summary>
		protected virtual void OnCompleted()
		{
			EnsureFollowUpStudyEventsHooked();
			EventHandler completed = Completed;
			if (completed != null)
				completed(this, null);
		}

		private void EnsureFollowUpStudyEventsHooked()
		{
			// Due to lazy loading, all we need to do is 
			// access the follow-up event to load it.
			Study study;
			if (FollowUpEvent != null)
				study = FollowUpEvent.Study;
		}
		/// <summary>
		/// Fires the <see cref="Resume"/> event.
		/// </summary>
		protected virtual void OnResume()
		{
			EventHandler resume = Resume;
			if (resume != null)
				resume(this, null);
		}

		#endregion

		#region Collections
		private System.Collections.IDictionary responses;
		private System.Collections.IDictionary ResponsesSet
		{
			get
			{
				if (responses == null) responses = new HybridDictionary();
				return responses;
			}
			set
			{
				responses = value;
				typedResponsesWrapper = null;
			}
		}
		private EvaluationResponses typedResponsesWrapper;
		/// <summary>
		/// Gets/sets the respondent's responses to the questions on the evaluation form.
		/// </summary>
		public virtual EvaluationResponses Responses
		{
			get
			{
				if (typedResponsesWrapper == null)
					typedResponsesWrapper = new EvaluationResponses(this, ResponsesSet);
				return typedResponsesWrapper;
			}
		}

		/// <remarks>
		/// Included for NHibernate support
		/// </remarks>
		private System.Collections.IDictionary UntypedResponses
		{
			get
			{
				System.Collections.IDictionary d = new System.Collections.Specialized.HybridDictionary(Responses.Count);
				foreach (KeyValuePair<string, object> de in Responses)
					d.Add(de.Key, de.Value);
				return d;
			}
			set
			{
				Responses.Clear();
				foreach (System.Collections.DictionaryEntry de in value)
					Responses.Add((string)de.Key, de.Value);
			}
		}

		private ISet followUpsSet;
		private ISet FollowUpsSet
		{
			get
			{
				if (followUpsSet == null)
					followUpsSet = new ListSet();
				return followUpsSet;
			}
			set
			{
				followUpsSet = value;
			}
		}
		private ITwoWayCollection<FollowUp.EvaluationTracker, Evaluation> followUps;
		public virtual ITwoWayCollection<FollowUp.EvaluationTracker, Evaluation> FollowUps
		{
			get
			{
				if (followUps == null)
					followUps = new TwoWaySet<EvaluationTracker, Evaluation>(this, FollowUpsSet);
				return followUps;
			}
		}

		private EvaluationScales scales;
		public virtual EvaluationScales Scales
		{
			get
			{
				if (scales == null)
					scales = new EvaluationScales(this);
				return scales;
			}
		}

		#endregion

		#region Operations
		/// <summary>
		/// Gets the web HTML that will allow a user to fill out the evaluation and submit the answers.
		/// </summary>
		/// <param name="page">
		/// The 1-based indexed of the page you want to get the HTML for.
		/// </param>
		/// <param name="writer">
		/// The <see cref="XmlWriter"/> to use to write the resulting HTML.
		/// </param>
		/// <returns>
		/// A string of HTML that will display the questionnaire in a browser.
		/// </returns>
		public void ToHtml(int page, XmlWriter writer)
		{
			HtmlEngine.ToHtml(page, writer);
		}
		public string ToHtml(int page)
		{
			return HtmlEngine.ToHtml(page);
		}
		/// <summary>
		/// Adds the evaluation to any applicable follow-up studies.
		/// </summary>
		/// <remarks>
		/// Sign up for the highest priority exclusive study and no others,
		/// or if no exclusive studies qualify this evaluation, sign up for
		/// all the (non-exclusive) studies that qualify.
		/// </remarks>
		public bool RegisterForFollowUps(INAssessManager manager)
		{
			if (manager == null) throw new ArgumentNullException("manager");
			if (!IsCompleted)
				throw new InvalidOperationException("Evaluation.IsCompleted must be true before follow-ups can be registered.");
			bool matchingStudyFound = false;

			List<Study> nonExclusiveQualifyingStudies = new List<Study>(3);
			Study bestExclusiveStudy = null;
			foreach (Study study in manager.GetStudies())
				if (study.Qualifies(this))
				{
					if (study.IsExclusive)
					{
						if (bestExclusiveStudy == null)
							bestExclusiveStudy = study;
						else if (bestExclusiveStudy.Priority < study.Priority)
							bestExclusiveStudy = study;
					}
					else
						nonExclusiveQualifyingStudies.Add(study);
				}
			// It is possible that this evaluation has already been hooked by another
			// study.  If an exclusive study now wants to nab it, we can unhook the
			// existing studies if they have not actually started yet.
			if (bestExclusiveStudy != null && ClearAnyExistingTrackersIfPossible())
			{
				bestExclusiveStudy.Track(this);
				matchingStudyFound = true;
			}
			else
			{
				foreach (Study study in nonExclusiveQualifyingStudies)
					study.Track(this);
				matchingStudyFound = nonExclusiveQualifyingStudies.Count > 0;
			}
			return matchingStudyFound;
		}
		/// <summary>
		/// Checks for any existing hooks into follow-up studies.  If there are any,
		/// and if none of them have been started yet, then they are all
		/// removed.
		/// </summary>
		/// <returns>
		/// True if no follow-up study trackers remain at the end of this call.
		/// False if some remain.
		/// </returns>
		private bool ClearAnyExistingTrackersIfPossible()
		{
			if (FollowUps.Count == 0) return true;
			foreach (EvaluationTracker tracker in FollowUps)
				// fail if any of the studies already have begun any events.
				if (tracker.TrackedEvents.Count > 1 ||
					(tracker.TrackedEvents.Count == 1 && tracker.TrackedEvents[0].State != EventTrackerState.Dormant))
					return false; // indicate presence of non-removable trackers

			// All trackers are removable!
			// First remove them from the study's collection.
			foreach (EvaluationTracker tracker in FollowUps)
				tracker.Study.Trackers.Remove(tracker);
			// Now remove them all from this evaluation.
			FollowUps.Clear();
			return true;
		}
		#endregion

		#region IParentReferencingNode<Evaluation, Questionnaire> Members

		Questionnaire IParentReferencingNode<Evaluation, Questionnaire>.Parent
		{
			get { return questionnaire; }
			set { questionnaire = value; }
		}

		#endregion

		#region IReferenceableParent<EvaluationTracker,Evaluation> Members

		ITwoWayCollection<EvaluationTracker, Evaluation> IReferenceableParent<EvaluationTracker, Evaluation>.ContainingCollection
		{
			get { return FollowUps; }
		}

		#endregion

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Evaluation other = obj as Evaluation;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}

		public class EvaluationScales
		{
			protected Evaluation Evaluation;

			#region Construction
			internal EvaluationScales(Evaluation evaluation)
			{
				Evaluation = evaluation;
			}

			#endregion

			#region Attributes
			/// <summary>
			/// Get the number of scales available.
			/// </summary>
			public int Count
			{
				get
				{
					return Evaluation.Codebook.FindAllScales().Count;
				}
			}

			/// <summary>
			/// Get the value of the specified scale.
			/// </summary>
			public float this[string scaleId]
			{
				get
				{
					return ScaleAverage(Evaluation.Codebook.FindScaleById(scaleId));
				}
			}
			#endregion

			#region Iterator
			public virtual Enumerator GetEnumerator()
			{
				return new Enumerator(this);
			}

			public class Enumerator
			{
				EvaluationScales es;
				protected int index;
				protected readonly int count;

				protected internal Enumerator(EvaluationScales es)
				{
					if (es == null) throw new ArgumentNullException("es");
					this.es = es;
					count = es.Count;
				}

				public bool MoveNext()
				{
					index++;
					return (index <= count);
				}

				public class Scale
				{
					private string name;
					private Single _value;
					public Scale(string name, float value)
					{
						this.name = name;
						_value = value;
					}
					public string Name
					{
						get
						{
							return name;
						}
					}
					public float Value
					{
						get
						{
							return _value;
						}
					}
				}

				public Scale Current
				{
					get
					{
						XPathNavigator nScale = es.Evaluation.Codebook.FindScaleByIndex(index);
						return new Scale(nScale.GetAttribute("ID", ""), es[nScale.GetAttribute("ID", "")]);
					}
				}

			}
			#endregion

			#region Operations
			/// <summary>
			/// Creates an XmlNode that contains attributes to describe every scale defined
			/// in the set of codebooks that make up this evaluation.
			/// </summary>
			/// <param name="docSource">
			/// Optional. The document that should be used to create the XmlNode, if necessary.
			/// </param>
			/// <returns>
			/// The XmlNode containing the scales.  In the format &lt;scales scale1="value" ... /&gt;
			/// </returns>
			public virtual XmlNode ToXml(XmlDocument docSource)
			{
				XmlDocument docScales = docSource == null ? new XmlDocument() : docSource;
				XmlNode nScales;
				nScales = docScales.CreateElement("scales");
				foreach (Enumerator.Scale scale in this)
				{
					XmlAttribute attr = docScales.CreateAttribute(scale.Name);
					if (!float.IsNaN(scale.Value))
					{
						attr.Value = scale.Value.ToString();
						nScales.Attributes.Append(attr);
					}
				}

				return nScales;
			}
			protected float ScaleAverage(XPathNavigator scaleNode)
			{
				if (scaleNode == null) throw new ArgumentNullException("scaleNode");

				XPathNavigator[] variables;
				StringCollection reversedVariables;
				Evaluation.Codebook.GetAllVariablesInScale(scaleNode, out variables, out reversedVariables);
				int sumCount = 0;
				float sum = 0;
				foreach (XPathNavigator varNavigator in variables)
				{
					object objValue = Evaluation.Responses[varNavigator.GetAttribute("name", "")];
					if (objValue == null) continue; // skip to next one
					float varValue = Convert.ToSingle(objValue);
					if (varValue < 0) continue; // negative values indicate N/A, and shouldn't be included in the scale
					sum += reversedVariables.Contains(varNavigator.GetAttribute("ID", "")) ? ReverseVariableValue(varNavigator, varValue) : varValue;
					sumCount++;
				}
				if (sumCount == 0) return float.NaN; // no data
				return sum / sumCount;
			}
			protected static float ReverseVariableValue(XPathNavigator var, float value)
			{
				if (var == null) throw new ArgumentNullException("var");

				// Just take the first range that is specified as the one to use
				XPathNavigator nRange = Ddi.SelectSingleNode("ddi:valrng/ddi:range[@min][@max]", var);
				if (nRange == null) throw new ApplicationException("No range available to use to reverse the scale.");
				float min = Convert.ToSingle(nRange.GetAttribute("min", ""));
				float max = Convert.ToSingle(nRange.GetAttribute("max", ""));

				if (value < min || value > max) throw new ArgumentOutOfRangeException("value", value,
					string.Format(Resources.ValueOutsideReversibleRange, var.GetAttribute("ID", ""), min, max));
				// If responses 1,2,3,4,5 are options, we transform to 5,4,3,2,1.  But we do not tamper with negatives.
				// Formula: (Min + Max - Value)
				// Applied to example: (3 + 5 - 3) = 5, (3 + 5 - 4) = 4, (3 + 5 - 5) = 3
				return min + max - value;
			}
			#endregion
		}
	}
}
