using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.XPath;

using NerdBank.Tools;
using MyNHibernateContrib.GenericCollections;

namespace NAssess.Core.Evaluating
{
	public class EvaluationResponses : IDictionaryWrapper<string, object>
	{
		protected internal EvaluationResponses(Evaluation evaluation, System.Collections.IDictionary dictionary)
			: base(dictionary)
		{
			if (evaluation == null) throw new ArgumentNullException("evaluation");
			Evaluation = evaluation;
		}

		private Evaluation evaluation;
		protected Evaluation Evaluation
		{
			get { return evaluation; }
			private set { evaluation = value; }
		}
		protected Ddi Codebook
		{
			get { return (Evaluation != null) ? Evaluation.Codebook : null; }
		}

		private bool inBatchChangeMode;

		/// <summary>
		/// Provides access to the individual questions and responses in the questionnaire.
		/// If you try to set a response that is of type int by using a string, the string
		/// will be converted to an int automatically, if possible.  
		/// </summary>
		/// <exception cref="ArgumentOutOfRangeException">
		/// Thrown when illegal input is set using this accessor, or when
		/// the variableName is invalid.
		/// </exception>
		public new object this[string variableName]
		{
			get
			{
				if (variableName == null) throw new ArgumentNullException("variableName");
				if (Codebook != null)
				{
					if (!Codebook.ContainsVariable(variableName))
						throw new ArgumentOutOfRangeException("variableName", variableName, Resources.VariableNotFound);
					return ContainsKey(variableName) ? Codebook.PrepareDataValue(variableName, base[variableName]) : null;
				}
				else
					return ContainsKey(variableName) ? base[variableName] : null;
			}
			set
			{
				if (variableName == null) throw new ArgumentNullException("variableName");
				if (Codebook != null)
				{
					if (!Codebook.ContainsVariable(variableName))
						throw new ArgumentOutOfRangeException("variableName", variableName, Resources.VariableNotFound);
					Codebook.CheckConstraints(variableName, value);
					// we convert to some other type, and then back to a string for the database table
					if (value != null && value.ToString().Length > 0)
						base[variableName] = Convert.ToString(Codebook.PrepareDataValue(variableName, value));
					else if (this.ContainsKey(variableName))  // if question was left blank
						this.Remove(variableName);  // remove empty entry from database
				}
				else
				{
					if (value != null && value.ToString().Length > 0)
						base[variableName] = Convert.ToString(value);
					else if (this.ContainsKey(variableName))  // if question was left blank
						this.Remove(variableName);  // remove empty entry from database
				}
				if (!inBatchChangeMode) OnChanged(new string[] { variableName });
			}
		}

		#region Events
		public delegate void ResponsesChangedEventHandler(Evaluation sender, string[] variableNames);
		/// <summary>
		/// Fired when a response or set of responses has changed.
		/// </summary>
		public event ResponsesChangedEventHandler Changed;
		protected void OnChanged(string[] variableNames)
		{
			ResponsesChangedEventHandler changed = Changed;
			if (changed != null) changed(Evaluation, variableNames);
		}
		#endregion

		#region Import/export
		/// <summary>
		/// Assembles a single XML tag whose attributes supply all the user's responses that have
		/// been provided thus far.
		/// </summary>
		/// <param name="docSource">
		/// The source XML document that should be used to
		/// generate the XmlNode.  Optional, but can be useful if you will be adding the
		/// returned XmlNode to some XmlDocument.
		/// </param>
		/// <returns>An XML &lt;responses/&gt; tag in a System.String object.</returns>
		public virtual XmlNode ToXml(XmlDocument docSource)
		{
			XmlDocument docResponses = (docSource == null) ? new XmlDocument() : docSource;
			XmlNode nResponses;
			nResponses = docResponses.CreateElement("responses");
			XPathNodeIterator nlVars = Codebook.Select(
				"/ddi:codeBook/ddi:dataDscr/ddi:var[@name]");
			while (nlVars.MoveNext())
			{
				XPathNavigator nVar = nlVars.Current;

				XmlAttribute attr = docResponses.CreateAttribute(nVar.GetAttribute("name", ""));
				object val = this[attr.Name];
				if (val != null)
				{
					attr.Value = val.ToString();
					nResponses.Attributes.Append(attr);
				}
			}

			return nResponses;
		}
		/// <summary>
		/// This function is identical to changing each answer directly using the indexer,
		/// except that for changing many answers at once, this method will be faster,
		/// especially for out-of-process calls.
		/// </summary>
		/// <param name="node">An XML node of answers in the format 
		/// <responses v1="1" v2="5" />
		/// </param>
		public virtual void FromXml(XmlNode node)
		{
			if (node.HasChildNodes || node.Name != "responses")
				throw new ArgumentException("Responses not provided in required schema.", "xmlResponses");
			lock (this)
			{
				try
				{
					inBatchChangeMode = true;
					List<string> variableNames = new List<string>(node.Attributes.Count);
					foreach (XmlNode response in node.Attributes)
					{
						this[response.Name] = response.Value == "" ? null : response.Value;
						variableNames.Add(response.Name);
					}
					OnChanged(variableNames.ToArray());
				}
				finally
				{
					inBatchChangeMode = false;
				}
			}
		}
		/// <summary>
		/// Saves the responses just submitted on a PostBack operation in ASP.NET.  
		/// </summary>
		/// <returns>
		/// An array of field names that caused errors on submission.
		/// </returns>
		/// <exception cref="System.InvalidOperationException">
		/// Thrown when no page PostBack context could be found.
		/// </exception>
		public string[] FromHtml()
		{
			System.Web.HttpContext context = System.Web.HttpContext.Current;
			if (context == null) throw new InvalidOperationException("No context page postback could be detected.");
			System.Web.HttpRequest Request = context.Request;
			if (Request.Form["VarNamesHint"] == null) throw new InvalidOperationException("No VarNamesHint incoming field could be found.");

			List<string> troubleColumns = new List<string>();

			lock (this)
			{
				try
				{
					inBatchChangeMode = true;
					string[] vars = StringUtilities.Tokenize(Request.Form["VarNamesHint"]);
					List<string> goodColumns = new List<string>(vars.Length);
					foreach (string varName in vars)
					{
						try
						{
							this[varName] = Codebook.PrepareDataValue(varName, Request.Form[varName]);
							goodColumns.Add(varName);
						}
						catch (ArgumentOutOfRangeException)
						{
							troubleColumns.Add(varName);
						}
					}
					OnChanged(goodColumns.ToArray());
				}
				finally
				{
					inBatchChangeMode = false;
				}
			}

			return troubleColumns.ToArray();
		}
		#endregion
	}
}
