using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Globalization;
using System.Data;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.IO;
using System.Data.SqlClient;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;

using Iesi.Collections;

using NAssess.Core.Reporting;
using NerdBank.Tools;
using MetaData;
using MyNHibernateContrib.TwoWayCollections;
using MyNHibernateContrib.GenericCollections;

namespace NAssess.Core.Evaluating
{
	/// <summary>
	/// An entity object that represents a questionnaire that is associated with a 
	/// particular site.  This object provides access to
	/// design questionnaires, and also to access associated assessments.
	/// </summary>
	public class Questionnaire
		: IReferenceableParent<Assessment, Questionnaire>
	{
		#region Constructors
		public Questionnaire()
		{
		}
		#endregion

		#region Attributes
		private int id;
		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}
		private string name;
		public virtual string Name
		{
			get { return name; }
			set { name = value; }
		}
		private DateTime createdOn = DateTime.Now;
		/// <summary>
		/// Gets the date this questionnaire was created on.
		/// </summary>
		public virtual DateTime CreatedOn
		{
			get { return createdOn; }
			private set { createdOn = value; }
		}

		private FORenderingEngine foEngine;
		public virtual FORenderingEngine FOEngine
		{
			get
			{
				if (foEngine == null)
				{
					foEngine = new NAssess.Core.Engines.Questionnaire.StandardFO();
					foEngine.Questionnaire = this;
				}
				return foEngine;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("FOEngine", Resources.FOEngineRequired);
				if (foEngine != null) foEngine.Questionnaire = null;
				foEngine = value;
				foEngine.Questionnaire = this;
			}
		}
		private PdfRenderingEngine pdfEngine;
		public virtual PdfRenderingEngine PdfEngine
		{
			get { return pdfEngine; }
			set
			{
				pdfEngine = value;
				if (pdfEngine != null)
					pdfEngine.Questionnaire = this;
			}
		}

		private CultureInfo primaryCulture = CultureInfo.CurrentUICulture;
		public virtual CultureInfo PrimaryCulture
		{
			get { return primaryCulture; }
			set
			{
				if (value == null) throw new ArgumentNullException("PrimaryCulture");
				primaryCulture = value;
			}
		}
		/// <summary>
		/// Gets or sets the DDI-compliant XML that describes the codebook.  
		/// Upon being set, the codebook immediately creates/adjusts the 
		/// DataTable schema that stores responses collected by this questionnaire.
		/// </summary>
		public virtual Ddi Codebook
		{
			get
			{
				// Search for the codebook under the current culture.
				return GetMostRelevantLocalResource<Ddi>(Codebooks);
			}
			set
			{
				Codebooks[Thread.CurrentThread.CurrentUICulture] = value;
			}
		}
		/// <summary>
		/// Gets the codebook culture to use when no culture is specified.
		/// </summary>
		public CultureInfo DefaultCodebookCulture
		{
			get { return CultureAssist.GetMostRelevantCulture(Codebooks.Keys, GetRelevantCultures()); }
		}

		private System.Collections.IDictionary consentsMap;
		private System.Collections.IDictionary ConsentsMap
		{
			get
			{
				if (consentsMap == null)
					consentsMap = new ListDictionary();
				return consentsMap;
			}
			set
			{
				consentsMap = value;
				consents = null; // reset accessor
			}
		}
		private IDictionary<CultureInfo, string> consents;
		/// <summary>
		/// The consent statements for various cultures that any respondent
		/// must agree to before taking the questionnaire.
		/// </summary>
		public virtual IDictionary<CultureInfo, string> Consents
		{
			get
			{
				if (consents == null)
					consents = new IDictionaryWrapper<CultureInfo, string>(ConsentsMap);
				return consents;
			}
		}
		/// <summary>
		/// The consent statement (for the active culture) that any respondent must agree to before taking the questionnaire.
		/// </summary>
		public string Consent
		{
			get
			{
				return GetMostRelevantLocalResource<string>(Consents);
			}
			set
			{
				Consents[Thread.CurrentThread.CurrentUICulture] = value;
			}
		}

		private System.Collections.IDictionary introductionsMap;
		private System.Collections.IDictionary IntroductionsMap
		{
			get
			{
				if (introductionsMap == null)
					introductionsMap = new ListDictionary();
				return introductionsMap;
			}
			set
			{
				introductionsMap = value;
				introductions = null; // reset accessor
			}
		}
		private IDictionary<CultureInfo, string> introductions;
		/// <summary>
		/// The introduction for various cultures that any respondent
		/// may read before taking the questionnaire.
		/// </summary>
		public virtual IDictionary<CultureInfo, string> Introductions
		{
			get
			{
				if (introductions == null)
					introductions = new IDictionaryWrapper<CultureInfo, string>(IntroductionsMap);
				return introductions;
			}
		}
		/// <summary>
		/// The consent statement (for the active culture) that any 
		/// respondent may read before taking the questionnaire.
		/// </summary>
		public string Introduction
		{
			get
			{
				return GetMostRelevantLocalResource<string>(Introductions);
			}
			set
			{
				Introductions[Thread.CurrentThread.CurrentUICulture] = value;
			}
		}

		private System.Collections.IDictionary closingStatementsMap;
		private System.Collections.IDictionary ClosingStatementsMap
		{
			get
			{
				if (closingStatementsMap == null)
					closingStatementsMap = new ListDictionary();
				return closingStatementsMap;
			}
			set
			{
				closingStatementsMap = value;
				closingStatements = null; // reset accessor
			}
		}
		private IDictionary<CultureInfo, string> closingStatements;
		/// <summary>
		/// The closing statement for various cultures that any respondent
		/// may read after taking the questionnaire.
		/// </summary>
		public virtual IDictionary<CultureInfo, string> ClosingStatements
		{
			get
			{
				if (closingStatements == null)
					closingStatements = new IDictionaryWrapper<CultureInfo, string>(ClosingStatementsMap);
				return closingStatements;
			}
		}
		/// <summary>
		/// The closing statement (for the active culture) that any 
		/// respondent may read after taking the questionnaire.
		/// </summary>
		public string ClosingStatement
		{
			get
			{
				return GetMostRelevantLocalResource<string>(ClosingStatements);
			}
			set
			{
				ClosingStatements[Thread.CurrentThread.CurrentUICulture] = value;
			}
		}

		private string css;
		/// <summary>
		/// Custom CSS to apply to this questionnaire when rendered as HTML.
		/// </summary>
		public string Css
		{
			get { return css; }
			set { css = value; }
		}

		private int groupOptionsFrequency = NAssess.Core.Engines.Evaluation.StandardHtml.GroupSelectOneOptionFrequencyDefault;
		/// <summary>
		/// The number of rows to appear between value label rows
		/// in a select one grid.
		/// </summary>
		public int GroupOptionsFrequency
		{
			get { return groupOptionsFrequency; }
			set { groupOptionsFrequency = value; }
		}
		private bool groupOptionsOnLeft = NAssess.Core.Engines.Evaluation.StandardHtml.GroupSelectOneOptionsOnLeftDefault;
		/// <summary>
		/// Whether grid options will be to the left of the question text.
		/// </summary>
		public bool GroupOptionsOnLeft
		{
			get { return groupOptionsOnLeft; }
			set { groupOptionsOnLeft = value; }
		}
		#endregion

		#region Operations
		/// <summary>
		/// Gets the list of cultures, in precedence order, that should be checked through
		/// when parsing the arrays of culture-specific attributes in this class.
		/// </summary>
		protected virtual CultureInfo[] GetRelevantCultures()
		{
			return CultureAssist.GetRelevantCultures(PrimaryCulture);
		}
		/// <summary>
		/// Gets the most relevant localized resource from a dictionary of
		/// culture-specific resources.
		/// </summary>
		/// <typeparam name="T">The type of resource being sought.</typeparam>
		/// <param name="dictionary">The dictionary to search.</param>
		/// <returns>The relevant resource, or default(T) if none found.</returns>
		protected T GetMostRelevantLocalResource<T>(IDictionary<CultureInfo, T> dictionary)
		{
			return CultureAssist.GetMostRelevantLocalResource<T>(dictionary, GetRelevantCultures());
		}
		public virtual Evaluation StartEvaluation()
		{
			Evaluation evaluation = new Evaluation();
			evaluation.Questionnaire = this;
			return evaluation;
		}
		/// <summary>
		/// Used for getting a professional-looking FO version of a questionnaire.
		/// </summary>
		/// <returns>A data stream containing the contents of the PDF file.</returns>
		public Stream ToFO()
		{
			return ToFO(false);
		}
		/// <summary>
		/// Used for getting a professional-looking FO version of a questionnaire.
		/// </summary>
		/// <param name="monochrome">
		/// Whether the printed questionnaire should be without color.
		/// </param>
		/// <returns>A data stream containing the contents of the PDF file.</returns>
		public Stream ToFO(bool monochrome)
		{
			return FOEngine.ToFO(monochrome);
		}
		/// <summary>
		/// Used for getting a professional-looking FO version of a questionnaire.
		/// </summary>
		public void ToFO(Stream outStream)
		{
			ToFO(outStream, false);
		}
		/// <summary>
		/// Used for getting a professional-looking FO version of a questionnaire.
		/// </summary>
		public void ToFO(Stream outStream, bool monochrome)
		{
			FOEngine.ToFO(outStream, false);
		}

		/// <summary>
		/// Used Pdfr getting a professional-looking PDF version of a questionnaire.
		/// </summary>
		/// <returns>A data stream containing the contents of the PDF file.</returns>
		public Stream ToPdf()
		{
			return ToPdf(false);
		}
		/// <summary>
		/// Used Pdfr getting a professional-looking PDF version of a questionnaire.
		/// </summary>
		/// <param name="monochrome">
		/// Whether the printed questionnaire should be without color.
		/// </param>
		/// <returns>A data stream containing the contents of the PDF file.</returns>
		public Stream ToPdf(bool monochrome)
		{
			if (PdfEngine == null) throw new InvalidOperationException(Resources.PdfEngineRequired);
			return PdfEngine.ToPdf(monochrome);
		}
		/// <summary>
		/// Used Pdfr getting a professional-looking PDF version of a questionnaire.
		/// </summary>
		public void ToPdf(Stream outStream)
		{
			ToPdf(outStream, false);
		}
		/// <summary>
		/// Used Pdfr getting a professional-looking PDF version of a questionnaire.
		/// </summary>
		public void ToPdf(Stream outStream, bool monochrome)
		{
			if (PdfEngine == null) throw new InvalidOperationException(Resources.PdfEngineRequired);
			PdfEngine.ToPdf(outStream, false);
		}

		/// <summary>
		/// Fills the provided data structure with meta data on the variable specified within that structure.
		/// Meta data includes the variable label and value labels.
		/// </summary>
		/// <param name="var">
		/// The data structure that is initialized to the variable to get meta data on.
		/// </param>
		/// <remarks>
		/// Useful as a callback function to pass to SPSS.net for exporting meta data filled SAV files.
		/// </remarks>
		public void FillMetadata(VarMetaData var)
		{
			if (var == null) throw new ArgumentNullException("var");

			// Find the node for the variable
			XPathNavigator nVar = Codebook.FindVariableByName(var.Name);
			if (nVar == null) throw new ArgumentOutOfRangeException("var", var.Name,
								  "Could not find a variable in this codebook to match the desired name.");

			// Set the variable label
			var.Label = Ddi.SelectSingleNode("ddi:labl", nVar).Value;

			// Set the value labels
			XPathNodeIterator nlValueLabels = Ddi.Select("ddi:catgry", nVar);
			while (nlValueLabels.MoveNext())
			{
				XPathNavigator nValueLabel = nlValueLabels.Current;
				int val = Convert.ToInt32(Ddi.SelectSingleNode("ddi:catValu", nValueLabel).Value);
				string label = Ddi.SelectSingleNode("ddi:labl", nValueLabel).Value;
				var[val] = label;
			}
		}
		#endregion

		#region Data exporting functions
		/// <summary>
		/// Exports all evaluations collected by the <see cref="Questionnaire"/>.
		/// </summary>
		/// <returns>
		/// An enumerable list of DataRows filled with all the variables, scales 
		/// defined by the <see cref="Codebook"/> and all collected data.
		/// </returns>
		/// <exception cref="InvalidOperationException">
		/// Thrown when called before a <see cref="Codebook"/> has been set.
		/// </exception>
		/// <remarks>
		/// This method returns an enumerable list of rows rather than a fully
		/// filled <see cref="DataTable"/> because large datasets would run out
		/// of memory.
		/// </remarks>
		public IEnumerable<DataRow> GetDataTable(INAssessManager mgr)
		{
			if (Codebook == null) throw new InvalidOperationException("Codebook must be provided before a DataTable can be produced.");
			DataTable table = Codebook.BuildDataTable();
			AddExtraDataTableColumns(mgr, table);

			foreach (Evaluation eval in mgr.GetDataSetEvaluations(Id))
			{
				DataRow dr = table.NewRow();
				foreach (DataColumn dc in table.Columns)
					if (!AddExtraDataTableData(mgr, dr, dc, eval) &&
						!dc.ReadOnly) // avoid explicitly setting calculated columns (like scales)
						dr[dc] = NullToDBNullFilter(eval.Responses[dc.ColumnName]);
				// don't add to table, just stream out
				yield return dr;
			}
		}
		protected static object NullToDBNullFilter(object value)
		{
			return value == null ? DBNull.Value : value;
		}
		/// <summary>
		/// Adds any extra appropriate DataTable columns besides the raw 
		/// collected data to the DataTable structure before being filled.
		/// </summary>
		protected virtual void AddExtraDataTableColumns(INAssessManager mgr, DataTable dataTable)
		{
			if (mgr == null) throw new ArgumentNullException("mgr");
			if (dataTable == null) throw new ArgumentNullException("dataTable");

			// Add the primary key column
			DataColumn dcId = dataTable.Columns.Add("EvaluationId", typeof(int)); // The EvaluationId column
			dataTable.PrimaryKey = new DataColumn[] { dcId };
		}
		/// <summary>
		/// Adds any extra data to a given column of a given row that may not come out
		/// of the Evaluation.Responses dictionary.
		/// </summary>
		/// <returns>
		/// True if some value was inserted.  False otherwise.
		/// </returns>
		protected virtual bool AddExtraDataTableData(INAssessManager mgr, DataRow row, DataColumn column, Evaluation evaluation)
		{
			if (mgr == null) throw new ArgumentNullException("mgr");
			if (row == null) throw new ArgumentNullException("row");
			if (column == null) throw new ArgumentNullException("column");
			if (evaluation == null) throw new ArgumentNullException("evaluation");

			switch (column.ColumnName)
			{
				case "EvaluationId":
					row[column] = evaluation.Id;
					return true;
				default:
					return false;
			}
		}

		#endregion

		#region Collections
		private System.Collections.IDictionary codebooksSet;
		private System.Collections.IDictionary CodebooksSet
		{
			get
			{
				if (codebooksSet == null) codebooksSet = new ListDictionary();
				return codebooksSet;
			}
			set
			{
				codebooksSet = value;
				codebooks = null;
			}
		}
		private IDictionary<CultureInfo, Ddi> codebooks;
		public virtual IDictionary<CultureInfo, Ddi> Codebooks
		{
			get
			{
				if (codebooks == null)
					codebooks = new IDictionaryWrapper<CultureInfo, Ddi>(CodebooksSet);
				return codebooks;
			}
		}

		private System.Collections.IList assessmentsList;
		protected System.Collections.IList AssessmentsList
		{
			get
			{
				if (assessmentsList == null) assessmentsList = new System.Collections.ArrayList();
				return assessmentsList;
			}
			set
			{
				assessmentsList = value;
				assessments = null;
			}
		}
		private TwoWayList<Assessment, Questionnaire> assessments;
		/// <summary>
		/// Report description storage
		/// </summary>
		public virtual TwoWayList<Assessment, Questionnaire> Assessments
		{
			get
			{
				if (assessments == null)
					assessments = new TwoWayList<Assessment, Questionnaire>(this, AssessmentsList);
				return assessments;
			}
		}
		#endregion

		#region IReferenceableParent<Assessment,Questionnaire> Members

		ITwoWayCollection<Assessment, Questionnaire> IReferenceableParent<Assessment, Questionnaire>.ContainingCollection
		{
			get { return Assessments; }
		}

		#endregion


		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Questionnaire other = obj as Questionnaire;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}


	}
}
