using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using System.Net.Mail;

namespace NAssess.Core.FollowUp
{
	public abstract class Email
	{
		private MailAddress from;
		/// <summary>
		/// The person from whom the invitation email appears to be from.
		/// </summary>
		public virtual MailAddress From
		{
			get { return from; }
			set { from = value; }
		}
		private string persistFromAddress
		{
			get { return From != null ? From.Address : null; }
			set
			{
				if (value == null)
				{
					From = null;
					return;
				}
				if (From != null && From.DisplayName != null)
					From = new MailAddress(value, From.DisplayName);
				else
					From = new MailAddress(value);
			}
		}
		private string persistFromDisplayName
		{
			get { return From != null ? From.DisplayName : null; }
			set
			{
				if (value == null)
				{
					From = null;
					return;
				}
				if (From != null)
					From = new MailAddress(From.Address, value);
				else
					From = new MailAddress("sometemporary@email.com", value);
			}
		}
		private string body;
		/// <summary>
		/// The content of the email to send as the invitation.
		/// </summary>
		/// <remarks>
		/// Placeholders for various fill in values can and should be placed
		/// in the email body to customize the invite for each user and 
		/// allow access to the follow-up study.
		/// <list>
		/// <listheader><term>Placeholder</term><value>Meaning</value></listheader>
		/// <term>{0}<item>Salutation</item></term>
		/// <term>{1}<item>Link</item></term>
		/// </list>
		/// </remarks>
		public virtual string Body
		{
			get { return body; }
			set { body = value; }
		}

		private bool isHtml;
		/// <summary>
		/// Indicates whether the body of the email message is in Html or clear text.
		/// </summary>
		public virtual bool IsHtml
		{
			get { return isHtml; }
			set { isHtml = value; }
		}

		private string subject;

		public virtual string Subject
		{
			get { return subject; }
			set { subject = value; }
		}
		private int id;

		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}

		/// <summary>
		/// Sends this email to a recipient.
		/// </summary>
		/// <param name="recipient">The recipient name and email address for this invite.</param>
		/// <param name="args">The <see cref="String.Format(string, object[])"/> arguments to pass into the email before sending.</param>
		public virtual void Send(MailAddress recipient, params object[] args)
		{
			if (recipient == null) throw new ArgumentNullException("recipient");
			if (args == null) throw new ArgumentNullException("args");
			object[] argsWithName = new object[args.Length + 1];
			argsWithName[0] = recipient.DisplayName;
			args.CopyTo(argsWithName, 1);

			MailMessage message = new MailMessage(From, recipient);
			message.Subject = Subject;
			message.Body = string.Format(Body, argsWithName);
			message.IsBodyHtml = IsHtml;

			SmtpClient smtp = new SmtpClient();
			smtp.Send(message);
		}

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Email other = obj as Email;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		/// <summary>
		/// Gets a hash code for this object.
		/// </summary>
		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString(CultureInfo.InvariantCulture);
			return stringRepresentation.GetHashCode();
		}

				
	}
}
