using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;

using Iesi.Collections;
using MyNHibernateContrib.TwoWayCollections;

using NAssess.Core.Evaluating;

namespace NAssess.Core.FollowUp
{
	public class EvaluationTracker
		: IParentReferencingNode<EvaluationTracker, Study>,
		  IParentReferencingNode<EvaluationTracker, Evaluation>,
		  IReferenceableParent<EventTracker, EvaluationTracker>
	{
		private int id;

		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}
		private Study study;
		/// <summary>
		/// The study this evaluation tracker is associated with.
		/// </summary>
		public virtual Study Study
		{
			get { return study; }
			private set { study = value; }
		}
		private Evaluation evaluation;
		/// <summary>
		/// The original evaluation that started the study on this respondent.
		/// </summary>
		public virtual Evaluation Evaluation
		{
			get { return evaluation; }
			private set { evaluation = value; }
		}
		private System.Collections.IList trackedEventsList;
		private System.Collections.IList TrackedEventsList
		{
			get
			{
				if (trackedEventsList == null)
					trackedEventsList = new System.Collections.ArrayList();
				return trackedEventsList;
			}
			set
			{
				trackedEventsList = value;
				trackedEvents = null; // reset accessor
			}
		}
		private TwoWayList<EventTracker, EvaluationTracker> trackedEvents;
		/// <summary>
		/// The events this respondent has been scheduled for and/or participated in.
		/// </summary>
		public virtual TwoWayList<EventTracker, EvaluationTracker> TrackedEvents
		{
			get
			{
				if (trackedEvents == null)
					trackedEvents = new TwoWayList<EventTracker, EvaluationTracker>(this, TrackedEventsList);
				return trackedEvents;
			}
		}
		private bool isClosed;
		/// <summary>
		/// Whether the study has been closed for this evaluation.
		/// </summary>
		public virtual bool IsClosed
		{
			get { return isClosed; }
			set { isClosed = value; }
		}

		protected internal virtual void AddNextEvent()
		{
			if (Evaluation == null || Study == null)
				throw new InvalidOperationException("Cannot add first event until Evaluation and Study properties are set.");
			if (TrackedEvents.Count > 0 && !TrackedEvents[TrackedEvents.Count - 1].CompletedDate.HasValue)
				throw new InvalidOperationException("Last added event must be completed before next event can be added.");
			if (TrackedEvents.Count >= Study.TotalEventIterations)
				throw new InvalidOperationException("Study has no more events.");
			StudyEvent nextEvent = Study.GetEvent(TrackedEvents.Count);

			TrackedEvents.Add(new EventTracker(nextEvent));
		}


		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			EvaluationTracker other = obj as EvaluationTracker;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}

				

		#region IParentReferencingNode<EvaluationTracker, Study> Members

		Study IParentReferencingNode<EvaluationTracker, Study>.Parent
		{
			get { return study; }
			set { study = value; }
		}

		#endregion

		#region IParentReferencingNode<EvaluationTracker,Evaluation> Members

		Evaluation IParentReferencingNode<EvaluationTracker, Evaluation>.Parent
		{
			get { return evaluation; }
			set { evaluation = value; }
		}

		#endregion

		#region IReferenceableParent<EventTracker,EvaluationTracker> Members

		ITwoWayCollection<EventTracker, EvaluationTracker> IReferenceableParent<EventTracker, EvaluationTracker>.ContainingCollection
		{
			get { return TrackedEvents; }
		}

		#endregion

		/// <summary>
		/// The event currently open for participation for this study.  
		/// Null if no event is active.
		/// </summary>
		public EventTracker ActiveEvent
		{
			get
			{
				if (TrackedEvents.Count == 0) return null;
				EventTracker e = TrackedEvents[TrackedEvents.Count - 1];
				if (e.State == EventTrackerState.Active)
					return e;
				else
					return null;
			}
		}
	}
}
