using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;
using System.Net.Mail;

using Iesi.Collections;
using MyNHibernateContrib.GenericCollections;
using MyNHibernateContrib.TwoWayCollections;

using NAssess.Core.Evaluating;

namespace NAssess.Core.FollowUp
{
	/// <summary>
	/// Tracks the progress of a single evaluation through the follow-up study.
	/// </summary>
	public class EventTracker
		: IParentReferencingNode<EventTracker, EvaluationTracker>,
		IParentReferencingNode<EventTracker, IncentiveCollection>,
		IReferenceableParent<Incentive, EventTracker>
	{
		protected EventTracker() { }
		public EventTracker(StudyEvent studyEvent)
		{
			StudyEvent = studyEvent;
		}

		private int id;

		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}
		private EvaluationTracker evaluationTracker;

		public virtual EvaluationTracker EvaluationTracker
		{
			get { return evaluationTracker; }
			private set { evaluationTracker = value; }
		}
		private StudyEvent studyEvent;

		public virtual StudyEvent StudyEvent
		{
			get { return studyEvent; }
			private set { studyEvent = value; }
		}
		private int invitesSent;
		/// <summary>
		/// The number of invites sent to the respondent for this specific event.
		/// </summary>
		public virtual int InvitesSent
		{
			get { return invitesSent; }
			protected set { invitesSent = value; }
		}
		private DateTime? lastInviteSentDate;
		/// <summary>
		/// The time the last invite was sent out.
		/// </summary>
		public virtual DateTime? LastInviteSentDate
		{
			get { return lastInviteSentDate; }
			protected set { lastInviteSentDate = value; }
		}

		/// <summary>
		/// Gets the date that this event should take place.
		/// </summary>
		public DateTime ScheduledDate
		{
			get
			{
				DateTime? reference;

				if (StudyEvent.ScheduleFrom == ScheduleFrom.OriginalEvaluation || PreviousEvent == null)
					reference = EvaluationTracker.Evaluation.EndDate;
				else
					switch (StudyEvent.ScheduleFrom)
					{
						case ScheduleFrom.LastScheduledEvent:
							reference = PreviousEvent.ScheduledDate;
							break;
						case ScheduleFrom.LastCompletedEvent:
							reference = PreviousEvent.CompletedDate;
							break;
						default:
							throw new ApplicationException();
					}

				if (!reference.HasValue)
					throw new ApplicationException("Reference date has not been set yet.");

				return reference.Value + StudyEvent.WaitTime;
			}
		}

		/// <summary>
		/// Gets the index of this events among the events being tracked.
		/// </summary>
		public int EventIndex
		{
			get
			{
				if (EvaluationTracker == null)
					throw new InvalidOperationException("Unable to determine EventIndex because EvaluationTracker has not been set.");
				int thisEventIndex = EvaluationTracker.TrackedEvents.IndexOf(this);
				if (thisEventIndex < 0)
					throw new ApplicationException("Unable to find self in tracked events list.");
				return thisEventIndex;
			}
		}

		/// <summary>
		/// Gets the last event in this study for this respondent, or null if this is the first event.
		/// </summary>
		protected EventTracker PreviousEvent
		{
			get
			{
				int eventIndex = EventIndex;
				if (eventIndex == 0) return null;
				return EvaluationTracker.TrackedEvents[eventIndex - 1];
			}
		}

		private DateTime? respondedDate;
		/// <summary>
		/// The date the respondent began this event.
		/// </summary>
		/// <remarks>
		/// This event can be considered responded to either by the user
		/// clicking a link in an email invitation or a link on the web
		/// site when the user is visiting.
		/// </remarks>
		public virtual DateTime? RespondedDate
		{
			get { return respondedDate; }
			protected set { respondedDate = value; }
		}
		/// <summary>
		/// The date that the user completed this event.
		/// </summary>
		public DateTime? CompletedDate
		{
			get { return EventEvaluation != null ? EventEvaluation.EndDate : (DateTime?)null; }
		}

		/// <summary>
		/// Gets whether the response qualifies for a participation incentive.
		/// </summary>
		/// <remarks>
		/// The value of this property remains the same even after a participation
		/// incentive is sent.
		/// </remarks>
		public bool QualifiesForParticipationIncentive
		{
			get
			{
				return StudyEvent != null && EventEvaluation != null &&
					StudyEvent.ParticipatingIncentives != null &&
					EventEvaluation.AmountCompleted >= StudyEvent.ParticipationIncentiveMinimumProportion &&
					EventEvaluation.Responses.Count >= StudyEvent.ParticipationIncentiveMinimumResponses;
			}
		}

		public bool ParticipationIncentiveSent
		{
			get
			{
				if (StudyEvent == null || StudyEvent.ParticipatingIncentives == null)
					return false;
				// WARNING: this recognizes participation incentives only if
				// they are in the participation incentive pool that is currently
				// selected.  If the pool were ever switched, all issued incentives
				// would no longer be recognized as being participation incentives.
				// (I am not sure the best way to fix this at this time.)
				foreach (Incentive incentive in Incentives)
					if (StudyEvent.ParticipatingIncentives.Contains(incentive))
						return true;

				return false;
			}
		}

		private IncentiveCollection lotteryPool;
		/// <summary>
		/// Gets/sets the incentive pool that this <see cref="EventTracker"/> belongs
		/// to for some upcoming drawing.
		/// </summary>
		public virtual IncentiveCollection LotteryPool
		{
			get { return lotteryPool; }
			set { lotteryPool = value; }
		}

		#region Collections
		private ISet incentivesSet;
		private ISet IncentivesSet
		{
			get
			{
				if (incentivesSet == null)
					incentivesSet = new HashedSet();
				return incentivesSet;
			}
			set
			{
				incentivesSet = value;
				incentives = null; // reset accessor
			}
		}
		private TwoWaySet<Incentive, EventTracker> incentives;
		/// <summary>
		/// The set of rewards sent out to the participant for participating 
		/// in this particular event.
		/// </summary>
		public virtual TwoWaySet<Incentive, EventTracker> Incentives
		{
			get
			{
				if (incentives == null)
					incentives = new TwoWaySet<Incentive, EventTracker>(this, IncentivesSet);
				return incentives;
			}
		}

		#endregion
		/// <summary>
		/// Gets the state of this event, whether it be before, during or after the event.
		/// </summary>
		public EventTrackerState State
		{
			get
			{
				if (CompletedDate.HasValue)
					return EventTrackerState.Completed;
				if (DateTime.Now < ScheduledDate)
					return EventTrackerState.Dormant;
				return EventTrackerState.Active;
			}
		}

		/// <summary>
		/// Gets the study this event tracker is associated with.
		/// </summary>
		public Study Study
		{
			get { return (EvaluationTracker != null) ? EvaluationTracker.Study : null; }
		}

		/// <summary>
		/// Sends an invitation to the respondent to participate in this event.
		/// </summary>
		/// <param name="recipient">The name and email address of the person receiving the invite.</param>
		/// <param name="link">The link the respondent can use to return to a web site and begin the evaluation.</param>
		/// <remarks>
		/// This method sends the invite regardless of the value of 
		/// <see cref="InviteShouldBeSent"/>, so this value should
		/// be checked before calling this method.
		/// </remarks>
		public void SendInvite(MailAddress recipient, Uri link)
		{
			if (recipient == null) throw new ArgumentNullException("recipient");
			if (link == null) throw new ArgumentNullException("link");

			string inviteKey = Study.GetInvitationKey(this);
			SmtpClient smtp = new SmtpClient();
			Study.InvitationEmails[inviteKey].Send(recipient, link);
			InvitesSent++;
			LastInviteSentDate = DateTime.Now;
		}

		public Incentive SendParticipationIncentiveIfAppropriate(INAssessManager mgr)
		{
			if (mgr == null) throw new ArgumentNullException("mgr");
			if (!QualifiesForParticipationIncentive)
				return null; // This event does not qualify for a participation incentive.
			if (ParticipationIncentiveSent)
				return null; // A participation incentive has already been sent.

			Debug.Assert(StudyEvent != null);
			Debug.Assert(StudyEvent.ParticipatingIncentives != null);

			if (StudyEvent.ParticipatingIncentives.Style == IncentiveStyles.Guaranteed)
				return SendDigitalIncentive(StudyEvent.ParticipatingIncentives, mgr);
			else if (StudyEvent.ParticipatingIncentives.Style == IncentiveStyles.Lottery)
			{
				LotteryPool = StudyEvent.ParticipatingIncentives;
				return null;
			}
			else throw new NotSupportedException("The StudyEvent.ParticipatingIncentives.IncentiveStyle selected ("
				+ StudyEvent.ParticipatingIncentives.Style.ToString() + ") is not supported.");
		}

		/// <summary>
		/// Sends a digital incentive to a respondent via email.
		/// </summary>
		/// <returns>
		/// The digital incentive that was sent.
		/// </returns>
		public DigitalIncentive SendDigitalIncentive(IncentiveCollection incentivePool, INAssessManager mgr)
		{
			if (incentivePool == null) throw new ArgumentNullException("incentivePool");
			if (mgr == null) throw new ArgumentNullException("mgr");

			DigitalIncentive incentive = incentivePool.FindUnassignedIncentive(mgr) as DigitalIncentive;
			if (incentive == null)
				throw new InvalidOperationException("No digital incentives found/left.");

			Incentives.Add(incentive);
			incentivePool.Email.Send(mgr.GetMailAddressFor(this), incentive);
			return incentive;
		}

		/// <summary>
		/// Assigns a stock incentive (i.e. a DVD) to a respondent for snail mailing.
		/// </summary>
		/// <param name="recipientAddress">
		/// The postal mailing address of the recipient.
		/// </param>
		/// <param name="incentiveType">
		/// The type of incentive to send.
		/// </param>
		/// <param name="mgr">
		/// The object persistence manager to query for an available incentive.
		/// </param>
		/// <returns>
		/// The stock incentive that was sent.
		/// </returns>
		public StockIncentive AssignStockIncentive(string recipientAddress,
			IncentiveCollection incentivePool, INAssessManager mgr)
		{
			if (recipientAddress == null || recipientAddress.Length == 0)
				throw new ArgumentNullException("recipientAddress");
			if (incentivePool == null) throw new ArgumentNullException("incentivePool");

			StockIncentive incentive = incentivePool.FindUnassignedIncentive(mgr) as StockIncentive;
			if (incentive == null)
				throw new InvalidOperationException("No stock incentives found/left.");

			Incentives.Add(incentive);
			incentive.ShippingAddress = recipientAddress;
			return incentive;
		}

		/// <summary>
		/// Starts a new evaluation consistent for this stage in the study.
		/// </summary>
		public virtual Evaluation Start()
		{
			if (State != EventTrackerState.Active)
				throw new InvalidOperationException("This evaluation tracker is not in the Active state.  Is this follow-up study being activated too early?");
			if (EventEvaluation != null)
				throw new InvalidOperationException("An event evaluation has already been started.");
			if (StudyEvent.OnlyWhileIncentivesLast &&
				(StudyEvent.ParticipatingIncentives == null || StudyEvent.ParticipatingIncentives.Count == 0))
				throw new InvalidOperationException("Study event is out of incentives, and is closed.");

			if (!RespondedDate.HasValue) RespondedDate = DateTime.Now;

			return EventEvaluation = StudyEvent.Questionnaire.StartEvaluation();
		}

		private Evaluation eventEvaluation;
		public virtual Evaluation EventEvaluation
		{
			get { return eventEvaluation; }
			protected set
			{
				if (eventEvaluation != null && eventEvaluation.FollowUpEvent == this)
				{
					eventEvaluation.Completed -= new EventHandler(EventEvaluation_Completed);
					eventEvaluation.FollowUpEvent = null;
				}
				if (value != null)
				{
					value.FollowUpEvent = this;
					value.Completed += new EventHandler(EventEvaluation_Completed);
				}
				eventEvaluation = value;
			}
		}

		private void EventEvaluation_Completed(object sender, EventArgs e)
		{
			// only pay attention to our EventEvaluation.
			if (sender != EventEvaluation)
			{
				((Evaluation)sender).Completed -= new EventHandler(EventEvaluation_Completed);
				return;
			}
			Debug.Assert(State == EventTrackerState.Completed);

			// End study early if appropriate
			if (Study.StillQualifies(EvaluationTracker))
			{
				if (// if there are more events defined
					EventIndex < Study.TotalEventIterations - 1 &&
					// and this is the last one that has been created for this tracker
					EventIndex == EvaluationTracker.TrackedEvents.Count - 1
					)
					EvaluationTracker.AddNextEvent();
			}
			else
				EvaluationTracker.IsClosed = true;

			// Notify study that an event has just been completed.
			Study.OnEventCompleted(this);
		}

		/// <summary>
		/// Gets whether it is time to send an invite for this event.
		/// </summary>
		public bool InviteShouldBeSent
		{
			get
			{
				return
					State == EventTrackerState.Active &&	// event must have started and not ended
					!RespondedDate.HasValue &&				// user has not already responded
					InvitesSent < StudyEvent.MaxInvites &&	// # of invites already sent below maximum
					(!LastInviteSentDate.HasValue			// and enough time elapsed since last invite
					|| LastInviteSentDate < DateTime.Now - StudyEvent.TimeBetweenInvites);
			}
		}

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			EventTracker other = obj as EventTracker;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}

		#region IParentReferencingNode<EventTracker, EvaluationTracker> Members

		EvaluationTracker IParentReferencingNode<EventTracker, EvaluationTracker>.Parent
		{
			get { return evaluationTracker; }
			set { evaluationTracker = value; }
		}

		#endregion

		#region IReferenceableParent<Incentive,EventTracker> Members

		ITwoWayCollection<Incentive, EventTracker> IReferenceableParent<Incentive, EventTracker>.ContainingCollection
		{
			get { return Incentives; }
		}

		#endregion

		#region IParentReferencingNode<EventTracker,IncentivePool> Members

		IncentiveCollection IParentReferencingNode<EventTracker, IncentiveCollection>.Parent
		{
			get { return LotteryPool; }
			set { LotteryPool = value; }
		}

		#endregion
	}

	public enum EventTrackerState
	{
		/// <summary>
		/// This is an event scheduled for the future, and the respondent is not invited to participate at this time.
		/// </summary>
		Dormant,
		/// <summary>
		/// This event is now active, and the user is welcome to participate.
		/// </summary>
		Active,
		/// <summary>
		/// The respondent has completed a response.
		/// </summary>
		Completed,
		/// <summary>
		/// This event ended before a response was completed.
		/// </summary>
		Expired,
	}
}
