using System;
using System.Collections.Generic;
using System.Text;

using Iesi.Collections;
using MyNHibernateContrib.TwoWayCollections;

namespace NAssess.Core.FollowUp
{
	public class IncentiveCollection
		: IParentReferencingNode<IncentiveCollection, Study>,
		IParentReferencingNode<IncentiveCollection, IncentiveEmail>,
		IReferenceableParent<EventTracker, IncentiveCollection>,
		IReferenceableParent<Incentive, IncentiveCollection>,
		IEnumerable<Incentive>
	{
		protected static readonly log4net.ILog Log =
				log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

		#region Attributes
		private int id;

		public virtual int Id
		{
			get { return id; }
			set { id = value; }
		}

		private Study study;
		/// <summary>
		/// Gets the <see cref="Study"/> that this <see cref="IncentivePool"/> 
		/// is associated with.
		/// </summary>
		public virtual Study Study
		{
			get { return study; }
			private set { study = value; }
		}

		/// <summary>
		/// Gets the name of this incentive pool, 
		/// as it appears as a key in the Study.IncentivePools dictionary.
		/// </summary>
		public virtual string Name
		{
			get
			{
				if (Study == null) return null;
				foreach (KeyValuePair<string, IncentiveCollection> pair in Study.IncentivePools)
					if (pair.Value.Equals(this)) return pair.Key;
				throw new ApplicationException("Unable to find pool in Study.IncentivePools dictionary.");
			}
			set
			{
				if (Study == null) throw new ApplicationException("This pool must first be associated with a Study.");
				Study study = Study;
				study.IncentivePools.Remove(Name);
				study.IncentivePools[value] = this;
			}
		}

		private IncentiveStyles style;
		/// <summary>
		/// Whether incentives are guaranteed to be mailed out or respondents will just be entered in a drawing.
		/// </summary>
		public virtual IncentiveStyles Style
		{
			get { return style; }
			set { style = value; }
		}

		private int lotteryOdds = 2; // default to a 1:2 chance
		/// <summary>
		/// The size of the pool when a respondent is drawn randomly and sent an incentive.  
		/// Applicable only if <see cref="IncentiveStyle"/> is set to 
		/// <see cref="NAssess.IncentiveStyle.Lottery"/>.
		/// </summary>
		public virtual int LotteryOdds
		{
			get { return lotteryOdds; }
			set
			{
				if (value < 2) throw new ArgumentOutOfRangeException("LotteryOddsAgainst", value, "Must be at least 2.");
				lotteryOdds = value;
			}
		}

		private ISet lotteryBowlSet;
		/// <summary>
		/// The NHibernate <see cref="ISet"/> that backs the <see cref="LotteryBowl"/> collection.
		/// </summary>
		protected ISet LotteryBowlSet
		{
			get
			{
				if (lotteryBowlSet == null) lotteryBowlSet = new Iesi.Collections.HashedSet();
				return lotteryBowlSet;
			}
			set
			{
				lotteryBowlSet = value;
				lotteryBowl = null;
			}
		}

		private TwoWaySet<EventTracker, IncentiveCollection> lotteryBowl;
		/// <summary>
		/// The collection of <see cref="EventTracker"/> objects that should be 
		/// chosen from in the next drawing.
		/// </summary>
		public virtual TwoWaySet<EventTracker, IncentiveCollection> LotteryBowl
		{
			get
			{
				if (lotteryBowl == null)
					lotteryBowl = new TwoWaySet<EventTracker, IncentiveCollection>(this, LotteryBowlSet);
				return lotteryBowl;
			}
		}

		private string outOfIncentivesMessage;
		/// <summary>
		/// The message to display to a new respondent when the incentives have run out.
		/// </summary>
		public virtual string OutOfIncentivesMessage
		{
			get { return outOfIncentivesMessage; }
			set { outOfIncentivesMessage = value; }
		}

		private ISet incentivesSet;
		/// <summary>
		/// The NHibernate <see cref="ISet"/> that backs the <see cref="Incentives"/> collection.
		/// </summary>
		protected ISet IncentivesSet
		{
			get
			{
				if (incentivesSet == null) incentivesSet = new Iesi.Collections.HashedSet();
				return incentivesSet;
			}
			set
			{
				incentivesSet = value;
				incentives = null;
			}
		}

		private TwoWaySet<Incentive, IncentiveCollection> incentives;
		/// <summary>
		/// A set of Incentives.
		/// </summary>
		protected virtual TwoWaySet<Incentive, IncentiveCollection> Incentives
		{
			get
			{
				if (incentives == null)
					incentives = new TwoWaySet<Incentive, IncentiveCollection>(this, IncentivesSet);
				return incentives;
			}
		}

		private IncentiveEmail email;
		/// <summary>
		/// The email to be used to mail out these incentives.
		/// </summary>
		/// <remarks>
		/// If these are stock incentives, the email may simply
		/// notify the user of pending shipment, or it may be null.
		/// </remarks>
		public virtual IncentiveEmail Email
		{
			get { return email; }
			set { email = value; }
		}

		public int Count
		{
			get { return Incentives.Count; }
		}
		public bool HasUnassignedIncentivesLeft(INAssessManager mgr)
		{
			return FindUnassignedIncentive(mgr) != null;
		}
		/// <summary>
		/// Gets whether stock incentives exist in the Incentives collection.
		/// </summary>
		public bool HasStockIncentives
		{
			get
			{
				foreach (Incentive i in Incentives)
					return i is StockIncentive;
				return false;
			}
		}
		/// <summary>
		/// Gets whether digital incentives exist in the Incentives collection.
		/// </summary>
		public bool HasDigitalIncentives
		{
			get
			{
				foreach (Incentive i in Incentives)
					return i is DigitalIncentive;
				return false;
			}
		}
		public bool IsTimeToDrawForLottery
		{
			get
			{
				return Style == IncentiveStyles.Lottery
					&& LotteryBowl.Count >= LotteryOdds;
			}
		}
		#endregion

		#region Operations
		/// <summary>
		/// Adds some incentive to this pool of incentives.
		/// </summary>
		/// <param name="incentive">
		/// The incentive to add.
		/// </param>
		/// <remarks>
		/// An <see cref="IncentivePool"/> may only contain one type of
		/// incentive.  Adding 
		/// </remarks>
		/// <exception cref="ArgumentException">
		/// Thrown when the new incentive does not match the type of
		/// incentives that already belong to the pool.
		/// </exception>
		public void Add(Incentive incentive)
		{
			if (incentive == null) throw new ArgumentNullException("incentive");

			// Make sure that this incentive's type matches the type of
			// the rest of the pool, if any are already there.
			if (Incentives.Count > 0)
			{
				// TODO: code here
			}

			Incentives.Add(incentive);
		}
		public void Remove(Incentive incentive)
		{
			if (incentive == null) throw new ArgumentNullException("incentive");
			Incentives.Remove(incentive);
		}
		public bool Contains(Incentive incentive)
		{
			if (incentive == null) throw new ArgumentNullException("incentive");
			return Incentives.Contains(incentive);
		}
		/// <summary>
		/// Returns an unassigned incentive from the pool.
		/// </summary>
		/// <returns>
		/// The first unassigned incentive, or null if no available ones are found.
		/// </returns>
		public Incentive FindUnassignedIncentive(INAssessManager mgr)
		{
			if (mgr == null) throw new ArgumentNullException("mgr");
			return mgr.FindUnassignedIncentive(this);
		}
		/// <summary>
		/// Chooses a random <see cref="EventTracker"/> from 
		/// <see cref="LotteryBowl"/> and issues the incentive to
		/// the winner.
		/// </summary>
		public void DrawForLottery(INAssessManager manager)
		{
			if (Style != IncentiveStyles.Lottery)
				throw new InvalidOperationException("IncentiveStyle must be set to Lottery.");
			if (LotteryBowl.Count < LotteryOdds)
				Log.WarnFormat("Drawing for lottery before odds match.  IncentivePool.Id: {0}, LotteryBowl.Count: {1}, LotteryOdds: {2}.",
					Id, LotteryBowl.Count, LotteryOdds);
			if (LotteryBowl.Count > LotteryOdds)
				Log.WarnFormat("Drawing for lottery after odds match.  IncentivePool.Id: {0}, LotteryBowl.Count: {1}, LotteryOdds: {2}.",
					Id, LotteryBowl.Count, LotteryOdds);
			int randomWinnerIndex = (new Random()).Next(LotteryBowl.Count);
			EventTracker winner = null;
			int index = 0;
			foreach (EventTracker tracker in LotteryBowl)
			{
				tracker.LotteryPool = null; // Only in drawing once, and this was it.
				if (index++ == randomWinnerIndex)
					winner = tracker; // tag the winner
			}
			if (winner == null)
				throw new ApplicationException("Somehow no winner was picked during the drawing.");
			winner.SendDigitalIncentive(this, manager);
		}
		#endregion

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			IncentiveCollection other = obj as IncentiveCollection;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}

		#region IParentReferencingNode<IncentivePool,Study> Members

		Study IParentReferencingNode<IncentiveCollection, Study>.Parent
		{
			get { return Study; }
			set { Study = value; }
		}

		#endregion

		#region IEnumerable<Incentive> Members

		public IEnumerator<Incentive> GetEnumerator()
		{
			return Incentives.GetEnumerator();
		}

		#endregion

		#region IEnumerable Members

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return this.GetEnumerator();
		}

		#endregion

		#region IParentReferencingNode<IncentivePool,IncentiveEmail> Members

		IncentiveEmail IParentReferencingNode<IncentiveCollection, IncentiveEmail>.Parent
		{
			get { return Email; }
			set { Email = value; }
		}

		#endregion

		#region IReferenceableParent<EventTracker,IncentivePool> Members

		ITwoWayCollection<EventTracker, IncentiveCollection> IReferenceableParent<EventTracker, IncentiveCollection>.ContainingCollection
		{
			get { return LotteryBowl; }
		}

		#endregion

		#region IReferenceableParent<Incentive,IncentivePool> Members

		ITwoWayCollection<Incentive, IncentiveCollection> IReferenceableParent<Incentive, IncentiveCollection>.ContainingCollection
		{
			get { return Incentives; }
		}

		#endregion
	}
}
