using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;

using Iesi.Collections;
using MyNHibernateContrib.TwoWayCollections;
using MyNHibernateContrib.GenericCollections;

using NAssess.Core.Evaluating;

namespace NAssess.Core.FollowUp
{
	// This used to be an abstract class, but NHibernate's Castle.DynamicProxy 
	// would throw exceptions, so it was changed to this.
	public class Study
		: IReferenceableParent<StudyEvent, Study>,
		IReferenceableParent<EvaluationTracker, Study>,
		IReferenceableParentDictionary<string, InvitationEmail, Study>,
		IReferenceableParentDictionary<string, IncentiveCollection, Study>
	{
		private int id;
		/// <summary>
		/// The unique identifier for this study.
		/// </summary>
		public virtual int Id
		{
			get { return id; }
			set { id = value; }
		}

		private string name;
		/// <summary>
		/// The common name used to refer to this study.
		/// </summary>
		public virtual string Name
		{
			get { return name; }
			set { name = value; }
		}

		private string description;
		/// <summary>
		/// Gets/sets the description for this study that will be
		/// displayed to the user when he/she is invited to participate.
		/// </summary>
		public virtual string Description
		{
			get { return description; }
			set { description = value; }
		}

		private System.Collections.IDictionary invitationEmailsMap;
		private System.Collections.IDictionary InvitationEmailsMap
		{
			get
			{
				if (invitationEmailsMap == null)
					invitationEmailsMap = new System.Collections.Specialized.ListDictionary();
				return invitationEmailsMap;
			}
			set
			{
				invitationEmailsMap = value;
				invitationEmails = null; // reset accessor
			}
		}
		private ITwoWayDictionary<string, InvitationEmail, Study> invitationEmails;
		/// <summary>
		/// The invitations that can be used to invite someone to participate in this study.
		/// </summary>
		public virtual ITwoWayDictionary<string, InvitationEmail, Study> InvitationEmails
		{
			get
			{
				if (invitationEmails == null)
					invitationEmails = new TwoWayDictionary<string, InvitationEmail, Study>(this, InvitationEmailsMap);
				return invitationEmails;
			}
		}
		private System.Collections.IDictionary incentivePoolsMap;
		private System.Collections.IDictionary IncentivePoolsMap
		{
			get
			{
				if (incentivePoolsMap == null)
					incentivePoolsMap = new System.Collections.Specialized.ListDictionary();
				return incentivePoolsMap;
			}
			set
			{
				incentivePoolsMap = value;
				incentivePools = null; // reset accessor
			}
		}
		private ITwoWayDictionary<string, IncentiveCollection, Study> incentivePools;
		/// <summary>
		/// The emails that can be used to deliver digital incentives to respondents.
		/// </summary>
		public virtual ITwoWayDictionary<string, IncentiveCollection, Study> IncentivePools
		{
			get
			{
				if (incentivePools == null)
					incentivePools = new TwoWayDictionary<string, IncentiveCollection, Study>(this, IncentivePoolsMap);
				return incentivePools;
			}
		}

		/// <summary>
		/// Gets whether a given Evaluation qualifies for this longitudinal study.
		/// </summary>
		/// <remarks>
		/// The default implementation is to qualify any completed evaluation
		/// that is within the <see cref="StartDate"/> and <see cref="EndDate"/>
		/// attributes.
		/// </remarks>
		public virtual bool Qualifies(Evaluation evaluation)
		{
			if (evaluation == null) throw new ArgumentNullException("evaluation");
			return
				// The evaluation must be finished.
				evaluation.IsCompleted &&
				// The evaluation must be finished at or after the earliest allowable date, if any is set.
				(!StartDate.HasValue || evaluation.EndDate.Value >= StartDate.Value) &&
				// The evaluation must be finished before the latest allowable date, if any is set.
				(!EndDate.HasValue || evaluation.EndDate.Value < EndDate.Value);
		}

		/// <summary>
		/// Reviews a particular follow-up study to determine whether it 
		/// still warrants further study.
		/// </summary>
		/// <param name="tracker">
		/// The study to evaluate the merit of continuing the study.
		/// </param>
		/// <returns>
		/// True if the study should be continued.  False if 
		/// it should be discontinued.
		/// </returns>
		/// <remarks>
		/// If this method returns false, the 
		/// <see cref="EvaluationTracker.IsClosed"/> attribute ought 
		/// to be set to true.
		/// <para>The default implementation is to return true.</para>
		/// </remarks>
		public virtual bool StillQualifies(EvaluationTracker tracker)
		{
			// Default implementation is that all studies should continue through 
			// all of their defined events.
			return true;
		}

		/// <summary>
		/// Registers an <see cref="Evaluation"/> for this longitudinal study
		/// </summary>
		public virtual EvaluationTracker Track(Evaluation evaluation)
		{
			if (evaluation == null) throw new ArgumentNullException("evaluation");
			Debug.Assert(Qualifies(evaluation), "Evaluation does not qualify for this study.");
			if (IsExclusive && evaluation.FollowUps.Count > 0)
				throw new InvalidOperationException("This exclusive follow-up study requires the evaluation to not have any other follow-up studies attached.");
			if (!evaluation.EndDate.HasValue)
				throw new InvalidOperationException("Evaluation must be completed before tracking for a follow-up study.");
			// Don't track evaluation twice
			EvaluationTracker tracker = FindAnyExistingTracker(evaluation);
			if (tracker != null) return tracker;
			tracker = new EvaluationTracker();
			Trackers.Add(tracker);
			evaluation.FollowUps.Add(tracker);
			tracker.AddNextEvent();
			return tracker;
		}

		protected EvaluationTracker FindAnyExistingTracker(Evaluation evaluation)
		{
			if (evaluation == null) throw new ArgumentNullException("evaluation");
			foreach (EvaluationTracker tracker in evaluation.FollowUps)
				if (tracker.Study == this)
					return tracker;
			return null;
		}

		protected internal virtual void OnEventCompleted(EventTracker eventTracker)
		{
		}

		private ISet trackersSet;
		private ISet TrackersSet
		{
			get
			{
				if (trackersSet == null)
					trackersSet = new HashedSet();
				return trackersSet;
			}
			set
			{
				trackersSet = value;
			}
		}
		private ITwoWayCollection<EvaluationTracker, Study> trackers;
		/// <summary>
		/// The set of evaluations that have been collected by this longitudinal study.
		/// </summary>
		public virtual ITwoWayCollection<EvaluationTracker, Study> Trackers
		{
			get
			{
				if (trackers == null)
					trackers = new TwoWaySet<EvaluationTracker, Study>(this, TrackersSet);
				return trackers;
			}
		}

		/// <summary>
		/// Gets a <see cref="System.Data.DataTable"/> filled with all data collected over the course of this study.
		/// </summary>
		public System.Data.DataTable GetDataTable()
		{
			throw new System.NotImplementedException();
		}

		private DateTime? startDate;
		private DateTime? endDate;

		/// <summary>
		/// The date of the first evaluation to consider for participation in this study.
		/// </summary>
		public virtual DateTime? StartDate
		{
			get { return startDate; }
			set { startDate = value; }
		}

		/// <summary>
		/// The date of the last evaluation to consider for participation in this study.
		/// </summary>
		public virtual DateTime? EndDate
		{
			get { return endDate; }
			set { endDate = value; }
		}

		private bool isExclusive;
		/// <summary>
		/// Gets/sets whether participation in this study requires that the respondent not be invited to any other studies for this evaluation.
		/// </summary>
		public virtual bool IsExclusive
		{
			get { return isExclusive; }
			set { isExclusive = value; }
		}

		private int priority;
		/// <summary>
		/// Gets/sets the priority this study has relative to others in the case of an exclusivity conflict.
		/// </summary>
		public virtual int Priority
		{
			get { return priority; }
			set { priority = value; }
		}

		private System.Collections.IList eventsList;
		private System.Collections.IList EventsList
		{
			get
			{
				if (eventsList == null)
					eventsList = new System.Collections.ArrayList();
				return eventsList;
			}
			set { eventsList = value; }
		}
		private TwoWayList<StudyEvent, Study> events;

		public virtual TwoWayList<StudyEvent, Study> Events
		{
			get
			{
				if (events == null)
					events = new TwoWayList<StudyEvent, Study>(this, EventsList);
				return events;
			}
		}

		// The GetInvitationKey method used to be abstract, but NHibernate's Castle.DynamicProxy 
		// would throw exceptions, so it was changed to this.
		/// <summary>
		/// Gets the key to use to look up the email invitation to send to a respondent.
		/// </summary>
		public virtual string GetInvitationKey(EventTracker tracker)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Gets the appropriate event for a given iteration within the study.
		/// </summary>
		/// <param name="iteration">The number of times the person has responded to invitations to this study.</param>
		public StudyEvent GetEvent(int iteration)
		{
			int i;
			for (i = 0; i < Events.Count && iteration > Events[i].Repeats; i++)
				iteration -= Events[i].Repeats + 1;
			if (i < 0 || i >= Events.Count)
				throw new ArgumentOutOfRangeException("iteration", "Must be between 0 and TotalEventIterations - 1.");
			return Events[i];
		}

		/// <summary>
		/// Gets the number of total event iterations possible in this study.
		/// </summary>
		public int TotalEventIterations
		{
			get
			{
				int total = 0;
				for (int i = 0; i < Events.Count; i++)
				{
					StudyEvent se = Events[i];
					if (se == null)
						throw new ApplicationException("Null element in Study.Events list.  Check for incontinuous studyIndex field values.");
					total += 1 + se.Repeats;
				}
				return total;
			}
		}

		/// <summary>
		/// Emails users inviting them to participate in this study.
		/// </summary>
		/// <param name="manager">
		/// The database manager to use.
		/// </param>
		/// <param name="maximumCount">
		/// The maximum number of invites to send.
		/// </param>
		/// <returns>
		/// The number of invites actually sent.
		/// </returns>
		public IEnumerable<EventTracker> EventsWithInvitesToSend()
		{
			foreach (EvaluationTracker tracker in Trackers)
			{
				EventTracker et = tracker.ActiveEvent;
				if (et == null) continue;
				if (et.InviteShouldBeSent)
					yield return et;
			}
		}

		/// <summary>
		/// Emails users inviting them to participate in any existing study.
		/// </summary>
		/// <param name="manager">
		/// The database manager to use.
		/// </param>
		/// <param name="maximumCount">
		/// The maximum number of invites to send.
		/// </param>
		/// <returns>
		/// The number of invites actually sent.
		/// </returns>
		public static IEnumerable<EventTracker> EventsWithInvitesToSendForAllStudies(INAssessManager manager)
		{
			if (manager == null) throw new ArgumentNullException("manager");
			foreach (Study s in manager.GetStudies())
				foreach (EventTracker et in s.EventsWithInvitesToSend())
					yield return et;
		}

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Study other = obj as Study;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}

		#region IReferenceableParentDictionary<string, InvitationEmail, Study> Members

		ITwoWayDictionary<string, InvitationEmail, Study> IReferenceableParentDictionary<string, InvitationEmail, Study>.ContainingDictionary
		{
			get { return InvitationEmails; }
		}

		#endregion

		#region IReferenceableParentDictionary<string,IncentivePool,Study> Members

		ITwoWayDictionary<string, IncentiveCollection, Study> IReferenceableParentDictionary<string, IncentiveCollection, Study>.ContainingDictionary
		{
			get { return IncentivePools; }
		}

		#endregion

		#region IReferenceableParent<StudyEvent,Study> Members

		ITwoWayCollection<StudyEvent, Study> IReferenceableParent<StudyEvent, Study>.ContainingCollection
		{
			get { return Events; }
		}

		#endregion

		#region IReferenceableParent<EvaluationTracker,Study> Members

		ITwoWayCollection<EvaluationTracker, Study> IReferenceableParent<EvaluationTracker, Study>.ContainingCollection
		{
			get { return Trackers; }
		}

		#endregion
	}
}
