using System;
using System.Collections.Generic;
using System.Text;

using Iesi.Collections;
using MyNHibernateContrib.TwoWayCollections;

using NAssess.Core.Evaluating;

namespace NAssess.Core.FollowUp
{
	public class StudyEvent
		: IParentReferencingNode<StudyEvent, Study>
	{
		private int id;

		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}

		private string description;
		/// <summary>
		/// Gets/sets the description for this event that will be
		/// displayed to the user when he/she is invited to participate.
		/// </summary>
		public virtual string Description
		{
			get { return description; }
			set { description = value; }
		}

		private TimeSpan waitTime;
		/// <summary>
		/// The time between when the tracked evaluation is completed and when this event becomes active.  This is also when the first invite is sent out.
		/// </summary>
		public virtual TimeSpan WaitTime
		{
			get { return waitTime; }
			set { waitTime = value; }
		}
		private TimeSpan timeBetweenInvites = TimeSpan.FromDays(14);

		/// <summary>
		/// How long to wait after sending an invite before sending another.
		/// </summary>
		public virtual TimeSpan TimeBetweenInvites
		{
			get { return timeBetweenInvites; }
			set { timeBetweenInvites = value; }
		}
		private ScheduleFrom scheduleFrom = ScheduleFrom.LastCompletedEvent;

		public virtual ScheduleFrom ScheduleFrom
		{
			get { return scheduleFrom; }
			set
			{
				// We don't support OriginalEvaluation until we eliminate the
				// potential timing conflicts when an earlier event has a 
				// sliding schedule like LastCompletedEvent, but a later event
				// is fixed with OriginalEvaluation, since the order of 
				// actual events can actually change the order of events,
				// which would be disastrous.
				// Also, Repeats and OriginalEvaluation scheduling doesn't 
				// make sense.
				if (value == ScheduleFrom.OriginalEvaluation)
					throw new NotSupportedException();
				scheduleFrom = value;
			}
		}
		private int maxInvites;
		/// <summary>
		/// The maximum number of email invites to send someone if they do not respond.
		/// </summary>
		public virtual int MaxInvites
		{
			get { return maxInvites; }
			set { maxInvites = value; }
		}
		private Questionnaire questionnaire;
		/// <summary>
		/// The follow-up questionnaire for the respondent to take when participating in this study.
		/// </summary>
		public virtual Questionnaire Questionnaire
		{
			get { return questionnaire; }
			set { questionnaire = value; }
		}

		private int repeats;
		/// <summary>
		/// The number of times in a row that this event is repeated for each tracked evaluation.
		/// </summary>
		public virtual int Repeats
		{
			get { return repeats; }
			set
			{
				if (value < 0) throw new ArgumentOutOfRangeException("Repeats", value, Resources.NonNegativeNumberRequired);
				repeats = value;
			}
		}
		private Study study;

		public virtual Study Study
		{
			get { return study; }
			private set { study = value; }
		}

		private TimeSpan timeToRespond;
		/// <summary>
		/// Time allowed to the participant to complete their response before the event closes.
		/// </summary>
		public virtual TimeSpan TimeToRespond
		{
			get { return timeToRespond; }
			set { timeToRespond = value; }
		}

		#region IParentReferencingNode<StudyEvent, Study> Members

		Study IParentReferencingNode<StudyEvent, Study>.Parent
		{
			get { return study; }
			set { study = value; }
		}

		#endregion

		private IncentiveCollection participatingIncentives;
		/// <summary>
		/// The pool of incentives to draw from for awarding to respondents.
		/// </summary>
		public IncentiveCollection ParticipatingIncentives
		{
			get { return participatingIncentives; }
			set { participatingIncentives = value; }
		}

		private float participationIncentiveMinimumProportion;
		/// <summary>
		/// The proportion of the event evaluation that must be 
		/// completed to qualify for the participation incentive.
		/// </summary>
		public float ParticipationIncentiveMinimumProportion
		{
			get { return participationIncentiveMinimumProportion; }
			set
			{
				if (value < 0 || value > 1)
					throw new ArgumentOutOfRangeException("ParticipationIncentiveMinimumProportion", value, "Must be a value between 0 and 1, inclusive.");
				participationIncentiveMinimumProportion = value;
			}
		}

		private int participationIncentiveMinimumResponses;
		/// <summary>
		/// The minimum number of responses required to qualify for the 
		/// participation incentive.
		/// </summary>
		public int ParticipationIncentiveMinimumResponses
		{
			get { return participationIncentiveMinimumResponses; }
			set
			{
				if (value < 0) throw new ArgumentOutOfRangeException(
					"ParticipationIncentiveMinimumResponses", value, "Must be a non-negative value.");
				participationIncentiveMinimumResponses = value;
			}
		}

		private bool onlyWhileIncentivesLast;
		/// <summary>
		/// Gets/sets whether the study should shutdown and prevent new
		/// respondents when the supply of incentives run out.
		/// </summary>
		public virtual bool OnlyWhileIncentivesLast
		{
			get { return onlyWhileIncentivesLast; }
			set { onlyWhileIncentivesLast = value; }
		}

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			StudyEvent other = obj as StudyEvent;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}
	}

	public enum ScheduleFrom
	{
		/// <summary>
		/// Wait time is in reference to the completion date of the original evaluation that led to respondent being in this study.
		/// </summary>
		OriginalEvaluation,
		/// <summary>
		/// Wait time is in reference to the scheduled start date of the last event participated in, or the completion date of the original evaluation if no previous event exists.
		/// </summary>
		LastScheduledEvent,
		/// <summary>
		/// Wait time is in reference to the completion date of the last event participated in, or the completion date of the original evaluation if no previous event exists.
		/// </summary>
		LastCompletedEvent,
	}
}
