using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Data.SqlClient;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;
using System.Globalization;
using System.Collections.Specialized;
using System.Threading;

using Iesi.Collections;

using NerdBank.Tools;
using NAssess.Core.Evaluating;
using NAssess.Core.Engines.Reporting;
using MyNHibernateContrib.TwoWayCollections;
using MyNHibernateContrib.GenericCollections;

namespace NAssess.Core.Reporting
{
	/// <summary>
	/// An entity object that represents an assessment.  This object provides access to
	/// design assessments, and also to produce Report objects.
	/// </summary>
	public class Assessment
		: IParentReferencingNode<Assessment, Questionnaire>
	{
		#region Attributes
		private int id;
		public virtual int Id
		{
			get { return id; }
			private set { id = value; }
		}
		private string name;
		public virtual string Name
		{
			get { return name; }
			set { name = value; }
		}
		private string description;
		/// <summary>
		/// Gets/sets a longer name, or description, of this assessment.
		/// </summary>
		public virtual string Description
		{
			get { return description; }
			set { description = value; }
		}

		private Questionnaire questionnaire;
		/// <summary>
		/// The <see cref="Questionnaire"/> to which this Assessment belongs.
		/// </summary>
		public virtual Questionnaire Questionnaire
		{
			get { return questionnaire; }
			private set { questionnaire = value; }
		}

		private CultureInfo primaryCulture = CultureInfo.CurrentUICulture;
		public virtual CultureInfo PrimaryCulture
		{
			get { return primaryCulture; }
			set
			{
				if (value == null) throw new ArgumentNullException("PrimaryCulture");
				primaryCulture = value;
			}
		}
		private System.Collections.IDictionary templatesSet;
		private System.Collections.IDictionary TemplatesSet
		{
			get
			{
				if (templatesSet == null) templatesSet = new ListDictionary();
				return templatesSet;
			}
			set
			{
				templatesSet = value;
				templates = null;
			}
		}
		private IDictionary<CultureInfo, ReportTemplate> templates;
		public virtual IDictionary<CultureInfo, ReportTemplate> Templates
		{
			get
			{
				if (templates == null)
					templates = new IDictionaryWrapper<CultureInfo, ReportTemplate>(TemplatesSet);
				return templates;
			}
		}
		/// <summary>
		/// Sets the template for reports that are generated.  
		/// Sensitive to the culture of the thread.
		/// </summary>
		/// <remarks>
		/// <p>Changes made to this attribute may take effect immediately. 
		/// (i.e. before the call to Commit() is made).</p>
		/// <p>To get/set the XML for a different culture, either switch the thread
		/// to that culture first, or use the Get/SetReportXmlForCulture methods.</p>
		/// </remarks>
		public virtual ReportTemplate Template
		{
			get
			{
				return GetMostRelevantLocalResource<ReportTemplate>(Templates);
			}
			set
			{
				Templates[Thread.CurrentThread.CurrentUICulture] = value;
			}
		}
		//private HybridDictionary reportXPathInternalCache = new HybridDictionary(4, false);
		private const string reportXPathCachePrefix = "ReportXPath: ";
		private string GetReportXPathCacheName()
		{
			return reportXPathCachePrefix + Id + " " + System.Threading.Thread.CurrentThread.CurrentUICulture.Name;
		}
		private int minParticipants = 1;
		public virtual int MinParticipants
		{
			get { return minParticipants; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("MinParticipants", Resources.PositiveValueRequired);
				if (value > MaxParticipants)
					throw new ArgumentOutOfRangeException("MinParticipants", Resources.MinParticipantsBelowMax);
				minParticipants = value;
			}
		}
		private int maxParticipants = 1;
		public virtual int MaxParticipants
		{
			get { return maxParticipants; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("MaxParticipants", Resources.PositiveValueRequired);
				if (value < MinParticipants)
					throw new ArgumentOutOfRangeException("MaxParticipants", Resources.MaxParticipantsAboveMin);
				maxParticipants = value;
			}
		}

		private int minEvalsPerParticipant = 1;
		public virtual int MinEvalsPerParticipant
		{
			get { return minEvalsPerParticipant; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("MinEvalsPerParticipant", Resources.PositiveValueRequired);
				if (value > MaxEvalsPerParticipant)
					throw new ArgumentOutOfRangeException("MinEvalsPerParticipant", Resources.MinEvaluationsPerParticipantBelowMax);
				minEvalsPerParticipant = value;
			}
		}
		private int maxEvalsPerParticipant = 1;
		public virtual int MaxEvalsPerParticipant
		{
			get { return maxEvalsPerParticipant; }
			set
			{
				if (value < 1)
					throw new ArgumentOutOfRangeException("MaxEvalsPerParticipant", Resources.PositiveValueRequired);
				if (value < MinEvalsPerParticipant)
					throw new ArgumentOutOfRangeException("MaxEvalsPerParticipant", Resources.MaxEvaluationsPerParticipantAboveMin);
				maxEvalsPerParticipant = value;
			}
		}
		private bool allowGuestQuestionnaires;
		public virtual bool AllowGuestQuestionnaires
		{
			get { return allowGuestQuestionnaires; }
			set { allowGuestQuestionnaires = value; }
		}
		#endregion

		#region Operations
		/// <summary>
		/// Gets the list of cultures, in precedence order, that should be checked through
		/// when parsing the arrays of culture-specific attributes in this class.
		/// </summary>
		protected virtual CultureInfo[] GetRelevantCultures()
		{
			return CultureAssist.GetRelevantCultures(PrimaryCulture);
		}
		/// <summary>
		/// Gets the most relevant localized resource from a dictionary of
		/// culture-specific resources.
		/// </summary>
		/// <typeparam name="T">The type of resource being sought.</typeparam>
		/// <param name="dictionary">The dictionary to search.</param>
		/// <returns>The relevant resource, or default(T) if none found.</returns>
		protected T GetMostRelevantLocalResource<T>(IDictionary<CultureInfo, T> dictionary)
		{
			return CultureAssist.GetMostRelevantLocalResource<T>(dictionary, GetRelevantCultures());
		}

		/// <summary>
		/// Prepares a Report object for merging an assessment template and evaluation
		/// data into a complete, client-readable report.
		/// </summary>
		/// <returns>A Report object that is ready to be attached to data.</returns>
		public virtual Report ProduceReport()
		{
			if (Template == null) throw new InvalidOperationException(Resources.ReportTemplateRequired);
			if (Questionnaire == null) throw new InvalidOperationException(Resources.AssessmentAddedToQuestionnaireRequired);
			if (Name == null || Name.Length == 0) throw new InvalidOperationException(Resources.AssessmentNameRequired);
			return new Report(this);
		}
		#endregion

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			Assessment other = obj as Assessment;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}




		#region IParentReferencingNode<Assessment, Questionnaire> Members

		Questionnaire IParentReferencingNode<Assessment, Questionnaire>.Parent
		{
			get { return questionnaire; }
			set { questionnaire = value; }
		}

		#endregion
	}
}
