using System;
using System.IO;
using System.Collections;
using NAssess.Core.Evaluating;
using System.Xml;

namespace NAssess.Core.Reporting
{
	/// <summary>
	/// Summary description for Report.
	/// </summary>
	public class Report
	{
		public Questionnaire Questionnaire
		{
			get { return Assessment.Questionnaire; }
		}
		private Assessment assessment;
		public Assessment Assessment
		{
			get { return assessment; }
			protected set { assessment = value; }
		}
		private ReportParticipants participants;
		public ReportParticipants Participants
		{
			get { return participants; }
			protected set { participants = value; }
		}
		protected internal Report(Assessment assessment)
		{
			this.assessment = assessment;
			participants = new ReportParticipants(this);
			Participants.Changed += new EventHandler(Participants_Changed);
		}


		#region Attributes
		public event EventHandler Changed;
		public int HtmlPageCount
		{
			get
			{
				return HtmlEngine.PagesCount;
			}
		}
		private HtmlReportRenderingEngine _HtmlEngine;
		public virtual HtmlReportRenderingEngine HtmlEngine
		{
			get
			{
				if (_HtmlEngine == null)
					HtmlEngine = new NAssess.Core.Engines.Reporting.StandardHtml();

				return _HtmlEngine;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("HtmlEngine");
				_HtmlEngine = value;
				_HtmlEngine.Report = this;
			}
		}

		private FORenderingEngine _FOEngine;
		public virtual FORenderingEngine FOEngine
		{
			get
			{
				if (_FOEngine == null)
					FOEngine = new NAssess.Core.Engines.Reporting.StandardFO();

				return _FOEngine;
			}
			set
			{
				if (value == null) throw new ArgumentNullException("FOEngine");
				_FOEngine = value;
				_FOEngine.Report = this;
			}
		}

		private PdfRenderingEngine pdfEngine;
		public virtual PdfRenderingEngine PdfEngine
		{
			get { return pdfEngine; }
			set
			{
				pdfEngine = value;
				if (pdfEngine != null)
					pdfEngine.Report = this;
			}
		}


		/// <summary>
		/// Sets the WebDirectUrl attribute on both the 
		/// <see cref="HtmlEngine"/> and the <see cref="PdfEngine"/>.
		/// </summary>
		public void SetWebDirectUrl(Uri value)
		{
			HtmlEngine.WebDirectUrl = value;
			FOEngine.WebDirectUrl = value;
		}

		#endregion

		#region Report generating methods
		void VerifyProperUserAndEvalCount()
		{
			if (Participants.Count < Assessment.MinParticipants)
				throw new InvalidOperationException(Assessment.MinParticipants + " participants must be " +
					"attached before this report can be generated.  Only " + Participants.Count +
					" have been attached.");
			foreach (ReportParticipant guy in Participants)
			{
				if (guy.Evaluations.Count < Assessment.MinEvalsPerParticipant)
					throw new InvalidOperationException(Assessment.MinEvalsPerParticipant + " evaluations " +
						"per participant must be attached to generate this report.  Participant with name " +
						guy.Name + " has only " + guy.Evaluations.Count + " attached.");
			}
		}
		public virtual void ToHtml(int page, XmlWriter writer)
		{
			if (page < 1) throw new ArgumentOutOfRangeException("page", page,
			   Resources.PositiveValueRequired);
			if (page > HtmlEngine.PagesCount)
				throw new ArgumentOutOfRangeException("page", page, Resources.PageBeyondUpperBound);
			VerifyProperUserAndEvalCount();
			HtmlEngine.ToHtml(page, writer);
		}
		public virtual string ToHtml(int page)
		{
			if (page < 1) throw new ArgumentOutOfRangeException("page", page,
			   Resources.PositiveValueRequired);
			if (page > HtmlEngine.PagesCount)
				throw new ArgumentOutOfRangeException("page", page, Resources.PageBeyondUpperBound);
			VerifyProperUserAndEvalCount();
			return HtmlEngine.ToHtml(page);
		}
		public Stream ToFO()
		{
			return ToFO(false);
		}
		public Stream ToFO(bool monochrome)
		{
			VerifyProperUserAndEvalCount();
			return FOEngine.ToFO(monochrome);
		}
		public void ToFO(Stream outStream)
		{
			ToFO(outStream, false);
		}
		public void ToFO(Stream outStream, bool monochrome)
		{
			VerifyProperUserAndEvalCount();
			FOEngine.ToFO(outStream, monochrome);
		}

		public Stream ToPdf()
		{
			return ToPdf(false);
		}
		public Stream ToPdf(bool monochrome)
		{
			if (PdfEngine == null) throw new InvalidOperationException(Resources.PdfEngineMustBeSetFirst);
			return PdfEngine.ToPdf(monochrome);
		}
		public void ToPdf(Stream outStream)
		{
			ToPdf(outStream, false);
		}
		public void ToPdf(Stream outStream, bool monochrome)
		{
			if (PdfEngine == null) throw new InvalidOperationException(Resources.PdfEngineMustBeSetFirst);
			PdfEngine.ToPdf(outStream, monochrome);
		}
		#endregion
		#region Events
		protected virtual void OnChanged()
		{
			if (Changed != null) Changed(this, null);
		}
		#endregion
		#region Methods for obtaining embedded data
		#endregion

		private void Participants_Changed(object sender, EventArgs e)
		{
			// Propogate Changed event to whoever wants it
			OnChanged();
		}
	}

}
