using System;
using System.Collections.Generic;
using System.Diagnostics;
using MetaData;

namespace NAssess.Core.Reporting
{
	/// <summary>
	/// Summary description for ReportParticipants.
	/// </summary>
	public class ReportParticipants
	{
		internal event EventHandler Changed;

		private Report report;
		protected IList<ReportParticipant> reportParticipantsList = new List<ReportParticipant>(1);

		protected internal ReportParticipants(Report report)
		{
			this.report = report;
		}

		#region Attributes
		public Report Report
		{
			get
			{
				return report;
			}
			protected set
			{
				report = value;
			}
		}
		public ReportParticipant this [string name] 
		{
			get
			{
				for(int i=0;i<reportParticipantsList.Count;i++)
					if( reportParticipantsList[i].Equals(name) ) 
						return reportParticipantsList[i];
				throw new ArgumentOutOfRangeException("name", name);
			}
		}
		public ReportParticipant this [int index]
		{
			get
			{
				return reportParticipantsList[index];
			}
		}
		public bool Contains(string friendlyName)
		{
			foreach( ReportParticipant participant in reportParticipantsList )
				if( participant.Name == friendlyName )
					return true;
			return false;
		}
		public int Count
		{
			get
			{
				return reportParticipantsList.Count;
			}
		}
		#endregion
		#region Operations
		/// <summary>
		/// Adds a participant to the report
		/// </summary>
		/// <param name="friendlyName">
		/// The name of the participant.
		/// </param>
		public ReportParticipant Attach(string friendlyName)
		{
			return Attach(friendlyName, Demographics.GenderType.Unknown);
		}
		/// <summary>
		/// Adds a participant to the report
		/// </summary>
		/// <param name="friendlyName">
		/// The name of the participant.
		/// </param>
		/// <param name="gender">
		/// The gender of the participant.
		/// </param>
		public ReportParticipant Attach(string friendlyName, Demographics.GenderType gender)
		{
			if( friendlyName == null ) throw new ArgumentNullException("friendlyName");
			// Check for alias uniqueness
			if( Contains(friendlyName) ) 
				throw new ArgumentException("A participant with that FriendlyName already exists in report.  All FriendlyName must be unique.");
			// Make sure we do not go over our limit on allowed participants
			if( reportParticipantsList.Count + 1 > Report.Assessment.MaxParticipants )
				throw new TooManyParticipantsException();

			// All tests passed.  Go ahead and add participant.
			ReportParticipant newParticipant = new ReportParticipant(this, friendlyName, gender);
			reportParticipantsList.Add( newParticipant );

			newParticipant.Changed += new EventHandler(Participant_Changed);
			OnChanged();

			return newParticipant;
		}

		internal object Serialize()
		{
			List<ReportParticipantDetail> list = new List<ReportParticipantDetail>(Count);
			foreach( ReportParticipant rp in this )
			{
				ReportParticipantDetail detail = new ReportParticipantDetail(rp.Name, rp.Gender);
				foreach( Evaluating.Evaluation eval in rp.Evaluations )
					detail.Evaluations.Add( eval.Id );
				list.Add(detail);
			}
			return list;
		}
		internal void Deserialize(object obj)
		{
			throw new NotImplementedException();
			if( obj == null ) throw new ArgumentNullException("obj");
			List<ReportParticipantDetail> list = (List<ReportParticipantDetail>) obj;
			foreach( ReportParticipantDetail detail in list )
			{
				ReportParticipant rp = Attach(detail.Name, detail.Gender);
				//foreach( int evalId in detail.Evaluations )
				//    rp.Evaluations.Attach( Parent.Questionnaire.Transaction.FindEvaluation(evalId) );
			}
		}
		[Serializable] private struct ReportParticipantDetail
		{
			public ReportParticipantDetail(string friendlyName, MetaData.Demographics.GenderType gender)
			{
				Name = friendlyName;
				Gender = gender;
				Evaluations = new System.Collections.ArrayList(1);
			}
			public readonly string Name;
			public readonly MetaData.Demographics.GenderType Gender;
			public readonly System.Collections.ArrayList Evaluations;
		}
		#endregion
		#region Events
		protected virtual void OnChanged()
		{
			if( Changed != null ) Changed(this, null);
		}
		#endregion
		#region Enumerator
		public Enumerator GetEnumerator()
		{
			return new Enumerator(this);
		}
		public class Enumerator
		{
			int position = -1;
			ReportParticipants parent;
			protected internal Enumerator( ReportParticipants parent )
			{
				this.parent = parent;
			}
			public bool MoveNext()
			{
				return ++position < parent.reportParticipantsList.Count;
			}
			public ReportParticipant Current
			{
				get 
				{
					Trace.Assert( position < parent.reportParticipantsList.Count, 
						"This enumerator is being called incorrectly at the end of the list.");
					return parent.reportParticipantsList[position];
				}
			}

		}
		#endregion

		private void Participant_Changed(object sender, EventArgs e)
		{
			// Propogate changed event upward
			OnChanged();
		}
	}
}
