using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;

using NAssess.Core.Engines.Reporting;
using NerdBank.Tools;

namespace NAssess.Core.Reporting
{
	public class ReportTemplate
	{
		private ReportTemplate() { } // NHibernate

		public ReportTemplate(XmlDocument xml)
		{
			Xml = xml;
		}

		#region Attributes
		private int id;
		public int Id
		{
			get { return id; }
			private set { id = value; }
		}

		private string xmlString;
		/// <remarks>
		/// For NHibernate support.
		/// </remarks>
		public string XmlString
		{
			get
			{
				if (xmlString == null && xml != null)
					xmlString = xml.OuterXml;
				return xmlString;
			}
			set
			{
				xmlString = value;
				xml = null;
				xpath = null;
			}
		}

		private XmlDocument xml;
		public XmlDocument Xml
		{
			get
			{
				if (xml == null && xmlString != null)
				{
					xml = new XmlDocument();
					xml.LoadXml(xmlString);
				}
				return xml;
			}
			set
			{
				ValidateTemplate(value);
				xml = value;
				xmlString = null;
				xpath = null;
			}
		}

		private XPathDocument xpath;
		public XPathDocument XPath
		{
			get
			{
				if (xpath == null && XmlString != null)
					xpath = new XPathDocument(new StringReader(XmlString));
				return xpath;
			}
		}

		#endregion

		public static implicit operator XmlDocument(ReportTemplate template)
		{
			if (template == null) return null;
			return template.Xml;
		}

		#region Validation
		private System.Text.StringBuilder xmlValidationErrors;
		protected virtual void ValidateTemplate(XmlDocument value)
		{
			// Validate the incoming report XML to make sure it is fully
			// compliant with workable code.
			XmlReaderSettings settings = new XmlReaderSettings();

			// Prepare to validate incoming report XML against filtering schemas
			settings.Schemas.Add(ReportingTools.NAssessReportingNamespace,
				new XmlTextReader(EmbeddedResources.GetFileStreamFromAssembly(
					"/schema/Report.xsd", Project.DefaultNamespace)));
			//settings.Schemas.Add(ReportingTools.XHtmlNamespace,
			//    new XmlTextReader(Byu.Utilities.Project.EmbeddedFile.GetFileStreamFromAssembly(
			//        "/schema/xhtml-transitional.xsd", Project.DefaultNamespace)));
			settings.ValidationType = ValidationType.Schema;
			settings.ValidationEventHandler += new ValidationEventHandler(report_ValidationEventHandler);

			using (XmlReader reader = XmlReader.Create(new StringReader(value.OuterXml), settings))
			{
				XmlDocument newReport = new XmlDocument();
				newReport.PreserveWhitespace = true;
				xmlValidationErrors = new StringBuilder();
				newReport.Load(reader);
				if (xmlValidationErrors.Length > 0)
				{
					// There were schema validation errors.  
					throw new ArgumentException(string.Format(Resources.ReportValidationErrorSummary,
						xmlValidationErrors));
				}

				SuperValidateNewReportXml(newReport);
			}
		}
		private void report_ValidationEventHandler(object sender, ValidationEventArgs e)
		{
			xmlValidationErrors.Append(e.Message + "\n");
		}
		virtual protected void SuperValidateNewReportXml(XmlDocument newReport)
		{
			// Validate the new report in ways that extend beyond the schema's capacity to check.
			// Throw an exception if something fails.

			// Verify root element, because if that's not what we expect,
			// then the schema would have been ineffective.
			if (newReport.DocumentElement.Name != "nassess:report")
				throw new ArgumentException(Resources.ReportValidationExpectedRootTag);
		}


		#endregion
		#region Namespace generators
		protected internal static void AddNamespaces(XmlNamespaceManager nsmgr)
		{
			if (nsmgr == null) throw new ArgumentNullException("nsmgr");
			nsmgr.AddNamespace("nassess", Engines.Reporting.ReportingTools.NAssessReportingNamespace);
			nsmgr.AddNamespace("xhtml", Engines.Reporting.ReportingTools.XHtmlNamespace);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNameTable nameTable)
		{
			if (nameTable == null) throw new ArgumentNullException("nametable");
			XmlNamespaceManager nsmgr = new XmlNamespaceManager(nameTable);
			AddNamespaces(nsmgr);
			return nsmgr;
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNode node)
		{
			if (node == null) throw new ArgumentNullException("node");
			XmlDocument nodeDoc = node as XmlDocument;
			return AddNamespaces(nodeDoc != null ? nodeDoc.NameTable : node.OwnerDocument.NameTable);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XmlNodeList nodeList)
		{
			if (nodeList == null) throw new ArgumentNullException("nodeList");
			if (nodeList.Count == 0) return null;
			return AddNamespaces(nodeList[0].OwnerDocument.NameTable);
		}
		protected internal static XmlNamespaceManager AddNamespaces(XPathNavigator xpath)
		{
			if (xpath == null) throw new ArgumentNullException("xpath");
			return AddNamespaces(xpath.NameTable);
		}
		#endregion
		#region Searches
		/// <summary>
		/// Performs a fast XPath search of the assessment.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <param name="startNode">
		/// The node to consider the root node in the search.
		/// </param>
		/// <returns>
		/// A <see cref="XPathNodeIterator">nodeset</see> of all search hits.
		/// </returns>
		public static XPathNodeIterator Select(string xpath, XPathNavigator startNode)
		{
			if (xpath == null || xpath.Length == 0) throw new ArgumentNullException("xpath");
			if (startNode == null) throw new ArgumentNullException("startNode");
			XPathExpression exp = startNode.Compile(xpath);
			exp.SetContext(AddNamespaces(startNode));
			return startNode.Select(exp);
		}
		/// <summary>
		/// Performs a fast XPath search of the assessment, 
		/// considering the root node to be the starting node.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <returns>
		/// A <see cref="XPathNodeIterator">nodeset</see> of all search hits.
		/// </returns>
		public XPathNodeIterator Select(string xpath)
		{
			return Select(xpath, XPath.CreateNavigator());
		}
		/// <summary>
		/// Performs a fast XPath search of the assessment.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <param name="startNode">
		/// The node to consider the root node in the search.
		/// </param>
		/// <returns>
		/// An <see cref="XPathNavigator"/> pointing at the first search hit.
		/// </returns>
		public static XPathNavigator SelectSingleNode(string xpath, XPathNavigator startNode)
		{
			XPathNodeIterator nodeset = Select(xpath, startNode);
			return nodeset.MoveNext() ? nodeset.Current : null;
		}
		/// <summary>
		/// Performs a fast XPath search of the assessment,
		/// considering the root node to be the starting node.
		/// </summary>
		/// <param name="xpath">
		/// The XPath search criteria.
		/// </param>
		/// <returns>
		/// An <see cref="XPathNavigator"/> pointing at the first search hit.
		/// </returns>
		public XPathNavigator SelectSingleNode(string xpath)
		{
			return SelectSingleNode(xpath, XPath.CreateNavigator());
		}
		#endregion

		/// <summary>
		/// Tests whether this and another object are equal in a way that 
		/// will still pass when proxy objects are being used.
		/// </summary>
		public override bool Equals(object obj)
		{
			ReportTemplate other = obj as ReportTemplate;
			if (other == null) return false;
			if (Id == 0 && other.Id == 0)
				return (object)this == other;
			else
				return Id == other.Id;
		}

		public override int GetHashCode()
		{
			if (Id == 0) return base.GetHashCode();
			string stringRepresentation =
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType.FullName
				+ "#" + Id.ToString();
			return stringRepresentation.GetHashCode();
		}


	}
}
