using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.ComponentModel;
using System.Xml;

using NerdBank.Tools;
using NAssess.Core.Engines.Evaluation;

namespace NAssess.Core.Web
{
	/// <summary>
	/// A web control that renders/persists an evaluation and its responses for the client.
	/// </summary>
	[
		DefaultProperty("PageNumber"),
		ToolboxData("<{0}:EvaluationArea runat=server></{0}:EvaluationArea>"),
		PersistChildren(false), // persist properties, not controls
		Designer(typeof(AreaDesigner)) // the designer type
	]
	public class EvaluationArea : System.Web.UI.WebControls.WebControl
	{
		private const string NoQuestionnaireSpecifiedResource = "NoQuestionnaireSpecified";
		private const string NoRenderAtDesignTimeResource = "NoRenderAtDesignTime";

		protected readonly CultureAssist CultureAssist;

		#region Controls
		/// <summary>
		/// The control that will be filled with the evaluation html.
		/// </summary>
		protected Literal EvaluationLiteral;
		/// <summary>
		/// The validation control that checks whether all questions have
		/// been (properly) filled out.
		/// </summary>
		protected CustomValidator validator;
		/// <summary>
		/// The label for introducing a list of fields with input errors on them.
		/// </summary>
		protected Label IntroduceBadInputFieldsLabel;
		/// <summary>
		/// The label for listing all fields with input errors.
		/// </summary>
		protected Label FieldsWithErrorsListLabel;
		#endregion

		#region Construction
		public EvaluationArea()
		{
			CultureAssist = new CultureAssist(this);
		}

		protected override void OnInit(EventArgs e)
		{
			base.OnInit(e);

			AddAspNetValidationControl();
			AddBadInputFieldsListControls();
			AddEvaluationHtmlPlaceHolderControl();
		}

		private void AddAspNetValidationControl()
		{
			validator = new CustomValidator();

			validator.ClientValidationFunction = "dotNetSubmit";
			validator.ErrorMessage = EntryErrorMessage;
			validator.Display = ValidatorDisplay.Dynamic;
			validator.Visible = !DesignMode;

			Controls.Add(validator);
		}

		private void AddEvaluationHtmlPlaceHolderControl()
		{
			EvaluationLiteral = new Literal();

			EvaluationLiteral.ID = "phEvaluationHtml";
			EvaluationLiteral.Text = CultureAssist.GetResourceString(
				DesignMode ? NoRenderAtDesignTimeResource : NoQuestionnaireSpecifiedResource);

			EvaluationLiteral.EnableViewState = false;

			Controls.Add(EvaluationLiteral);
		}
		private void AddBadInputFieldsListControls()
		{
			IntroduceBadInputFieldsLabel = new Label();
			FieldsWithErrorsListLabel = new Label();

			IntroduceBadInputFieldsLabel.Text = IntroduceBadInputFieldsMessage;

			IntroduceBadInputFieldsLabel.Visible = false;
			FieldsWithErrorsListLabel.Visible = false;

			FieldsWithErrorsListLabel.EnableViewState = false;

			Controls.Add(IntroduceBadInputFieldsLabel);
			Controls.Add(FieldsWithErrorsListLabel);
		}
		#endregion

		#region Attributes
		private const string GroupOptionsOnLeftViewStateName = "EMARGroupOptionsOnLeft";
		[Bindable(true), Category("Appearance"),
			DefaultValue(StandardHtml.GroupSelectOneOptionsOnLeftDefault),
			Description("Whether matrix question groups will have options on the right of the questions rather than the left.")]
		public bool GroupOptionsOnLeft
		{
			get
			{
				bool defaultValue = StandardHtml.GroupSelectOneOptionsOnLeftDefault;
				if (Evaluation != null && Evaluation.Questionnaire != null)
					defaultValue = Evaluation.Questionnaire.GroupOptionsOnLeft;
				return (bool)GetViewState(GroupOptionsOnLeftViewStateName, defaultValue);
			}
			set
			{
				ViewState[GroupOptionsOnLeftViewStateName] = value;
			}
		}
		private const string GroupOptionsFrequencyViewStateName = "EMARGroupOptionsFrequency";
		[Bindable(true), Category("Appearance"),
			DefaultValue(StandardHtml.GroupSelectOneOptionFrequencyDefault),
			Description("How many rows to render between response labels.")]
		public int GroupOptionsFrequency
		{
			get
			{
				int defaultValue = StandardHtml.GroupSelectOneOptionFrequencyDefault;
				if (Evaluation != null && Evaluation.Questionnaire != null)
					defaultValue = Evaluation.Questionnaire.GroupOptionsFrequency;
				return (int)GetViewState(GroupOptionsFrequencyViewStateName, defaultValue);
			}
			set
			{
				ViewState[GroupOptionsFrequencyViewStateName] = value;
			}
		}
		private const string IncludeClientValidationViewStateName = "EMARIncludeClientValidation";
		[Bindable(true), Category("Behavior"),
			DefaultValue(StandardHtml.IncludeClientValidationDefault),
			Description("Whether client-side data validation scripts are used.")]
		public bool IncludeClientValidation
		{
			get
			{
				return (bool)GetViewState(IncludeClientValidationViewStateName, StandardHtml.IncludeClientValidationDefault);
			}
			set
			{
				ViewState[IncludeClientValidationViewStateName] = value;
			}
		}
		private const string PageNumberViewStateName = "EMARPageNumber";
		private const int PageNumberDefault = 1;
		[Bindable(true), Category("Appearance"),
			DefaultValue(PageNumberDefault),
			Description("The page of the questionnaire to display.")]
		public int PageNumber
		{
			get
			{
				return ViewState[PageNumberViewStateName] == null ? PageNumberDefault : (int)ViewState[PageNumberViewStateName];
			}
			set
			{
				if (value < 1) throw new ArgumentOutOfRangeException("PageNumber", value, Resources.PositiveNumberRequired);
				if (Evaluation != null && value > Evaluation.PagesCount)
					throw new ArgumentOutOfRangeException("PageNumber", value, Resources.PageOutOfRange);
				ViewState[PageNumberViewStateName] = value;
				OnPageNumberChanged();
			}
		}
		private const string EntryErrorMessageViewStateName = "EMAREntryErrorMessage";
		private const string EntryErrorMessageDefault = "Correct entry errors before continuing.";
		[Bindable(true),
			Category("Messages"),
			DefaultValue(EntryErrorMessageDefault),
			Description("The error message to display to the user when some questions are left blank or mis-input."),
			LocalizableAttribute(true)]
		public string EntryErrorMessage
		{
			get
			{
				return (string)GetViewState(EntryErrorMessageViewStateName, EntryErrorMessageDefault);
			}
			set
			{
				ViewState[EntryErrorMessageViewStateName] = value;
				if (validator != null)
					validator.ErrorMessage = value;
			}
		}
		private const string IntroduceBadInputFieldsMessageViewStateName = "EMARIntroduceBadInputFieldsMessage";
		private const string IntroduceBadInputFieldsMessageDefault = "The following fields had errors in them: ";
		[Bindable(true), Category("Messages"),
			DefaultValue(IntroduceBadInputFieldsMessageDefault),
			Description("The message to introduce the list of fields with bad input in them."),
			LocalizableAttribute(true)]
		public string IntroduceBadInputFieldsMessage
		{
			get
			{
				return (string)GetViewState(IntroduceBadInputFieldsMessageViewStateName, IntroduceBadInputFieldsMessageDefault);
			}
			set
			{
				ViewState[IntroduceBadInputFieldsMessageViewStateName] = value;
				if (IntroduceBadInputFieldsLabel != null)
					IntroduceBadInputFieldsLabel.Text = value;
			}
		}
		private const bool AutoSaveDefault = true;
		private bool autoSave = AutoSaveDefault;
		[Bindable(true),
			Category("Behavior"),
			DefaultValue(AutoSaveDefault),
			Description("Whether the evaluation responses are automatically saved on every postback.")]
		public bool AutoSave
		{
			get
			{
				return autoSave;
			}
			set
			{
				autoSave = value;
			}
		}
		public const string WebDirectUrlViewStateName = "EMARWebDirectUrl";
		private const string WebDirectUrlDefault = null;
		[Bindable(true),
			Category("Behavior"), DefaultValue(null),
			Description("The Url to the WebDirect accessor on the web site.")]
		public string WebDirectUrl
		{
			get
			{
				return (string)GetViewState(WebDirectUrlViewStateName, WebDirectUrlDefault);
			}
			set
			{
				if (value != null && value.Length == 0) value = null;

				ValidateWebDirectUrl(value);

				ViewState[WebDirectUrlViewStateName] = value;
			}
		}
		private bool isInErrorState;
		/// <summary>
		/// Gets whether this control encountered an error while saving
		/// user responses.
		/// </summary>
		[Browsable(false)]
		public bool IsInErrorState
		{
			get
			{
				return isInErrorState;
			}
		}

		private INAssessManager manager;
		[Browsable(false)]
		public INAssessManager Manager
		{
			get { return manager; }
			set { manager = value; }
		}

		private const string EvaluationIdViewState = "EMAREvaluationId";
		private int evaluationId;
		[Browsable(false)]
		private int EvaluationId
		{
			get
			{
				if (evaluationId == 0 && ViewState[EvaluationIdViewState] != null)
					evaluationId = (int)ViewState[EvaluationIdViewState];
				return evaluationId;
			}
			set
			{
				if (value == evaluationId) return; // nothing to do
				if (value < 0) throw new ArgumentOutOfRangeException("EvaluationId", value, Resources.NonNegativeNumberRequired);
				Evaluating.Evaluation eval = null;
				if (Manager != null && value > 0) eval = Manager.GetEvaluation(value);

				evaluationId = value;
				if (evaluation == null || (evaluation != null && evaluation.Id != value))
					evaluation = eval;

				ViewState[EvaluationIdViewState] = value;
			}
		}
		private Evaluating.Evaluation evaluation;
		[Browsable(false)]
		public Evaluating.Evaluation Evaluation
		{
			get
			{
				if (evaluation == null && Manager != null && EvaluationId > 0)
					evaluation = Manager.GetEvaluation(EvaluationId);

				return evaluation;
			}
			set
			{
				if (value != null && value.Id < 1)
					throw new ArgumentException(Resources.EvaluationIdMustBePositive);
				evaluation = value;
				EvaluationId = (evaluation == null) ? 0 : evaluation.Id;
			}
		}
		#endregion

		#region Hidden attributes
		// These attributes are inherited from our parent Control class,
		// and show up in the Visual Studio designer.  But they don't
		// apply to this type of control, so by overriding them
		// and setting Browsable=false, we hide them in the designer.

		[Browsable(false), Bindable(false)]
		public override string AccessKey
		{
			get
			{
				return base.AccessKey;
			}
			set
			{
				base.AccessKey = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit Height
		{
			get
			{
				return base.Height;
			}
			set
			{
				base.Height = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit Width
		{
			get
			{
				return base.Width;
			}
			set
			{
				base.Width = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override string ToolTip
		{
			get
			{
				return base.ToolTip;
			}
			set
			{
				base.ToolTip = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override FontInfo Font
		{
			get
			{
				return base.Font;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color ForeColor
		{
			get
			{
				return base.ForeColor;
			}
			set
			{
				base.ForeColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color BackColor
		{
			get
			{
				return base.BackColor;
			}
			set
			{
				base.BackColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color BorderColor
		{
			get
			{
				return base.BorderColor;
			}
			set
			{
				base.BorderColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override BorderStyle BorderStyle
		{
			get
			{
				return base.BorderStyle;
			}
			set
			{
				base.BorderStyle = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit BorderWidth
		{
			get
			{
				return base.BorderWidth;
			}
			set
			{
				base.BorderWidth = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override string CssClass
		{
			get
			{
				return base.CssClass;
			}
			set
			{
				base.CssClass = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override short TabIndex
		{
			get
			{
				return base.TabIndex;
			}
			set
			{
				base.TabIndex = value;
			}
		}

		#endregion

		#region Events
		/// <summary>
		/// Fired when the <see cref="PageNumber"/> attribute changes.
		/// </summary>
		[Category("Behavior"), Description("Fires when the PageNumber attribute changes.")]
		public event EventHandler PageNumberChanged;
		/// <summary>
		/// Fires the <see cref="PageNumberChanged"/> event.
		/// </summary>
		protected virtual void OnPageNumberChanged()
		{
			EventHandler pageNumberChanged = PageNumberChanged;
			if (pageNumberChanged != null)
				pageNumberChanged(this, null);
		}
		protected override void OnLoad(EventArgs e)
		{
			base.OnLoad(e);

			if (Page.IsPostBack)
			{
				if (AutoSave) SaveResponses();
			}
		}
		protected override void OnPreRender(EventArgs e)
		{
			base.OnPreRender(e);

			if (Evaluation != null)
			{
				if (Evaluation.HtmlEngine is StandardHtml)
				{
					StandardHtml engine = (StandardHtml)Evaluation.HtmlEngine;
					engine.GroupSelectOneOptionsOnLeft = GroupOptionsOnLeft;
					engine.GroupSelectOneOptionFrequency = GroupOptionsFrequency;
					engine.IncludeClientValidation = IncludeClientValidation;
				}
				if (WebDirectUrl != null)
					Evaluation.HtmlEngine.WebDirectUrl = new Uri(Page.Request.Url, ResolveUrl(WebDirectUrl));

				EvaluationLiteral.Text = Evaluation.ToHtml(PageNumber);
			}
		}
		#endregion

		#region Operations
		/// <summary>
		/// Saves the incoming evaluation responses to the database.
		/// </summary>
		/// <returns>
		/// Whether all responses were saved without errors.
		/// </returns>
		/// <remarks>
		/// This method may be called automatically on every postback if 
		/// <see cref="AutoSave"/> is set to true.
		/// </remarks>
		public bool SaveResponses()
		{
			if (Evaluation == null) throw new InvalidOperationException(Resources.EvaluationPropertyMustBeSet);
			if (DesignMode) throw new InvalidOperationException(Resources.SavingResponsesAtDesignTimeNotAllowed);

			string[] errorColumns = Evaluation.Responses.FromHtml();
			// Replace column names with labels for easier reading for the user
			for (int i = 0; i < errorColumns.Length; i++)
			{
				MetaData.VarMetaData vmd = new MetaData.VarMetaData(errorColumns[i], false);
				Evaluation.Questionnaire.FillMetadata(vmd);
				errorColumns[i] = vmd.Label;
			}
			if (errorColumns.Length > 0)
			{
				FieldsWithErrorsListLabel.Text += string.Join(", ", errorColumns);
				FieldsWithErrorsListLabel.Visible = true;
				isInErrorState = true;
				return false; // at least partial failure
			}
			isInErrorState = false;
			if (Manager != null)
				Manager.SaveEvaluation(Evaluation);
			else
				System.Diagnostics.Trace.TraceWarning("Manager property not set.  Evaluation responses may not be saved.");
			return true; // total success
		}
		protected object GetViewState(string key, object defaultValue)
		{
			object val = ViewState[key];
			return val == null ? defaultValue : val;
		}
		/// <summary>
		/// Gets a list of the headings for each page in the questionnaire.
		/// </summary>
		/// <remarks>
		/// This list can be used to generate questionnaire navigation links.
		/// </remarks>
		public string[] GetPageHeadings()
		{
			if (Evaluation == null) throw new InvalidOperationException(Resources.EvaluationPropertyMustBeSet);
			if (!(Evaluation.HtmlEngine is StandardHtml))
				throw new InvalidOperationException(Resources.SupportedOnlyWithStandardHtmlEngine);

			return ((StandardHtml)Evaluation.HtmlEngine).GetPageHeadings();
		}

		/// <summary>
		/// Validates a WebDirect URL.
		/// </summary>
		/// <exception cref="ArgumentException">
		/// Thrown when the passed value is not valid.
		/// </exception>
		protected internal static void ValidateWebDirectUrl(string url)
		{
			if (url == null || url.Length == 0) return;

			// Verify that the URL ends with .axd.
			if (!url.EndsWith(".axd"))
				throw new ArgumentException(Resources.InvalidWebDirectUrl, "url");

			// If the path begins with a "~/", then that is acceptable.  
			if (System.Text.RegularExpressions.Regex.IsMatch(url, @"\A~/\w+\.axd\z"))
				return; // good.

			// Also acceptable is a fully-qualified URL.
			try
			{
				Uri validateUrl = new Uri(url);
			}
			catch (UriFormatException ex)
			{
				throw new ArgumentException(Resources.InvalidWebDirectUrl, ex);
			}
		}
		#endregion
	}
}
