using System;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Drawing.Design;
using System.Web.UI;

namespace NAssess.Core.Web
{
	/// <summary>
	/// Summary description for LinkedReportUrlsCollection.
	/// </summary>
	[
	Editor(typeof(System.ComponentModel.Design.CollectionEditor), typeof(UITypeEditor))
	]
	public sealed class LinkedReportUrlsCollection : IStateManager, IList
	{
		/// <summary>
		/// Creates a new instance of the <see cref="LinkedReportUrlsCollection"/> class.
		/// </summary>
		internal LinkedReportUrlsCollection()
		{
		}

		protected ArrayList list = new ArrayList();
		private bool isTrackingViewState;
		private bool saveAll;

		#region Attributes
		/// <summary>
		/// The indexer for the collection.
		/// </summary>
		public LinkedReportUrl this[int index]
		{
			get
			{
				return (LinkedReportUrl)list[index];
			}
			set
			{
				list[index] = value;
			}
		}
		/// <summary>
		/// Determines whether a <see cref="LinkedReportUrl"/> is already
		/// in the collection.
		/// </summary>
		public bool Contains(LinkedReportUrl item)
		{
			return list.Contains(item);
		}
		/// <summary>
		/// Gets the index of the given <see cref="LinkedReportUrl"/>.
		/// </summary>
		public int IndexOf(LinkedReportUrl item)
		{
			if (item == null) throw new ArgumentNullException("item");
			return list.IndexOf(item);
		}
		#endregion

		#region Operations
		/// <summary>
		/// Adds a new <see cref="LinkedReportUrl"/> to the collection.
		/// </summary>
		/// <returns>
		/// The index of the newly added <see cref="LinkedReportUrl"/> in the collection.
		/// </returns>
		public int Add(LinkedReportUrl item)
		{
			if (item == null) throw new ArgumentNullException("item");
			int idx = list.Add(item);

			if (((IStateManager)this).IsTrackingViewState)
			{
				((IStateManager)item).TrackViewState();
				item.SetDirty();
			}

			return idx;
		}
		/// <summary>
		/// Adds a <see cref="LinkedReportUrl"/> to the list at a specific index.
		/// </summary>
		public void Insert(int index, LinkedReportUrl item)
		{
			if (item == null) throw new ArgumentNullException("item");
			list.Insert(index, item);
			if (((IStateManager)this).IsTrackingViewState)
			{
				((IStateManager)item).TrackViewState();
				saveAll = true;
			}
		}
		/// <summary>
		/// Removes a <see cref="LinkedReportUrl"/> from the collection.
		/// </summary>
		public void Remove(LinkedReportUrl item)
		{
			if (item == null) throw new ArgumentNullException("item");
			list.Remove(item);
			if (((IStateManager)this).IsTrackingViewState) saveAll = true;
		}
		#endregion

		#region IStateManager Members
		void IStateManager.TrackViewState()
		{
			isTrackingViewState = true;
			foreach (LinkedReportUrl url in list)
				((IStateManager)url).TrackViewState();
		}

		bool IStateManager.IsTrackingViewState
		{
			get
			{
				return isTrackingViewState;
			}
		}

		object IStateManager.SaveViewState()
		{
			if (saveAll)
			{
				// save all items
				ArrayList states = new ArrayList(Count);
				foreach (LinkedReportUrl url in this)
				{
					url.SetDirty();
					states.Add(((IStateManager)url).SaveViewState());
				}
				return (states.Count > 0) ? states : null;
			}
			else
			{
				// Save only the dirty items
				ArrayList indices = new ArrayList();
				ArrayList states = new ArrayList();
				for (int i = 0; i < Count; i++)
				{
					LinkedReportUrl url = this[i];
					object state = ((IStateManager)url).SaveViewState();
					if (state != null)
					{
						states.Add(state);
						indices.Add(i);
					}
				}
				return (indices.Count > 0) ? new Pair(indices, states) : null;
			}
		}

		void IStateManager.LoadViewState(object savedState)
		{
			if (savedState == null) return;
			ArrayList states = savedState as ArrayList;
			if (states != null)
			{
				// All items were saved.
				// Create new internal list of LinkedReportUrls  using view state.
				saveAll = true;

				list = new ArrayList(states.Count);
				foreach (object itemState in states)
				{
					LinkedReportUrl url = new LinkedReportUrl();
					Add(url);
					((IStateManager)url).LoadViewState(itemState);
				}
			}
			else
			{
				// Load modified items.
				Pair p = (Pair)savedState;
				ArrayList indices = (ArrayList)p.First;
				states = (ArrayList)p.Second;

				for (int i = 0; i < indices.Count; i++)
				{
					int index = (int)indices[i];
					if (index < this.Count)
						((IStateManager)this[index]).LoadViewState(states[i]);
					else
					{
						LinkedReportUrl url = new LinkedReportUrl();
						Add(url);
						((IStateManager)url).LoadViewState(states[i]);
					}
				}
			}
		}

		#endregion

		#region IList Members
		bool IList.IsReadOnly
		{
			get
			{
				return false;
			}
		}

		bool IList.IsFixedSize
		{
			get
			{
				return false;
			}
		}

		object IList.this[int index]
		{
			get
			{
				return this[index];
			}
			set
			{
				this[index] = (LinkedReportUrl)value;
			}
		}
		/// <summary>
		/// Removes the <see cref="LinkedReportUrl"/> at the specified index.
		/// </summary>
		public void RemoveAt(int index)
		{
			list.RemoveAt(index);
			if (((IStateManager)this).IsTrackingViewState) saveAll = true;
		}

		void IList.Insert(int index, object item)
		{
			Insert(index, (LinkedReportUrl)item);
		}

		void IList.Remove(object item)
		{
			Remove((LinkedReportUrl)item);
		}

		bool IList.Contains(object item)
		{
			return Contains((LinkedReportUrl)item);
		}

		/// <summary>
		/// Clears the collection.
		/// </summary>
		public void Clear()
		{
			list.Clear();
			if (((IStateManager)this).IsTrackingViewState)
				saveAll = true;
		}

		int IList.IndexOf(object item)
		{
			return IndexOf((LinkedReportUrl)item);
		}

		int IList.Add(object item)
		{
			return Add((LinkedReportUrl)item);
		}

		#endregion

		#region ICollection Members

		public bool IsSynchronized
		{
			get
			{
				return list.IsSynchronized;
			}
		}

		public int Count
		{
			get
			{
				return list.Count;
			}
		}

		public void CopyTo(Array array, int index)
		{
			list.CopyTo(array, index);
		}

		public object SyncRoot
		{
			get
			{
				return list.SyncRoot;
			}
		}

		#endregion

		#region IEnumerable Members

		public IEnumerator GetEnumerator()
		{
			return list.GetEnumerator();
		}

		#endregion
	}
}
