using System;
using System.IO;
using System.Diagnostics;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.ComponentModel;

using NerdBank.Tools;
using NAssess.Core.Reporting;

namespace NAssess.Core.Web
{
	/// <summary>
	/// Summary description for ReportArea.
	/// </summary>
	[
		DefaultProperty("PageNumber"),
		ToolboxData("<{0}:ReportArea runat=server></{0}:ReportArea>"),
		ParseChildren(true, "LinkedReportUrls"),
		PersistChildren(false), // persist properties, not controls
		Designer(typeof(AreaDesigner)) // the designer type
	]
	public class ReportArea : System.Web.UI.WebControls.WebControl
	{
		private const string NoReportSpecifiedResource = "NoReportSpecified";
		private const string NoRenderAtDesignTimeResource = "NoRenderAtDesignTime";

		protected readonly CultureAssist CultureAssist;

		#region Controls
		/// <summary>
		/// The control that will be filled with the report html.
		/// </summary>
		protected Literal phReport;
		#endregion

		#region Construction
		public ReportArea()
		{
			CultureAssist = new CultureAssist(this);
		}
		protected override void OnInit(EventArgs e)
		{
			base.OnInit(e);

			AddReportHtmlPlaceHolderControl();
		}
		private void AddReportHtmlPlaceHolderControl()
		{
			phReport = new Literal();

			phReport.ID = "phReportHtml";
			phReport.Text = CultureAssist.GetResourceString(
				DesignMode ? NoRenderAtDesignTimeResource : NoReportSpecifiedResource);

			phReport.EnableViewState = false;

			Controls.Add(phReport);
		}
		#endregion

		#region Attributes
		private const string PageNumberViewStateName = "EMARPageNumber";
		private const int PageNumberDefault = 1;
		[Bindable(true), Category("Appearance"),
			DefaultValue(PageNumberDefault),
			Description("The page of the report to display.")]
		public int PageNumber
		{
			get
			{
				return ViewState[PageNumberViewStateName] == null ? PageNumberDefault : (int)ViewState[PageNumberViewStateName];
			}
			set
			{
				if (value < 1) throw new ArgumentOutOfRangeException("PageNumber", value, Resources.NonNegativeNumberRequired);
				if (Report != null && value > Report.HtmlPageCount)
					throw new ArgumentOutOfRangeException("PageNumber", value, Resources.PageOutOfRange);
				ViewState[PageNumberViewStateName] = value;
				OnPageNumberChanged();
			}
		}

		public const string WebDirectUrlViewStateName = "EMARWebDirectUrl";
		private const string WebDirectUrlDefault = null;
		[Bindable(true),
		Category("Behavior"), DefaultValue(null),
		Description("The Url to the WebDirect accessor on the web site.")]
		public string WebDirectUrl
		{
			get
			{
				return (string)GetViewState(WebDirectUrlViewStateName, WebDirectUrlDefault);
			}
			set
			{
				if (value != null && value.Length == 0) value = null;

				EvaluationArea.ValidateWebDirectUrl(value);

				ViewState[WebDirectUrlViewStateName] = value;
			}
		}
		private INAssessManager manager;
		[Browsable(false)]
		public INAssessManager Manager
		{
			get { return manager; }
			set { manager = value; }
		}

		private const string AssessmentIdViewStateName = "EMARAssessmentId";
		private int AssessmentId
		{
			get
			{
				return (int)GetViewState(AssessmentIdViewStateName, 0);
			}
			set
			{
				if (AssessmentId == value) return; // nothing to do

				ViewState[AssessmentIdViewStateName] = value;
			}
		}
		private Report report;
		[Browsable(false)]
		public Report Report
		{
			get
			{
				if (report == null && Manager != null && AssessmentId > 0)
				{
					Assessment a = Manager.GetAssessment(AssessmentId);
					report = a.ProduceReport();
					report.Participants.Deserialize(ViewState["EMARParticipants"]);

					// Preset some attributes, in case a PDF or something
					// is generated based on this before the PreRender event is fired.
					report.SetWebDirectUrl(GenerateWebDirectUrl());
				}
				return report;
			}
			set
			{
				report = value;
				if (report != null)
				{
					// Store break-down of values in ViewState so we can reconstruct this
					// report later.
					AssessmentId = report.Assessment.Id;
					ViewState["EMARParticipants"] = report.Participants.Serialize();
				}
			}
		}
		private LinkedReportUrlsCollection linkedReportUrls = new LinkedReportUrlsCollection();
		[
			Browsable(true),
			Category("Behavior"),
			DesignerSerializationVisibility(DesignerSerializationVisibility.Content),
			NotifyParentProperty(true),
			PersistenceMode(PersistenceMode.InnerDefaultProperty)
		]
		public LinkedReportUrlsCollection LinkedReportUrls
		{
			get
			{
				if (IsTrackingViewState)
					((IStateManager)linkedReportUrls).TrackViewState();
				return linkedReportUrls;
			}
		}
		#endregion

		#region Hidden attributes
		// These attributes are inherited from our parent Control class,
		// and show up in the Visual Studio designer.  But they don't
		// apply to this type of control, so by overriding them
		// and setting Browsable=false, we hide them in the designer.

		[Browsable(false), Bindable(false)]
		public override string AccessKey
		{
			get
			{
				return base.AccessKey;
			}
			set
			{
				base.AccessKey = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit Height
		{
			get
			{
				return base.Height;
			}
			set
			{
				base.Height = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit Width
		{
			get
			{
				return base.Width;
			}
			set
			{
				base.Width = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override string ToolTip
		{
			get
			{
				return base.ToolTip;
			}
			set
			{
				base.ToolTip = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override FontInfo Font
		{
			get
			{
				return base.Font;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color ForeColor
		{
			get
			{
				return base.ForeColor;
			}
			set
			{
				base.ForeColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color BackColor
		{
			get
			{
				return base.BackColor;
			}
			set
			{
				base.BackColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override System.Drawing.Color BorderColor
		{
			get
			{
				return base.BorderColor;
			}
			set
			{
				base.BorderColor = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override BorderStyle BorderStyle
		{
			get
			{
				return base.BorderStyle;
			}
			set
			{
				base.BorderStyle = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override Unit BorderWidth
		{
			get
			{
				return base.BorderWidth;
			}
			set
			{
				base.BorderWidth = value;
			}
		}

		[Browsable(false), Bindable(false)]
		public override string CssClass
		{
			get
			{
				return base.CssClass;
			}
			set
			{
				base.CssClass = value;
			}
		}
		[Browsable(false), Bindable(false)]
		public override short TabIndex
		{
			get
			{
				return base.TabIndex;
			}
			set
			{
				base.TabIndex = value;
			}
		}

		#endregion

		#region Events
		/// <summary>
		/// Fired when the <see cref="PageNumber"/> attribute changes.
		/// </summary>
		[Category("Behavior"), Description("Fires when the PageNumber attribute changes.")]
		public event EventHandler PageNumberChanged;
		/// <summary>
		/// Fires the <see cref="PageNumberChanged"/> event.
		/// </summary>
		protected virtual void OnPageNumberChanged()
		{
			EventHandler pageNumberChanged = PageNumberChanged;
			if (pageNumberChanged != null)
				pageNumberChanged(this, null);
		}
		protected override void OnPreRender(EventArgs e)
		{
			base.OnPreRender(e);

			if (Report != null)
			{
				if (Report.HtmlEngine is Engines.Reporting.StandardHtml)
				{
					Engines.Reporting.StandardHtml engine = (Engines.Reporting.StandardHtml)Report.HtmlEngine;
					foreach (LinkedReportUrl url in LinkedReportUrls)
					{
						if (url.AssessmentName == Report.Assessment.Name)
							engine.LinkWithinReportUrl = new Uri(url.UrlTemplate);
						else
							engine.ExternalReportUrls.Add(Manager.GetAssessment(url.AssessmentId), new Uri(url.UrlTemplate));
					}
				}
				report.SetWebDirectUrl(GenerateWebDirectUrl());

				phReport.Text = Report.ToHtml(PageNumber);
			}
		}

		class FOtoPdf : NAssess.Core.Engines.IFOtoPdfEngine
		{
			public FOtoPdf(ReportArea reportArea)
			{
				if (reportArea == null) throw new ArgumentNullException("reportArea");
				this.reportArea = reportArea;
			}

			private ReportArea reportArea;
			#region IFOtoPdfEngine Members

			private Stream pdfStream;
			public Stream PdfStream
			{
				get { return pdfStream; }
				set { pdfStream = value; }
			}


			public void Render(Stream foStream)
			{
				reportArea.OnRenderFO(foStream, PdfStream);
			}

			#endregion
		}
		public class RenderFOEventArgs : EventArgs
		{
			internal RenderFOEventArgs(Stream foStream, Stream pdfStream)
			{
				FOStream = foStream;
				PdfStream = pdfStream;
			}
			public readonly Stream FOStream;
			public readonly Stream PdfStream;
		}
		[Category("Behavior"), Description("Fires when FO needs to be rendered as a PDF or other format.")]
		public event EventHandler<RenderFOEventArgs> RenderFO;
		protected virtual void OnRenderFO(Stream foStream, Stream pdfStream)
		{
			EventHandler<RenderFOEventArgs> renderFO = RenderFO;
			if (renderFO != null)
				renderFO(this, new RenderFOEventArgs(foStream, pdfStream));
		}
		#endregion

		#region Operations
		private Uri GenerateWebDirectUrl()
		{
			if (WebDirectUrl == null) return null;
			return new Uri(Page.Request.Url, ResolveUrl(WebDirectUrl));
		}
		protected object GetViewState(string key, object defaultValue)
		{
			object val = ViewState[key];
			return val == null ? defaultValue : val;
		}
		/// <summary>
		/// Generates a PDF of the report and streams it to the browser.
		/// </summary>
		public void SendPdf()
		{
			SendPdf(false);
		}
		/// <summary>
		/// Generates a PDF of the report and streams it to the browser.
		/// </summary>
		/// <param name="monochrome">
		/// Whether the generated PDF should be monochrome.  If false, color is used where available.
		/// </param>
		public void SendPdf(bool monochrome)
		{
			if (Report.PdfEngine == null)
				Report.PdfEngine = new NAssess.Core.Engines.Reporting.PluggablePdf(new FOtoPdf(this));
			WebUtilities.StreamToBrowser(Report.ToPdf(monochrome), "application/pdf", "report.pdf");
		}
		#endregion

		#region Custom State Management Implementation
		protected override void LoadViewState(object savedState)
		{
			object baseState = null;
			object[] myState = null;

			if (savedState != null)
			{
				myState = (object[])savedState;
				if (myState.Length != 2) throw new ArgumentException(Resources.InvalidViewState);
				baseState = myState[0];
			}

			base.LoadViewState(baseState);

			if (myState != null && myState[1] != null)
				((IStateManager)LinkedReportUrls).LoadViewState(myState[1]);
		}
		protected override object SaveViewState()
		{
			object baseState = base.SaveViewState();
			object linkedReportUrlsState = null;

			Debug.Assert(LinkedReportUrls != null);
			if (linkedReportUrls.Count > 0)
				linkedReportUrlsState = ((IStateManager)LinkedReportUrls).SaveViewState();

			if (baseState != null || linkedReportUrlsState != null)
			{
				object[] savedState = new object[2];
				savedState[0] = baseState;
				savedState[1] = linkedReportUrlsState;

				return savedState;
			}

			return null;
		}
		protected override void TrackViewState()
		{
			base.TrackViewState();
			Debug.Assert(LinkedReportUrls != null);
			((IStateManager)LinkedReportUrls).SaveViewState();
		}

		#endregion
	}
}
