﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Xml;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NAssess.Core.Evaluating;
using NerdBank.Tools;

namespace TestNAssess.Evaluating
{
	/// <summary>
	///This is a test class for NAssess.Core.Evaluating.Ddi and is intended
	///to contain all NAssess.Core.Evaluating.Ddi Unit Tests
	///</summary>
	[TestClass()]
	public class DdiTest : TestBase
	{
		XmlDocument ddi;
		XmlNamespaceManager nsmgr;

        [TestInitialize]
        public void Initialize()
        {
			ddi = new XmlDocument();
			ddi.LoadXml(EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/rawddi.xml", TestBase.DefaultNamespace));
			nsmgr = NAssess_Evaluating_DdiAccessor.AddNamespaces(ddi);
		}

		/// <summary>
		///A test case for GroupSimilarVariables (XmlDocument, XmlNamespaceManager)
		///</summary>
		[TestMethod()]
		public void GroupSimilarVariablesTest()
		{
			int grpIdx = Ddi.GroupSimilarVariables(ddi, nsmgr);
			Ddi.GroupUngroupedVariables(ddi, nsmgr, grpIdx);

			XmlNodeList grids = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='grid']", nsmgr);
			Assert.AreEqual(3, grids.Count);
			Assert.AreEqual("v1 v2", ((XmlElement)grids[0]).GetAttribute("var"));
			Assert.AreEqual("v5 v6", ((XmlElement)grids[1]).GetAttribute("var"));
			Assert.AreEqual("v7 v8", ((XmlElement)grids[2]).GetAttribute("var"));

			XmlNodeList subjects = ddi.SelectNodes("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='subject']", nsmgr);
			Assert.AreEqual(1, subjects.Count);
			Assert.AreEqual("v3 v4", ((XmlElement)subjects[0]).GetAttribute("var"));
		}


		/// <summary>
		///A test case for PrepareEssentialVarGrps (XmlDocument, XmlNamespaceManager)
		///</summary>
		[TestMethod()]
		public void PrepareEssentialVarGrpsTest()
		{
			if (ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='display']", nsmgr) != null)
				Assert.Inconclusive("DDI already has display section.");
			Ddi.PrepareEssentialVarGrps(ddi, nsmgr);

			Assert.IsNotNull(ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='display']", nsmgr));
			Assert.AreEqual("v1 v2 v3 v4 v5 v6 v7 v8", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='subject']/@var", nsmgr).Value.Trim());
		}

		/// <summary>
		///A test case for FixContinuousMultipleChoiceVars (XmlDocument, XmlNamespaceManager)
		///</summary>
		[TestMethod()]
		public void FixContinuousMultipleChoiceVarsTest()
		{
			Ddi.FixCommonVariableProblems(ddi, nsmgr);

			Assert.AreEqual("discrete", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v1']/@intrvl", nsmgr).Value);
			Assert.AreEqual("discrete", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v2']/@intrvl", nsmgr).Value);
			Assert.AreEqual("contin", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v3']/@intrvl", nsmgr).Value);
		}

		/// <summary>
		///A test case for FixNumericVariableLengths (XmlDocument, XmlNamespaceManager)
		///</summary>
		[TestMethod()]
		public void FixNumericVariableLengthsTest()
		{
			Ddi.FixCommonVariableProblems(ddi, nsmgr);

			Assert.AreEqual("1", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v1']/ddi:location/@width", nsmgr).Value);
			Assert.AreEqual("1", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v2']/ddi:location/@width", nsmgr).Value);
			Assert.AreEqual("3", ddi.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:var[@ID='v3']/ddi:location/@width", nsmgr).Value);
		}

		/// <summary>
		/// Set the codebook DDI to a invalid value.
		/// </summary>
		[TestMethod()] [ExpectedException(typeof(ArgumentException))]
		public void InvalidCodebookTest()
		{
			XmlDocument xml = new XmlDocument();
			xml.AppendChild(xml.CreateElement("badelement"));

			Ddi val = new Ddi(xml);
		}

		[TestMethod()]
		public void DdiPersistenceTest()
		{
			XmlDocument ddi = new XmlDocument();
			ddi.LoadXml(EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi.xml", TestBase.DefaultNamespace));

			Ddi ddiObj = new Ddi(ddi);
			Debug.Assert(ddi.OuterXml == ddiObj.Xml.OuterXml);
			mgr.Save(ddiObj);
            IList<Ddi> list = NAssessDbMgr.GetList<Ddi>();
			Assert.AreEqual(1, list.Count);
			ddiObj = list[0];
			Assert.AreEqual(ddi.OuterXml, ddiObj.Xml.OuterXml);
		}
	}
}
