﻿using System;
using System.Diagnostics;
using System.Xml;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Globalization;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NAssess.Core.Evaluating;
using NAssess.Core.FollowUp;
using NerdBank.Tools;

namespace TestNAssess.Evaluating
{
	/// <summary>
	///This is a test class for NAssess.Core.Evaluating.Evaluation and is intended
	///to contain all NAssess.Core.Evaluating.Evaluation Unit Tests
	///</summary>
	[TestClass()]
	public class EvaluationTest : TestBase
	{
		Questionnaire q;
		Evaluation e;

        [TestInitialize]
        public void Initialize()
        {
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi.xml", TestBase.DefaultNamespace);
			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);

			q = new Questionnaire();
			q.Codebook = new Ddi(xml);
			e = q.StartEvaluation();
		}

		[TestMethod()]
		public void SetResponsesWithoutCodebookTest()
		{
			Evaluation loneEval = new Evaluation();
			Assert.IsNull(loneEval.Questionnaire);
			loneEval.Responses["v1"] = 3;
			Assert.AreEqual(3, Convert.ToInt32(loneEval.Responses["v1"]));
		}

		/// <summary>
		/// Tests for retention of valid responses.
		/// </summary>
		[TestMethod()]
		public void GoodResponsesTest()
		{
			Evaluation target = q.StartEvaluation();

			target.Responses["v1"] = 1;
			target.Responses["v2"] = 5;
			target.Responses["v3"] = -1;

			Assert.AreEqual(1, target.Responses["v1"]);
			Assert.AreEqual(5, target.Responses["v2"]);
			Assert.AreEqual(-1, target.Responses["v3"]);
		}

		/// <summary>
		/// Tests for rejection of responses to non-existing variables.
		/// </summary>
		[TestMethod()]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void InvalidVariablesResponsesTest()
		{
			Evaluation target = q.StartEvaluation();

			target.Responses["somebadvariable"] = 3;
		}

		/// <summary>
		/// Tests for rejection of out of range responses.
		/// </summary>
		[TestMethod()]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void OutOfRangeResponsesTest1()
		{
			Evaluation target = q.StartEvaluation();

			target.Responses["v1"] = 6;
		}

		/// <summary>
		/// Tests for rejection of out of range responses.
		/// </summary>
		[TestMethod()]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void OutOfRangeResponsesTest2()
		{
			Evaluation target = q.StartEvaluation();

			target.Responses["v1"] = 0;
		}

		/// <summary>
		/// Tests for rejection of out of range responses.
		/// </summary>
		[TestMethod()]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void OutOfRangeResponsesTest3()
		{
			Evaluation target = q.StartEvaluation();

			target.Responses["v1"] = -2;
		}

		[TestMethod()]
		public void PageCompletionTest1()
		{
			Assert.AreEqual(0f, e.PageCompletion(1));
		}

		[TestMethod()]
		public void PageCompletionTest2()
		{
			const int totalQuestionsOnPage = 6;

			// Put in some data
			e.Responses["v1"] = 5;
			e.Responses["v2"] = 3;

			Assert.AreEqual((float)2 / totalQuestionsOnPage, e.PageCompletion(1));
		}

		[TestMethod()]
		public void PageCompletionTest3()
		{
			// Put in some data
			e.Responses["v1"] = 3;
			e.Responses["v2"] = 4;
			e.Responses["v3"] = 1;
			e.Responses["v4"] = 1;
			e.Responses["Name"] = "andrew";
			e.Responses["v20"] = 2;

			Assert.AreEqual(1f, e.PageCompletion(1));
		}

		[TestMethod()]
		public void GetPageHeadings()
		{
			string[] expectedHeadings = new string[] { "Section A", "Section B", "Section C" };
			Debug.Assert(e.Questionnaire.Codebook.SelectSingleNode("/ddi:codeBook/ddi:dataDscr/ddi:varGrp[@type='section']").Value.Trim() == "Section B", "Section B should come before Section A in the codebook Xml to make a good test for ordering of headings.");

			string[] headings = ((NAssess.Core.Engines.Evaluation.StandardHtml)e.HtmlEngine).GetPageHeadings();
			Assert.IsNotNull(headings, "GetPageHeadings() returned null.");
			// Although there are three headings, only those on pages that would be displayed 
			// should be in the returned array.
			Assert.AreEqual(e.PagesCount, headings.Length, "Arrays of unexpected length.");
			for (int i = 0; i < headings.Length; i++)
				Assert.AreEqual(expectedHeadings[i], headings[i].Trim());
		}

		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		[TestMethod()]
		public void ScaleNotAvailableThroughFields()
		{
			object o = e.Responses["happy1"];
		}
		[TestMethod()]
		public void ScaleValueNormal()
		{
			e.Responses["v1"] = 1;
			e.Responses["v2"] = 1;
			e.Responses["v3"] = 5;
			e.Responses["v4"] = 5;
			Assert.AreEqual(3f, Convert.ToSingle(e.Scales["happy1"]));
		}
		[TestMethod()]
		public void ScaleValueSomeMissing()
		{
			e.Responses["v1"] = 1;
			e.Responses["v2"] = 5;
			e.Responses["v3"] = null;
			e.Responses["v4"] = null;
			Assert.AreEqual(3f, Convert.ToSingle(e.Scales["happy1"]));
		}
		[TestMethod()]
		public void ScaleValueAllMissing()
		{
			e.Responses["v1"] = null;
			e.Responses["v2"] = null;
			e.Responses["v3"] = null;
			e.Responses["v4"] = null;
			Assert.AreEqual(float.NaN, Convert.ToSingle(e.Scales["happy1"]));
		}
		[TestMethod()]
		public void ScaleValueReversed()
		{
			e.Responses["v1"] = 1;
			e.Responses["v2"] = 5; // this gets reversed to 1 in the scale
			e.Responses["v3"] = 5;
			e.Responses["v4"] = 5;
			Assert.AreEqual(3f, Convert.ToSingle(e.Scales["happy2"]));
		}
		[TestMethod()]
		public void ScaleValueCompositeAndReversed()
		{
			e.Responses["v1"] = 1;
			e.Responses["v2"] = 2; // this gets reversed to 4 in the scale
			e.Responses["v3"] = 4;
			e.Responses["v4"] = 3;
			Assert.AreEqual(3f, Convert.ToSingle(e.Scales["happy3"]));
		}
		[TestMethod()]
		public void ScaleValueNotApplicable()
		{
			e.Responses["v1"] = 5;
			e.Responses["v2"] = null;
			e.Responses["v3"] = -1;
			e.Responses["v4"] = null;
			Assert.AreEqual(5f, Convert.ToSingle(e.Scales["happy1"]));
		}
		[TestMethod()]
		public void ScaleValueNotApplicableReversed()
		{
			e.Responses["v1"] = 5;
			e.Responses["v2"] = -1;
			e.Responses["v3"] = -1;
			e.Responses["v4"] = null;
			Assert.AreEqual(5f, Convert.ToSingle(e.Scales["happy1"]));
		}

		[TestMethod()]
		public void ResponsesChangedTest()
		{
			Evaluation target = q.StartEvaluation();

			int eventFired = 0;
			target.Responses.Changed += new EvaluationResponses.ResponsesChangedEventHandler(
				delegate { eventFired++; } );

			target.Responses["v1"] = 2;
			Assert.AreEqual(1, eventFired);
			target.Responses["v2"] = 3;
			Assert.AreEqual(2, eventFired);
		}

		[TestMethod()]
		public void ResponsesChangedBulkTest()
		{
			Evaluation target = q.StartEvaluation();

			int eventFired = 0;
			target.Responses.Changed += new EvaluationResponses.ResponsesChangedEventHandler(
				delegate { eventFired++; } );
			
			XmlDocument responses = new XmlDocument();
			responses.LoadXml("<responses v1='2' v2='3'/>");
			target.Responses.FromXml(responses.DocumentElement);

			Assert.AreEqual(1, eventFired); // two variables in bulk -> just one event
		}

		/// <summary>
		///A test case for ToHtml (int)
		///</summary>
		[TestMethod()]
		public void ToHtmlNotNullTest()
		{
			Evaluation target = q.StartEvaluation();

			int page = 1;

			string actual;

			actual = target.ToHtml(page);

			Assert.IsNotNull(actual, "NAssess.Core.Evaluating.Evaluation.ToHtml returned a null value.");
		}

		[TestMethod()]
		public void ToHtmlChangedWithResponseTest()
		{
			Evaluation target = q.StartEvaluation();

			int page = 1;

			string before;
			string after;

			before = target.ToHtml(page);
			target.Responses["v1"] = 2;
			after = target.ToHtml(page);

			Assert.AreNotEqual(before, after, "After changing a response value, the HTML did not change.");
		}

		[TestMethod()]
		public void EvaluationPersistenceTest()
		{
			DateTime start = DateTime.Now.Subtract(TimeSpan.FromDays(1));
			DateTime? end = null;

			Evaluation e = new Evaluation();
			e.IncludeInDataSet = false;
			e.Responses["v1"] = 3;
			e.Responses["v2"] = 5;
			e.Responses["name"] = "Andrew";
			e.StartDate = start;
			e.EndDate = end;
			e.ConsentGiven = true;
			e.Culture = CultureInfo.GetCultureInfo("es-MX");

			mgr.Save(e);
			NAssessDbMgr.EndSession();
			e = null;

            IList<Evaluation> list = NAssessDbMgr.GetList<Evaluation>();
			Assert.AreEqual(1, list.Count);
			e = list[0];
			Assert.AreEqual(false, e.IncludeInDataSet);
			Assert.AreEqual(3, Convert.ToInt32(e.Responses["v1"]));
			Assert.AreEqual(5, Convert.ToInt32(e.Responses["v2"]));
			Assert.AreEqual("Andrew", e.Responses["name"]);
			Assert.AreEqual(start.ToString(), e.StartDate.ToString());
			Assert.AreEqual(end, e.EndDate);
			Assert.AreEqual(true, e.ConsentGiven);
			Assert.AreEqual(CultureInfo.GetCultureInfo("es-MX"), e.Culture);
		}

		[TestMethod]
		public void VeryLongValueLengthTest()
		{
			string tooLongValue = new string(' ', 10000);
			Evaluation e = new Evaluation();
			e.Responses["v1"] = tooLongValue;
			Assert.AreEqual(tooLongValue, e.Responses["v1"], "Long value never stuck to in-memory evaluation.");
			
			mgr.Save(e);
            NAssessDbMgr.EndSession();
			
			Evaluation e2 = NAssessDbMgr.Get<Evaluation>(e.Id);
			Assert.AreEqual(e.Responses["v1"], e2.Responses["v1"], "Long value lost after persistence.");
		}

		[TestMethod]
		public void RegisterForFollowUpsTest()
		{
			Study s = FollowUp.FollowUpTestHelper.CreateConfiguration();
			mgr.Save(s);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);
			Assert.AreEqual(1, e.FollowUps.Count);
		}

		[TestMethod]
		public void RegisterForFollowUpsExclusiveTest()
		{
			Study s1 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			mgr.Save(s1);
			
			Study s2 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s2.IsExclusive = true;
			mgr.Save(s2);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);
			Assert.AreEqual(1, e.FollowUps.Count);
			Assert.AreSame(s2, TestBase.GetOnlyItem<EvaluationTracker>(e.FollowUps).Study);
		}

		[TestMethod]
		public void RegisterForFollowUpsExclusiveAgainTest()
		{
			Study s1 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			mgr.Save(s1);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);

			Study s2 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s2.IsExclusive = true;
			mgr.Save(s2);

			e.RegisterForFollowUps(mgr);

            NAssessDbMgr.Commit(); // make sure orphans are properly disposed of

			Assert.AreEqual(1, e.FollowUps.Count);
			Assert.AreEqual(s2.Id, TestBase.GetOnlyItem<EvaluationTracker>(e.FollowUps).Study.Id);
		}

		[TestMethod]
		public void RegisterForFollowUpsMutualExclusiveTest1()
		{
			Study s1 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s1.IsExclusive = true;
			s1.Priority = 7;
			mgr.Save(s1);

			Study s2 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s2.IsExclusive = true;
			s2.Priority = 6;
			mgr.Save(s2);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);
			Assert.AreEqual(1, e.FollowUps.Count);
			Assert.AreSame(s1, TestBase.GetOnlyItem<EvaluationTracker>(e.FollowUps).Study);
		}
		
		[TestMethod]
		public void RegisterForFollowUpsMutualExclusiveTest2()
		{
			Study s1 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s1.IsExclusive = true;
			s1.Priority = 5;
			mgr.Save(s1);

			Study s2 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			s2.IsExclusive = true;
			s2.Priority = 6;
			mgr.Save(s2);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);
			Assert.AreEqual(1, e.FollowUps.Count);
			Assert.AreSame(s2, TestBase.GetOnlyItem<EvaluationTracker>(e.FollowUps).Study);
		}

		[TestMethod]
		public void RegisterForFollowUpsTwiceTest()
		{
			Study s1 = FollowUp.FollowUpTestHelper.CreateConfiguration();
			mgr.Save(s1);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			mgr.Save(e);

			e.RegisterForFollowUps(mgr);
			e.RegisterForFollowUps(mgr);
			Assert.AreEqual(1, e.FollowUps.Count);
		}

		[TestMethod]
		public void StickyCultureTest()
		{
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi-subculture.xml", TestBase.DefaultNamespace);
			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);

			CultureInfo alternateCulture = CultureInfo.GetCultureInfo("es-MX");
			q.Codebooks[alternateCulture] = new Ddi(xml);

			Evaluation e = q.StartEvaluation();
			Assert.AreSame(q.Codebook, e.Codebook);

			e.Culture = alternateCulture;
			Assert.AreSame(q.Codebooks[alternateCulture], e.Codebook);
		}
	}
}
