﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Xml;
using System.Globalization;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NAssess.Core.Evaluating;
using NAssess.Core.Reporting;
using NerdBank.Tools;

namespace TestNAssess.Evaluating
{
	/// <summary>
	///This is a test class for NAssess.Core.Evaluating.Questionnaire and is intended
	///to contain all NAssess.Core.Evaluating.Questionnaire Unit Tests
	///</summary>
	[TestClass()]
	public class QuestionnaireTest : TestBase
	{
		protected Questionnaire getQuestionnaire() 
		{ return getQuestionnaire("/Evaluating/testddi.xml"); }

		protected Questionnaire getQuestionnaire(string ddiFile)
		{
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace(ddiFile, TestBase.DefaultNamespace);
			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);

			Questionnaire q = new Questionnaire();
			q.Codebook = new Ddi(xml);
			q.Consent = "<p>p1</p><p>p2</p>";
			q.Introduction = "<p>p1</p><p>p2</p>";
			return q;
		}

		/// <summary>
		/// Set the codebook DDI to a good value.
		/// </summary>
		[TestMethod()]
		public void GoodCodebookTest()
		{
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi.xml", TestBase.DefaultNamespace);
			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);

			Questionnaire target = new Questionnaire();

			Ddi val = new Ddi(xml);

			target.Codebook = val;

			Assert.AreEqual(val, target.Codebook, "NAssess.Core.Evaluating.Questionnaire.Codebook was not set correctly.");
		}

		/// <summary>
		/// Set the codebook DDI to a null value.
		/// </summary>
		[TestMethod()]
		public void NullCodebookTest()
		{
			Questionnaire target = new Questionnaire();

			Ddi val = null;

			target.Codebook = val;
		}

		/// <summary>
		/// Checks that codebooks for various cultures are correctly managed.
		/// </summary>
		/// <remarks>
		/// A rigorous test of setting/updating codebooks in two 
		/// cultures to make sure that the codebooks are independently
		/// set and retrieved correctly.
		/// </remarks>
		[TestMethod()]
		public void CodebookMulticulturalTest()
		{
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi.xml", TestBase.DefaultNamespace);
			string ddi2 = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi-subculture.xml", TestBase.DefaultNamespace);
			string ddi3 = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi-modified.xml", TestBase.DefaultNamespace);

			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);
			Ddi val = new Ddi(xml);

			XmlDocument xml2 = new XmlDocument();
			xml2.LoadXml(ddi2);
			Ddi val2 = new Ddi(xml2);

			XmlDocument xml3 = new XmlDocument();
			xml3.LoadXml(ddi3);
			Ddi val3 = new Ddi(xml3);

			Questionnaire target = new Questionnaire();

			CultureInfo ci = Thread.CurrentThread.CurrentUICulture;
			try
			{
				target.Codebook = val;
				Assert.AreEqual(val, target.Codebook);
				
				Thread.CurrentThread.CurrentUICulture = CultureInfo.GetCultureInfo("es-MX");
				target.Codebook = val2;
				Assert.AreEqual(val2, target.Codebook);
				
				Thread.CurrentThread.CurrentUICulture = ci;
				Assert.AreEqual(val, target.Codebook);
				
				Thread.CurrentThread.CurrentUICulture = CultureInfo.GetCultureInfo("es-MX");
				Assert.AreEqual(val2, target.Codebook);
				
				Thread.CurrentThread.CurrentUICulture = ci;
				target.Codebook = val3;
				Assert.AreEqual(val3, target.Codebook);

				Thread.CurrentThread.CurrentUICulture = CultureInfo.GetCultureInfo("es-MX");
				Assert.AreEqual(val2, target.Codebook);
			}
			finally
			{
				Thread.CurrentThread.CurrentUICulture = ci;
			}
		}

		[TestMethod]
		public void ConsentMulticulturalTest()
		{
			Questionnaire q = new Questionnaire();
			CultureInfo std = Thread.CurrentThread.CurrentUICulture;
			CultureInfo other = CultureInfo.GetCultureInfo("pt");

			q.Consent = "standard consent";
			Assert.AreEqual(q.Consent, q.Consents[std]);

			Thread.CurrentThread.CurrentUICulture = other;
			Assert.AreEqual("standard consent", q.Consents[std]);
			Assert.AreEqual(q.Consents[std], q.Consent);
			Thread.CurrentThread.CurrentUICulture = std;

			q.Consents[other] = "pt consent";
			Assert.AreEqual("standard consent", q.Consent);
			Assert.AreEqual(q.Consent, q.Consents[std]);
			Assert.AreEqual("pt consent", q.Consents[other]);

			Thread.CurrentThread.CurrentUICulture = other;
			Assert.AreEqual("pt consent", q.Consent);
			Assert.AreEqual("standard consent", q.Consents[std]);

			Thread.CurrentThread.CurrentUICulture = std;
			q.Consents[other] = null;
			Assert.AreEqual(q.Consents[std], q.Consent);
		}

		[TestMethod]
		public void IntroductionMulticulturalTest()
		{
			Questionnaire q = new Questionnaire();
			CultureInfo std = Thread.CurrentThread.CurrentUICulture;
			CultureInfo other = CultureInfo.GetCultureInfo("pt");

			q.Introduction = "standard Introduction";
			Assert.AreEqual(q.Introduction, q.Introductions[std]);

			Thread.CurrentThread.CurrentUICulture = other;
			Assert.AreEqual("standard Introduction", q.Introductions[std]);
			Assert.AreEqual(q.Introductions[std], q.Introduction);
			Thread.CurrentThread.CurrentUICulture = std;

			q.Introductions[other] = "pt Introduction";
			Assert.AreEqual("standard Introduction", q.Introduction);
			Assert.AreEqual(q.Introduction, q.Introductions[std]);
			Assert.AreEqual("pt Introduction", q.Introductions[other]);

			Thread.CurrentThread.CurrentUICulture = other;
			Assert.AreEqual("pt Introduction", q.Introduction);
			Assert.AreEqual("standard Introduction", q.Introductions[std]);

			Thread.CurrentThread.CurrentUICulture = std;
			q.Introductions[other] = null;
			Assert.AreEqual(q.Introductions[std], q.Introduction);
		}

		/// <summary>
		/// Verify that adding an assessment to a questionnaire automatically
		/// makes that assessment point to the new parent questionnaire.
		/// </summary>
		[TestMethod()]
		public void AddingAssessmentMakesTwoWayRelationshipTest()
		{
			Questionnaire q = new Questionnaire();
			Assessment a = new Assessment();
			q.Assessments.Add(a);

			Assert.AreEqual(1, q.Assessments.Count);
			Assert.AreSame(a, TestBase.GetOnlyItem<Assessment>(q.Assessments));
			Assert.AreSame(q, a.Questionnaire);
		}

		/// <summary>
		/// Verify that adding an assessment to a questionnaire automatically
		/// removes it from a previous questionnaire.
		/// </summary>
		[TestMethod()]
		public void MoveAssessmentByAddingToOtherQuestionnaireTest()
		{
			Questionnaire q = new Questionnaire();
			Assessment a = new Assessment();
			q.Assessments.Add(a);

			Questionnaire q2 = new Questionnaire();
			q2.Assessments.Add(a);

			Assert.AreEqual(0, q.Assessments.Count);
			Assert.AreEqual(1, q2.Assessments.Count);
			Assert.AreSame(a, TestBase.GetOnlyItem<Assessment>(q2.Assessments));
			Assert.AreSame(q2, a.Questionnaire);
		}


		[TestMethod()]
		public void PrimaryKeyDefinedInDownloadedData()
		{
			Questionnaire q = getQuestionnaire();
			Evaluation e = q.StartEvaluation();
			e.EndDate = DateTime.Now;
			mgr.Save(e);
			IEnumerator<System.Data.DataRow> rows = q.GetDataTable(mgr).GetEnumerator();
			if (!rows.MoveNext()) Assert.Inconclusive("Cannot measure because questionnaire is empty.");
			Assert.IsTrue(rows.Current.Table.PrimaryKey.Length > 0, 
				"Primary key missing from downloaded DataTable.");
		}
		[TestMethod()]
		public void GetFlatDataTableColumnsCount()
		{
			Questionnaire q = getQuestionnaire();
			Evaluation e = q.StartEvaluation();
			e.EndDate = DateTime.Now;
			mgr.Save(e);
			IEnumerator<System.Data.DataRow> rows = q.GetDataTable(mgr).GetEnumerator();
			if (!rows.MoveNext()) Assert.Inconclusive("Cannot measure because questionnaire is empty.");

			const int nExpectedColumns = 11 + 1; // the fields plus the primary key column
			Assert.AreEqual(nExpectedColumns, rows.Current.Table.Columns.Count);
		}
		[TestMethod()]
		public void GetFlatDataTableRightData()
		{
			Questionnaire q = getQuestionnaire();
			const int numberOfEvals = 5;
			Evaluation[] e = new Evaluation[numberOfEvals];
			for (int i = 0; i < numberOfEvals; i++)
			{
				e[i] = q.StartEvaluation();
				e[i].Responses["v1"] = i + 1;
				mgr.Save(e[i]);
			}
            IEnumerator<System.Data.DataRow> rows = q.GetDataTable(mgr).GetEnumerator();
            int rowCount = 0;
            while( rows.MoveNext() )
			{
                rowCount++;
				Assert.IsTrue(Convert.ToInt32(rows.Current["evaluationId"]) > 0);
				Assert.AreEqual(rows.Current["evaluationId"], rows.Current["v1"]);
			}
            Assert.AreEqual(numberOfEvals, rowCount);
        }
		[TestMethod()]
		public void GetFlatDataTableWOdeletedRecords()
		{
			Questionnaire q = getQuestionnaire();
			Evaluation e = q.StartEvaluation();
			e.IncludeInDataSet = false;
            IEnumerator<System.Data.DataRow> rows = q.GetDataTable(mgr).GetEnumerator();
            int rowCount = 0;
            while(rows.MoveNext()) rowCount++;
			Assert.AreEqual(0, rowCount);
		}

		[TestMethod()]
		public void QuestionnairePersistanceTest()
		{
			string ddi = EmbeddedResources.LoadFileFromAssemblyWithNamespace("/Evaluating/testddi.xml", TestBase.DefaultNamespace);
			XmlDocument xml = new XmlDocument();
			xml.LoadXml(ddi);

			Questionnaire q = new Questionnaire();
			q.Name = "test questionnaire";
			q.Consent = new string('c', 10 * 1024); // some consent form.
			q.Consents[CultureInfo.GetCultureInfo("pt")] = "pt consent";
			q.Introduction = new string('i', 10 * 1024);
			q.Introductions[CultureInfo.GetCultureInfo("pt")] = "pt introduction";

			q.Codebook = new Ddi(xml);
			ddi = q.Codebook.Xml.OuterXml;

			Assessment a = new Assessment();
			a.Name = "test assessment";
			q.Assessments.Add(a);

			mgr.Save(q);
			q = null;
			a = null;

			NAssessDbMgr.EndSession();

            IList<Questionnaire> list = NAssessDbMgr.GetList<Questionnaire>();
			Assert.AreEqual(1, list.Count);
			q = (Questionnaire)list[0];
			Assert.AreEqual("test questionnaire", q.Name);
			Assert.AreEqual(new string('c', 10 * 1024), q.Consent);
			Assert.AreEqual("pt consent", q.Consents[CultureInfo.GetCultureInfo("pt")]);
			Assert.AreEqual(new string('i', 10 * 1024), q.Introduction);
			Assert.AreEqual("pt introduction", q.Introductions[CultureInfo.GetCultureInfo("pt")]);
			Assert.IsNotNull(q.Codebook);
			q.Codebook.Xml.Save(Path.Combine(TestContext.TestDeploymentDir, "testddi2.xml"));
			Assert.AreEqual(ddi, q.Codebook.Xml.OuterXml);

			Assert.AreEqual(1, q.Assessments.Count);
			a = TestBase.GetOnlyItem<Assessment>(q.Assessments);
			Assert.AreEqual("test assessment", a.Name);
			Assert.AreSame(q, a.Questionnaire, "Unpersisted assessment fails to point to parent questionnaire.");
		}
	}
}
