﻿using System;
using System.Diagnostics;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NAssess.Core.Evaluating;
using NAssess.Core.FollowUp;

namespace TestNAssess.FollowUp
{
	/// <summary>
	///This is a test class for NAssess.Core.FollowUp.EventTracker and is intended
	///to contain all NAssess.Core.FollowUp.EventTracker Unit Tests
	///</summary>
	[TestClass()]
	public class EventTrackerTest : TestBase
	{
		[TestMethod]
		public void EventTrackerEvaluationPointsBackTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			e.EndDate = DateTime.Now.Subtract(s.Events[0].WaitTime);
			EvaluationTracker et = s.Track(e);

			Evaluation e2 = et.ActiveEvent.Start();
			Assert.AreEqual(et.ActiveEvent, e2.FollowUpEvent);
			Assert.AreEqual(e2, et.ActiveEvent.EventEvaluation);
		}

		[TestMethod]
		public void EventTrackerPersistenceTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();
			Evaluation origE = new Evaluation();
			origE.Responses["qualifies"] = "y";
			origE.IsCompleted = true;
			origE.EndDate = DateTime.Now.Subtract(s.Events[0].WaitTime);
			EvaluationTracker evalT = s.Track(origE);

			EventTracker et = evalT.ActiveEvent;
            NAssess_FollowUp_EventTrackerAccessor eta = new NAssess_FollowUp_EventTrackerAccessor(et);
			eta.respondedDate = DateTime.Now.Subtract(TimeSpan.FromMinutes(1));
			Evaluation e = et.Start();
			
			Incentive iA = new Incentive();
			et.Incentives.Add(iA);
			Incentive iB = new Incentive();
			et.Incentives.Add(iB);

			mgr.Save(origE);
			mgr.Save(et);
			mgr.Save(e);
            NAssessDbMgr.RestartSession();
			
			Study s2 = NAssessDbMgr.Get<Study>(s.Id);
            EventTracker et2 = NAssessDbMgr.Get<EventTracker>(et.Id);
            Evaluation e2 = NAssessDbMgr.Get<Evaluation>(e.Id);
            Evaluation origE2 = NAssessDbMgr.Get<Evaluation>(origE.Id);
			Debug.Assert((object)et != (object)et2, "Test no good if they are the same object.");
			Assert.AreEqual(1, TestBase.GetOnlyItem<EvaluationTracker>(origE2.FollowUps).TrackedEvents.Count);
			Assert.AreSame(et2, TestBase.GetOnlyItem<EvaluationTracker>(origE2.FollowUps).TrackedEvents[0]);
			Assert.AreSame(s2.Events[0], et2.StudyEvent);
			Assert.AreEqual(et.InvitesSent, et2.InvitesSent);
			Assert.AreEqual(et.LastInviteSentDate, et2.LastInviteSentDate);
			Assert.AreEqual(et.RespondedDate.ToString(), et2.RespondedDate.ToString());
			Assert.AreSame(e2, et2.EventEvaluation, "The event tracker failed to point to its evaluation.");
			Assert.AreEqual(origE2.Id, et2.EvaluationTracker.Evaluation.Id);
			Assert.AreSame(et2, e2.FollowUpEvent);

			Assert.AreEqual(et.Incentives.Count, et2.Incentives.Count);
			foreach (Incentive i2 in et2.Incentives)
				Assert.IsTrue(i2.Id == iA.Id || i2.Id == iB.Id);
		}

		[TestMethod]
		[ExpectedException(typeof(InvalidOperationException))]
		public void StartFailsWhenOutOfIncentivesTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();
			s.Events[0].OnlyWhileIncentivesLast = true;

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			e.EndDate = DateTime.Now.Subtract(s.Events[0].WaitTime);
			EvaluationTracker et = s.Track(e);

			et.ActiveEvent.Start();
		}

		[TestMethod]
		public void EventTrackerDuplicityPersistenceTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();

			Evaluation e1 = new Evaluation();
			e1.Responses["qualifies"] = "y";
			e1.IsCompleted = true;

			Evaluation e2 = new Evaluation();
			e2.Responses["qualifies"] = "y";
			e2.IsCompleted = true;

			mgr.Save(s);
			mgr.Save(e1);
			mgr.Save(e2);

			e1.RegisterForFollowUps(mgr);
			e2.RegisterForFollowUps(mgr);

			NAssessDbMgr.EndSession();

            Evaluation re1 = NAssessDbMgr.Get<Evaluation>(e1.Id);
            Evaluation re2 = NAssessDbMgr.Get<Evaluation>(e2.Id);

			Assert.IsNull(re1.FollowUpEvent);
			Assert.IsNull(re2.FollowUpEvent);

			Assert.AreEqual(1, re1.FollowUps.Count);
			Assert.AreEqual(1, re1.FollowUps.Count);

			EvaluationTracker et1 = TestBase.GetOnlyItem<EvaluationTracker>(re1.FollowUps);
			EvaluationTracker et2 = TestBase.GetOnlyItem<EvaluationTracker>(re2.FollowUps);

			Assert.IsNull(et1.TrackedEvents[0].EventEvaluation);
			Assert.IsNull(et2.TrackedEvents[0].EventEvaluation);
		}
	}
}
