using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using NAssess.Core.Evaluating;
using NAssess.Core.FollowUp;

namespace TestNAssess.FollowUp
{
	class FollowUpStudy : Study
	{

		public override bool Qualifies(NAssess.Core.Evaluating.Evaluation evaluation)
		{
			return (string)evaluation.Responses["qualifies"] == "y";
		}

		public override string GetInvitationKey(EventTracker tracker)
		{
			return "only";
		}
	}

	[TestClass()]
	public class StudyTest : TestBase
	{
		[TestMethod]
		public void EventsTest()
		{
			Study s = new FollowUpStudy();
			Assert.AreEqual(0, s.Events.Count);

			StudyEvent[] ses = new StudyEvent[5];
			for (int i = 0; i < ses.Length; i++)
			{
				Assert.AreEqual(i, s.Events.Count);
				ses[i] = new StudyEvent();
				s.Events.Add(ses[i]);
				Assert.AreSame(ses[i], s.Events[i]);
			}
		}

		private static List<StudyEvent> GenerateComplexRepeatingEvents(Study s)
		{
			StudyEvent[] ses = new StudyEvent[10];
			List<StudyEvent> expected = new List<StudyEvent>();
			for (int i = 0; i < ses.Length; i++)
			{
				s.Events.Add(ses[i] = new StudyEvent());
				ses[i].Repeats = i < 5 ? i : 10 - i; // go up, and down in repeat count
				for (int j = 0; j <= ses[i].Repeats; j++)
					expected.Add(ses[i]);
			}

			return expected;
		}

		[TestMethod]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void GetEventOutOfRangeTest()
		{
			Study s = new FollowUpStudy();
			s.GetEvent(0); // no events at all
		}

		[TestMethod]
		[ExpectedException(typeof(ArgumentOutOfRangeException))]
		public void GetEventRepeatingOutOfRangeTest()
		{
			Study s = new FollowUpStudy();
			List<StudyEvent> expected = GenerateComplexRepeatingEvents(s);
			s.GetEvent(expected.Count);
		}

		[TestMethod]
		public void TotalEventIterationsRepeatingTest()
		{
			Study s = new FollowUpStudy();
			List<StudyEvent> expected = GenerateComplexRepeatingEvents(s);
			Assert.AreEqual(expected.Count, s.TotalEventIterations);
		}

		[TestMethod]
		public void GetEventNoRepeatingTest()
		{
			Study s = new FollowUpStudy();

			StudyEvent[] ses = new StudyEvent[5];
			for (int i = 0; i < ses.Length; i++)
				s.Events.Add(ses[i] = new StudyEvent());

			for (int i = 0; i < ses.Length; i++)
				Assert.AreSame(ses[i], s.GetEvent(i));
		}

		[TestMethod]
		public void GetEventRepeatingTest()
		{
			Study s = new FollowUpStudy();
			List<StudyEvent> expected = GenerateComplexRepeatingEvents(s);

			for (int i = 0; i < expected.Count; i++)
				Assert.AreSame(expected[i], s.GetEvent(i), "Failed when requesting event {0}.", i);
		}

		[TestMethod]
		public void GetEventRepeating2Test()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();

			List<StudyEvent> expected = new List<StudyEvent>();
			expected.Add(s.Events[0]);
			expected.Add(s.Events[1]);
			expected.Add(s.Events[1]);
			expected.Add(s.Events[1]);
			expected.Add(s.Events[1]);
			expected.Add(s.Events[2]);

			for (int i = 0; i < expected.Count; i++)
				Assert.AreSame(expected[i], s.GetEvent(i), "Failed when requesting event {0}.", i);
		}

		[TestMethod]
		public void TrackAddsFirstEventTest()
		{
			Study s = new FollowUpStudy();

			StudyEvent se = new StudyEvent();
			s.Events.Add(se);

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			EvaluationTracker et = s.Track(e);

			Assert.AreEqual(1, et.TrackedEvents.Count, "First event has not been scheduled.");
			Assert.AreSame(se, et.TrackedEvents[0].StudyEvent, "Incorrect first event.");
			Assert.AreSame(et.TrackedEvents[0], et.ActiveEvent);
		}

		[TestMethod]
		public void ScheduledReminderTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.IsCompleted = true;
			EvaluationTracker et = s.Track(e);

			Assert.IsNull(et.ActiveEvent, "The event should not have been activated yet.");
			Assert.IsFalse(et.TrackedEvents[0].InviteShouldBeSent);
			e.EndDate -= s.Events[0].WaitTime + TimeSpan.FromHours(1);
			Assert.IsNotNull(et.ActiveEvent, "The event should have started by now.");
			Assert.IsTrue(et.ActiveEvent.InviteShouldBeSent);
			et.ActiveEvent.SendInvite(new System.Net.Mail.MailAddress("andrewarnott@gmail.com"),
				new Uri("http://somestudy.byu.edu"));
			Assert.IsFalse(et.ActiveEvent.InviteShouldBeSent);
			Assert.IsTrue(et.ActiveEvent.LastInviteSentDate - DateTime.Now < TimeSpan.FromSeconds(5));
			Assert.AreEqual(1, et.ActiveEvent.InvitesSent);
		}

		[TestMethod]
		public void ResponseToEventTest()
		{
			Study s = FollowUpTestHelper.CreateConfiguration();

			Evaluation e = new Evaluation();
			e.Responses["qualifies"] = "y";
			e.EndDate = DateTime.Now - s.Events[0].WaitTime - TimeSpan.FromHours(1);
			EvaluationTracker et = s.Track(e);

			// go through first event
			Assert.IsNotNull(et.ActiveEvent);
			Assert.IsFalse(et.TrackedEvents[0].RespondedDate.HasValue);
			Evaluation fe = et.ActiveEvent.Start();
			Assert.IsTrue(et.TrackedEvents[0].RespondedDate.HasValue);
			Assert.AreEqual(EventTrackerState.Active, et.TrackedEvents[0].State);
			Assert.AreSame(s.Events[0].Questionnaire, fe.Questionnaire);
			fe.IsCompleted = true;
			Assert.AreEqual(EventTrackerState.Completed, et.TrackedEvents[0].State);
			Assert.AreEqual(2, et.TrackedEvents.Count);
			Assert.AreSame(et.ActiveEvent, et.TrackedEvents[1]);

			// go through second event (3x)
			for (int i = 1; i <= s.Events[1].Repeats + 1; i++)
			{
				fe = et.ActiveEvent.Start();
				Assert.IsTrue(et.TrackedEvents[i].RespondedDate.HasValue);
				Assert.AreEqual(EventTrackerState.Active, et.TrackedEvents[i].State);
				Assert.AreSame(s.Events[1].Questionnaire, fe.Questionnaire);
				fe.IsCompleted = true;
				Assert.AreEqual(EventTrackerState.Completed, et.TrackedEvents[i].State);
			}

			// check for third event
			Assert.IsNull(et.ActiveEvent); // null this time because of wait period before 3rd event
			Assert.AreEqual(6, et.TrackedEvents.Count);
			Assert.AreSame(s.Events[2], et.TrackedEvents[5].StudyEvent);
		}

		[TestMethod]
		public void StudyPersistenceTest()
		{
			Study s = new FollowUpStudy();
			s.Name = "some test name";
			s.Description = "some description";
			s.StartDate = DateTime.Parse("4/1/2000");
			s.EndDate = s.StartDate.Value.AddYears(2);
			s.IsExclusive = true;
			s.Priority = 5;

			StudyEvent seA = new StudyEvent();
			seA.Repeats = 1;
			s.Events.Add(seA);
			StudyEvent seB = new StudyEvent();
			seB.Repeats = 2;
			s.Events.Add(seB);

			EvaluationTracker etA = new EvaluationTracker();
			s.Trackers.Add(etA);
			EvaluationTracker etB = new EvaluationTracker();
			s.Trackers.Add(etB);

			IncentiveCollection pool = new IncentiveCollection();
			string poolKey = "some pool";
			s.IncentivePools.Add(poolKey, pool);

			mgr.Save(s);
			NAssessDbMgr.EndSession();

			Study s2 = NAssessDbMgr.Get<Study>(s.Id);
			Debug.Assert((object)s != (object)s2, "Test no good if they are the same object.");
			Assert.AreEqual(s.Name, s2.Name);
			Assert.AreEqual(s.Description, s2.Description);
			Assert.AreEqual(s.StartDate, s2.StartDate);
			Assert.AreEqual(s.EndDate, s2.EndDate);
			Assert.AreEqual(s.IsExclusive, s2.IsExclusive);
			Assert.AreEqual(s.Priority, s2.Priority);

			Assert.AreEqual(s.Events.Count, s2.Events.Count);
			Assert.AreEqual(seA.Repeats, s2.Events[0].Repeats);
			Assert.AreEqual(seB.Repeats, s2.Events[1].Repeats);

			Assert.AreEqual(s.Trackers.Count, s2.Trackers.Count);
			foreach (EvaluationTracker et2 in s2.Trackers)
				Assert.IsTrue(et2.Id == etA.Id || et2.Id == etB.Id);

			Assert.AreEqual(s.IncentivePools.Count, s2.IncentivePools.Count, "Incentive pools were not properly persisted.");
			Assert.IsTrue(s2.IncentivePools.Contains(new KeyValuePair<string, IncentiveCollection>(poolKey, pool)));
		}
	}
}
